
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
taudatabase.py: 
"""

from enums import TauEventType
import taumodel

DFT_DATABASE_DESCRIPTION = "A database"

class TauDatabaseQuery(taumodel.TauModel):
    
    def __init__(self,complete_name, parent=None):
        self.call__init__(taumodel.TauModel, complete_name, None)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauModel implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    

    @classmethod
    def buildModelName(cls, parent_model, relative_name):
        """build an 'absolute' model name from the parent model and the 'relative'
        name. 
        - If parent_model is a TauDatabase, the return is a composition of
        the database model name and is device name
        - If parent_model is a TauDatabaseQuery, the relative name is ignored and
        the parent name is returned
        """
        if parent_model is None:
            return relative_name
        parent_name = parent_model.getFullName()
        if not parent_name:
            return relative_name
        # if the parent type is the same as this type, just use the parent
        if isinstance(parent_model, cls):
            return parent_name
        return '%s%s' % (parent_name,relative_name)   
    
    @classmethod
    def getNameValidator(cls):
        import tauvalidator
        return tauvalidator.DatabaseQueryValidator()    
    
    def getValueObj(self,cache=True):
        if cache and self.db_value:
            return self.db_value
        try:
            if self.params:
                self.db_value = self.query(*self.params)
            else:
                self.db_value = self.query()
        except:
            self.debug("invalid query")
            self.traceback()
        return self.db_value
    
    def getDisplayValue(self,cache=True):
        return str(self.getValueObj())

    def getDisplayDescription(self,cache=True):
        return self.getSimpleName()
        
    def addListener(self, listener):
        ret = taumodel.TauModel.addListener(self, listener)
        if not ret:
            return ret
        self.fireEvent(TauEventType.Change, self.getValueObj(), listener)
        return ret
    
    def getChildObj(self,child_name):
        return None

    
class TauDatabase(taumodel.TauModel):
    
    def __init__(self, complete_name, parent=None):
        self._descr = None
        self.call__init__(taumodel.TauModel, complete_name, parent)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauModel implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    
    def cleanUp(self):
        self.trace("[TauDatabase] cleanUp")
        taumodel.TauModel.cleanUp(self)
    
    @classmethod
    def buildModelName(cls, parent_model, relative_name):
        """build an 'absolute' model name from the parent name and the 'relative'
        name. parent_model is ignored since there is nothing above the Database
        object"""
        return relative_name
    
    @classmethod
    def getNameValidator(cls):
        import tauvalidator
        return tauvalidator.DatabaseNameValidator()
    
    def getDescription(self,cache=True):
        if self._descr is None or not cache:
            try:
                self._descr = self.info()
            except:
                self._descr = self._getDefaultDescription()
        return self._descr

    def getDisplayValue(self,cache=True):
        return self.getDisplayDescription(cache)

    def getDisplayDescription(self,cache=True):
        return self.getFullName()
    
    def getDisplayDescrObj(self,cache=True):
        obj = []
        obj.append(('name', self.getDisplayName(cache=cache)))
        descr = self.getDescription(cache=cache)
        obj.append(('description', descr))
        return obj
    
    def addListener(self, listener):
        ret = taumodel.TauModel.addListener(self, listener)
        if not ret:
            return ret
        self.fireEvent(TauEventType.Change, self.getDisplayValue(), listener)
        return ret

    def getChildObj(self,child_name):
        if not child_name:
            return None
        if child_name.startswith('?query'):
            full_name = TauDatabaseQuery.buildModelName(self.getModelObj(),child_name)
            return self.factory().getDatabaseQuery(full_name)
        return self.getDevice(child_name)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Device access methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    
    
    def getDevice(self, devname):
        """Returns the device object given its name"""
        import taudevice
        return self.factory().getObject(taudevice.TauDevice, devname)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to overwrite
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    
    def getElementAlias(self, full_name):
        raise RuntimeError("TauDatabase::getElementAlias cannot be called")
        
    def getElementFullName(self, alias):
        raise RuntimeError("TauDatabase::getElementFullName cannot be called")
    
    def _getDefaultDescription(self):
        return DFT_DATABASE_DESCRIPTION