
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
taudevice.py: 
"""

import weakref
import sys

from enums import TauSWDevState, TauEventType
import taumodel

DFT_DEVICE_DESCRIPTION = "A device"

class TauDeviceQuery(taumodel.TauModel):
    
    def __init__(self,dev,query,params):
        self.dev = dev
        self.query_name = query
        if params:
            self.params = params.split('?param=')[1:]
            if not self.params:
                self.params = None
        else:
            self.params = None
        self.query_value_cache = None
        
        complete_name = '%s%s' % (self.dev.getFullName(),query)
        if self.params:
            complete_name += params
        
        self.call__init__(taumodel.TauModel, complete_name, None)
        
        self.query = None
        try:
            self.query = getattr(self.dev,self.query_name)
        except:
            self.debug("Query '%s' is not available in Device" % query)
            raise
                        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauModel implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    

    @classmethod
    def buildModelName(cls, parent_model, relative_name):
        """build an 'absolute' model name from the parent model and the 'relative'
        name. 
        - If parent_model is a TauDatabase, the return is a composition of
        the database model name and is device name
        - If parent_model is a TauDevice, the relative name is ignored and
        the parent name is returned
        """
        if parent_model is None:
            return relative_name
        parent_name = parent_model.getFullName()
        if not parent_name:
            return relative_name
        if isinstance(parent_model, cls):
            return parent_name
        return '%s%s' % (parent_name,relative_name)  
        
    @classmethod
    def getNameValidator(cls):
        import tauvalidator
        return tauvalidator.DeviceQueryValidator()    
    
    def getValueObj(self,cache=True):
        if cache and self.query_value_cache:
            return self.query_value_cache
        try:
            if self.params:
                self.query_value_cache = self.query(*self.params)
            else:
                self.query_value_cache = self.query()
        except:
            self.debug("invalid query")
            self.traceback()
        return self.query_value_cache
    
    def getDisplayValue(self,cache=True):
        return str(self.getValueObj())
        
    def addListener(self, listener):
        ret = taumodel.TauModel.addListener(self, listener)
        if not ret:
            return ret
        self.fireEvent(taumodel.TauEventType.Change, self.getValueObj(), listener)
        return ret
    
    def getChildObj(self,child_name):
        return None


class TauDevice(taumodel.TauModel):
    
    SHUTDOWNS = (TauSWDevState.Shutdown, TauSWDevState.Crash, 
                 TauSWDevState.EventSystemShutdown)

    """A Device object representing an abstraction of the PyTango.DeviceProxy
       object in the tau.core layer"""

    def __init__(self, name, **kw):
        """Object initialization."""
        parent = kw.pop('parent', None)
        storeCallback = kw.pop('storeCallback', None)
        self.__dict__.update(kw)
        self.call__init__(taumodel.TauModel, name, parent)
        
        self._deviceObj = self._createHWObject()
        self._deviceStateObj = None

        self._descr = None
        self._deviceSwState = self.decode(TauSWDevState.Uninitialized)
        
        if storeCallback:
            storeCallback(self)

    def cleanUp(self):
        self.trace("[TauDevice] cleanUp")
        self._deviceObj = None
        self._descr = None
        self._deviceSwObj
        if not self._deviceStateObj is None:
            self._deviceStateObj.removeListener(self)
        self._deviceStateObj = None
        taumodel.TauModel.cleanUp(self)

    # Export the DeviceProxy interface into this object.
    # This way we can call for example read_attribute on an object of this class
    def __getattr__(self, name):
        if not self._deviceObj is None: 
            return getattr(self._deviceObj,name)
        return None
    
    # Export the 'act like dictionary' feature of PyTango.DeviceProxy
    def __getitem__(self, key):
        attr = self.getAttribute(key)
        return attr.read()

    def __setitem__(self, key, value):
        attr = self.getAttribute(key)
        return attr.write(value)

    def __contains__(self, key):
        hw = self.getHWObj()
        if hw is None:
            return False
        return hw.__contains__(key)
    
    def getHWObj(self):
        return self._deviceObj
    
    def getStateObj(self):
        if self._deviceStateObj is None:
            self._deviceStateObj = self.factory().getAttribute("%s/state" % self.getFullName())
        return self._deviceStateObj

    def getState(self, cache=True):
        stateAttrValue = self.getStateObj().getValueObj(cache=cache)
        if not stateAttrValue is None:
            return stateAttrValue.value
        return None
        
    def getSWState(self, cache=True):
        return self.getValueObj(cache=cache).value

    def getAttribute(self, attrname):
        """Returns the attribute object given its name"""
        
        slashnb = attrname.count('/')
        if slashnb == 0:
            attrname = "%s/%s" % (self.getFullName(), attrname)
        import tauattribute
        return self.factory().getObject(tauattribute.TauAttribute,attrname)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory implementation in sub class
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    
    def _createHWObject(self):
        raise RuntimeError("TauDevice::_createHWObject cannot be called")

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauModel implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-    

    @classmethod
    def buildModelName(cls, parent_model, relative_name):
        """build an 'absolute' model name from the parent model and the 'relative'
        name. 
        - If parent_model is a TauDatabase, the return is a composition of
        the database model name and is device name
        - If parent_model is a TauDevice, the relative name is ignored and
        the parent name is returned
        """
        if parent_model is None:
            return relative_name
        parent_name = parent_model.getFullName()
        if not parent_name:
            return relative_name
        if isinstance(parent_model, cls):
            return parent_name
        return '%s/%s' % (parent_name,relative_name)   
    
    @classmethod
    def getNameValidator(cls):
        import tauvalidator
        return tauvalidator.DeviceNameValidator()

    def getValueObj(self, cache=True):
        if not self.hasListeners() or not cache:
            try:
                v = self.getStateObj().read(cache=cache)
            except Exception:
                v = sys.exc_info()[1]
            self._deviceSwState = self.decode(v)
        return self._deviceSwState
    
    def getDisplayValue(self,cache=True):
        return TauSWDevState.whatis(self.getValueObj(cache).value)

    def getDisplayDescrObj(self,cache=True):
        obj = []
        obj.append(('name', self.getDisplayName(cache=cache)))
        descr = self.getDescription(cache=cache)
        if descr.lower() != self._getDefaultDescription().lower():
            obj.append(('description', descr))
        obj.append(('device state', self.getStateObj().getDisplayValue()) or self.getNoneValue())
        obj.append(('SW state', self.getDisplayValue()))
        return obj
    
    def getDescription(self,cache=True):
        if self._descr is None or not cache:
            try:
                self._descr = self.description()
            except:
                self._descr = self._getDefaultDescription()
        return self._descr
    
    def removeListener(self, listener):
        ret = taumodel.TauModel.removeListener(self, listener)
        if not ret or self.hasListeners():
            return ret # False, None or True
        return self.getStateObj().removeListener(self)
    
    def addListener(self, listener):      
        weWereListening = self.hasListeners()
        ret = taumodel.TauModel.addListener(self, listener)
        if not ret:
            return ret
        
        # We are only listening to State if someone is listening to us
        if weWereListening:
            # We were listening already, so we must fake an event to the new
            # subscrived listener with the current value
            self.fireEvent(taumodel.TauEventType.Change, self.getValueObj(), listener)
        else:
            # We were not listening to events, but now we have to
            self.getStateObj().addListener(self)
        return ret
    
    def getChildObj(self,child_name):
        if not child_name:
            return None
        if child_name.startswith('?query'):
            full_name = TauDeviceQuery.buildModelName(self.getModelObj(),child_name)
            return factory().getDeviceQuery(full_name)
        return self.getAttribute(child_name)

    def eventReceived(self, src, event_type, event_value):
        new_sw_state = TauSWDevState.Uninitialized

        if event_type == TauEventType.Config:
            return
        value = self.decode(event_value)

        if value.value != self._deviceSwState.value:
            self.debug("SW Device State changed %s -> %s" % (TauSWDevState.whatis(self._deviceSwState.value), TauSWDevState.whatis(new_sw_state)))
            self._deviceSwState = value
            self.fireEvent(TauEventType.Change, value)


    def _getDefaultDescription(self):
        return DFT_DEVICE_DESCRIPTION

