
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
safeeval.py: Safe eval replacement with whitelist support
"""



SAFEEVAL_USE_MATH=True   #Turning this to False may improve loading time if you don't need to evaluate math functions
SAFEEVAL_USE_NUMPY=True  #Turning this to False may improve loading time if you don't need to evaluate array functions

if SAFEEVAL_USE_MATH: from math import *
if SAFEEVAL_USE_NUMPY:
    from numpy import *
    from numpy.random import *




class SafeEvaluator:
    """This class provides a safe eval replacement. 
    
    The method eval() will only evaluate the expressions considered safe (whitelisted).
    By default it has a whitelist of mathematical expressions that can be turn off using defaultSafe=False at init
    
    The user can add more safe functions passing a safedict to the addSafe() or init methods.
    
    Functions can be removed by name using removeSafe()
    
    Note: In order to use variables defined outside, the user must explicitelly declare them safe. 
    """
    def __init__(self, safedict={}, defaultSafe=True):
        self._default_safe_list = ['abs','array','arange','acos', 'asin', 'atan', 'atan2', 'ceil', 
                                   'cos', 'cosh', 'degrees', 'e', 'exp', 'fabs', 'floor', 'fmod', 
                                   'frexp', 'hypot', 'ldexp', 'log', 'log10', 'modf', 'pi', 'pow', 
                                   'radians','randn','rand', 'sin', 'sinh', 'sqrt', 'tan', 'tanh'] 
        if defaultSafe: 
            self.safe_list=self._default_safe_list
        else:
            self.safe_list=[]
        self.safe_dict = dict([ (k, globals().get(k, None)) for k in self.safe_list ])
        self.addSafe(safedict)
        
    def eval(self,expr):
        """safe eval"""
        return eval(expr, {"__builtins__":None}, self.safe_dict) 
    
    def addSafe(self,safedict):
        """The values in safedict will be evaluable (whitelisted)
        The safedict is as follows: {"eval_name":object, ...}. The evaluator will interpret eval_name as object.
        """
        self.safe_dict.update(safedict)
    
    def removeSafe(self, name):
        """Removes an object from the whitelist"""
        self.safe_dict.pop(name)






##-----------------------------------------------##
#A demo of use

if __name__ == '__main__':
    
    x=range(6)
    sev=SafeEvaluator()
    print "trying to evaluate a variable that has not been registered"
    try:
        print sev.safeEval('x+2')  #This will fail because 'x' is not registered in sev
    except:
        print "failed!!" 
   
    sev.addSafe({'x':x}) #After registering x, we can use it...
    f0='x'
    f1='sqrt(x)'
    f2='pow(2,8)'
    f3='ceil(array(x)/2.)'
    f4='x[3]*2'
    f5='open("/etc/passwd")' #This is something we do not want to be evaluated
    
    
    for f in [f0,f1,f2,f3,f4,f5]:
        print 'Evaluating "%s":'%f
        try:
            print sev.eval(f)
        except: 
            print 'ERROR: %s cannot be evaluated'%f
    
    vector=arange(6)
    sev2=SafeEvaluator({'x':x, 'y':vector},defaultSafe=False) #Another way of registering a variable is using the init method...
    print 'x*y=',sev2.eval('x*y')
    y=0 #note that the registered variable is local to the evaluator!!
    print 'x*y=',sev2.eval('x*y')  #here, y still has the previously registered value instead of 0
    