
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
actions.py: 
"""

import xml.dom.minidom

from PyQt4 import QtCore, QtGui, Qt

class TauMenu(QtGui.QMenu):
    """Base class for Tau Menus"""
    
    def __init__(self, parent):
        QtGui.QMenu.__init__(self, parent)
    
    def build(self, data):
        m_node = xml.dom.minidom.parseString(data).childNodes[0]
        self.buildFromXML(m_node)
    
    def getActionFactory(self):
        import actionfactory
        return actionfactory.ActionFactory()
        
    def buildFromXML(self, m_node):
        widget = self.parent()
        if m_node.hasAttribute('label'):
            self.setTitle(m_node.getAttribute('label'))
            
        for node in m_node.childNodes:
            name = node.nodeName
            if name == 'Menu':
                m = self.getActionFactory().buildMenu(widget, node)
                if m:
                    self.addMenu(m)
            else:
                a = self.getActionFactory().buildAction(widget, node)
                if a:
                    self.addAction(a)

import taubase

class TauAction(QtGui.QAction):
    """Base class for Tau Actions"""
    
    def __init__(self, parent):
        
        if (parent is None) or \
            (not isinstance(parent, QtGui.QWidget)) or \
            (not isinstance(parent, taubase.TauBaseComponent)):
            raise RuntimeError("Invalid parent. Must be a valid Tau widget.") 
        
        QtGui.QAction.__init__(self, parent)
        
        self.connect(parent, QtCore.SIGNAL('modelChanged(const QString &)'), self.modelChanged)
        self.connect(self, QtCore.SIGNAL("triggered()"), self.actionTriggered)
        
        self.update()
        
    def update(self):
        model = self.parent().getModelObj()
        self.setDisabled(model is None)
        
    @QtCore.pyqtSignature("modelChanged(const QString &)")
    def modelChanged(self, modelName):
        self.update()
        
    @QtCore.pyqtSignature("triggered()")
    def actionTriggered(self):
        pass


class SeparatorAction(TauAction):
    
    menuID = "_Separator_"
    
    def __init__(self, parent=None):
        QtGui.QAction.__init__(self, parent)
        self.setSeparator(True)


class AttributeHistoryAction(TauAction):
    
    menuID = "AttrHistory"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("Show history...")
        
    def actionTriggered(self):
        dialog = QtGui.QDialog()
        dialog.exec_()
    

class AttributeRangesAction(TauAction):
    
    menuID = "AttrRanges"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("Ranges...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        dialog = QtGui.QDialog()
        dialog.exec_()
        

class AttributeUnitsAction(TauAction):
    
    menuID = "AttrUnits"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("Units...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        dialog = QtGui.QDialog()
        dialog.exec_()


class AttributeDisplayAction(TauAction):
    
    menuID = "AttrDisplay"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("Display...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        dialog = QtGui.QDialog()
        dialog.exec_()


class AttributeAllConfigAction(TauAction):
    
    menuID = "AttrConfig"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("All...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        import dialog
        d = dialog.AttributeConfigDialog()
        d.setModel(tau_widget.getModelName())
        d.exec_()


class AttributeMonitorDeviceAction(TauAction):
    
    menuID = "MonitorDevice"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("&Monitor device ...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        model_name = tau_widget.getModelName()
        w = model_name.split("/")
        if len(w) == 3:
            dev_name = model_name
        elif len(w) == 4:
            dev_name = w[0] + "/" + w[1] + "/" + w[2]
        elif len(w) == 5:#FB: If the first parameter is the database
            dev_name = w[1] + "/" + w[2] + "/" + w[3]
        else:
            return
        
        import os
        cmd = "atkpanel " + dev_name + " &"
        os.system(cmd)
        
        
class AttributeImageDisplayAction(TauAction):
    
    menuID = "ImageDisplay"
    
    def __init__(self, parent=None):
        TauAction.__init__(self, parent)
        self.setText("&Image Display ...")
        
    def actionTriggered(self):
        tau_widget = self.parent()
        model_name = tau_widget.getModelName()
        w = model_name.split("/")
        if len(w) == 3:
            dev_name = model_name
        elif len(w) == 4:
            dev_name = w[0] + "/" + w[1] + "/" + w[2]
        else:
            return
        
        attr_name = dev_name + "/Image"
        
        import qub
        im = qub._TauQubDataImageDisplay(data=attr_name)
        im.show()
       

class AttributeMenu(TauMenu):
    
    menuID = "AttrMenu"
    menuData = "<Menu label='Attribute'>" \
    "<MenuItem class='AttrHistory'/>"\
    "<MenuItem class='_Separator_'/>"\
    "<Menu class='AttrConfig'/>"\
    "</Menu>"
    
    def __init__(self, parent):
        TauMenu.__init__(self, parent)
        self.build(self.menuData)
        
        
class ConfigurationMenu(TauMenu):
    menuID = "AttrConfigMenu"
    menuData = "<Menu label='Configuration'>" \
    "<MenuItem class='AttrConfig'/>"\
    "<MenuItem class='_Separator_'/>"\
    "<MenuItem class='AttrRanges'/>"\
    "<MenuItem class='AttrUnits'/>"\
    "<MenuItem class='AttrDisplay'/>"\
    "<MenuItem class='_Separator_'/>"\
    "</Menu>"
    
    def __init__(self, parent):
        TauMenu.__init__(self, parent)
        self.build(self.menuData)
        
