
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
buttons.py: 
"""



from PyQt4 import Qt
#import tau
#tau.setLogLevel(tau.Trace)     
from tau.widget import TauBaseWidget
from tau.core.utils import EventFilters
import PyTango

class TauLauncherButton(Qt.QPushButton, TauBaseWidget):
    '''This class provides a button that launches a modeless dialog containing
    an specified Tau widget which gets the same model as the button.
    The button does not use the model directly. Instead it passes it to the
    associated widget.
    
    Code examples::
        
        # a button that launches a TauAttrForm when clicked
        button =  TauLauncherButton(widget = TauAttrForm())
        button.setModel('a/b/c') #a device name, which will be given to the TauAttrForm when clicking
        
        # a button that launches a tauValueLabel (whose model is an attribute: 'a/b/c/attrname')
        button =  TauLauncherButton(widget = TauValueLabel())
        button.setModel('a/b/c/attrname') #a device name, which will be given to the TauValueLabel when clicking
        
        #same as the previous one, but using the parent model and putting a custom text and icon:
        button =  TauLauncherButton(widget = TauValueLabel(), text='click me', icon=':/tau.png')
        button.setUseParentModel(True)  #let's assume that the button's parent has a model of type "/a/b/c"
        button.setModel('/attrname')
    
    '''
    def __init__(self, parent = None, designMode = False, widget=None, icon = None, text = None):
        '''Constructor
        
        :param parent: (Qt.QWidget or None) parent of this widget
        :param designMode: (bool) flag for Qt designer
        :param widget: (Qt.QWidget) a QWidget that will be shown when clicking
                       in the button
        :param icon: (Qt.QIcon) icon for the button
        :param text: (str) the button text (if None passed, the widget's class name
                     is used) '''
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TauBaseWidget, name, designMode=designMode)
        if icon is not None: self.setIcon(Qt.QIcon(icon))
        self._text = text
        self.setWidget(widget)
        self.connect(self, Qt.SIGNAL('clicked()'), self.onClicked)
        self._dialog = None
        self.setDefault(False)
        self.setAutoDefault(False)
        self.insertEventFilter(EventFilters.IGNORE_CHANGE_AND_PERIODIC) #no need to listen to change events!
    
    def getModelClass(self):
        '''see :meth:`TauBaseComponent.getModelClass`. Note that in the case of
        :class:`TauLauncherButton`, the class is completely dependent on the
        widget's class'''
        try:
            return self._widget.getModelClass()
        except:
            #return None  #@TODO: Uncommenting this avoids the exception when TauBaseWidget.getModelClass chokes with relative classes. But the thing should be solved at TauBaseWidget.getModelClass level
            return TauBaseWidget.getModelClass(self)
    
    def setText(self, text):
        '''Sets the text of the button. see :meth:`Qt.QPushButton.setText`'''
        self._text = text
        Qt.QPushButton.setText(self, text)
    
    def setWidget(self, widget):
        '''sets the widget that will be shown when clicking the button
        
        :param widget: (Qt.QWidget)
        '''
        self._widget = widget
        if self._text is None and self._widget is not None:
            self._text = self._widget.__class__.__name__
        
    def displayValue(self,v):
        '''see :meth:`TauBaseComponent.displayValue`'''
        if self._text is not None: return self._text #make sure the text is not changed once set
        TauBaseWidget.displayValue(self,v)
        
    def getDisplayValue(self):
        '''see :meth:`TauBaseComponent.getDisplayValue`'''
        if self._text is not None: return self._text
        modelobj = self.getModelObj()
        if modelobj is None:
            return '---'
        config = modelobj.getConfig()
        if config.isSpectrum():
            return 'Spect[%i]' % modelobj.read().dim_x
        elif config.isImage():
            return 'Imag[%ix%i]'%(modelobj.read().dim_x,modelobj.read().dim_y)
        else:
            self.debug('Unknown model type for TauLauncherButton')
            return '---'
        
    @Qt.pyqtSignature("clicked()")        
    def onClicked(self):
        '''
        Slot called when the button is clicked.
        Note that the dialog will only be created once. Subsequent clicks on
        the button will only raise the existing dialog'''
        self._widget.setModel(self.getModelName())
        
        if self._dialog is None:
            self._dialog = Qt.QDialog(self, Qt.Qt.WindowTitleHint)
            self._dialog.setWindowTitle(str(self.getModel()))
            layout = Qt.QVBoxLayout(self._dialog)
            layout.addWidget(self._widget)
            #dialog.resize(400,300)
        
        self._dialog.show()
        self._dialog.raise_()
            
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    #                      Qt Properties                        #
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    Model = Qt.pyqtProperty("QString", TauBaseWidget.getModel, TauBaseWidget.setModel, TauBaseWidget.resetModel)    
    UseParentModel = Qt.pyqtProperty("bool", TauBaseWidget.getUseParentModel, TauBaseWidget.setUseParentModel, TauBaseWidget.resetUseParentModel)
        
class TauCommandButton(Qt.QPushButton, TauBaseWidget):
    '''This class provides a button that executes a tango command on its device.
    
    Code examples::
        
        # a button that executes the "status" command for the 'a/b/c' device server
        button =  TauCommandButton(command = 'Status', icon=':/tau.png')
        button.setModel('a/b/c')
        
        #  a button that executes the "exec" command for the 'a/b/c' device server with one parameter
        button =  TauCommandButton(command = 'Status', params=['2+2'],icon=':/tau.png')
        button.setModel('a/b/c')
        
    .. seealso:: :class:`TauCommandsForm` provides a good example of use of
                 TauCommandButton (including managing the return value) '''
    __pyqtSignals__ = ("commandExecuted()",)
    def __init__(self, parent = None, designMode = False, command=None, parameters=None, icon = None, text = None):
        '''Constructor

        :param parent: (Qt.QWidget or None) parent of this widget
        :param designMode: (bool) flag for Qt designer
        :param command: (str) the name of the command to be executed
        :param icon: (Qt.QIcon) icon for the button
        :param text: (str) the button text (if None passed, `command` is used)
        '''
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TauBaseWidget, name, designMode=designMode)
        self._command = command
        self._parameters = parameters
        if icon is not None: self.setIcon(Qt.QIcon(icon))
        self.setCustomText(text)
        self.setDefault(False)
        self.setAutoDefault(False)
        self.connect(self, Qt.SIGNAL('clicked()'), self.onClicked)
        
    def getDisplayValue(self):
        '''see :meth:`TauBaseComponent.displayValue`'''
        if self._customText is not None: return self._customText
        if self._command is None:
            return '---'
        modelobj = self.getModelObj()
        if modelobj is None or not hasattr(modelobj, self._command):
            return '---'
        return self._command
    
    @Qt.pyqtSignature("clicked()")
    def onClicked(self):
        '''Slot called when the button is clicked. It executes the command with
        parameters. It may issue a warning if the command is flagged as
        dangerous. On successful execution, it returns the command result and it
        emits a "commandExecuted" signal with the result as well.
        
        :return: The result of the command. The type depends on the command. It
                 may be None.
        
        .. seealso:: :meth:`setCommand`, :meth:`setParameters`, :meth:`TauBaseComponent.isDangerous`
        '''
        self.debug("launch command %s"%str(self._command))
        if self._command is None: return
        modelobj = self.getModelObj()
        if modelobj is None or not hasattr(modelobj, self._command):
            self.warning('Device %s does not implement command %s'%(modelobj, self._command))
            return
        
        if self.isDangerous() and not self.getForceDangerousOperations():
            result = Qt.QMessageBox.question(self, "Potentially dangerous action",
                                         "%s\nProceed?"%self.getDangerMessage(),
                                         Qt.QMessageBox.Ok|Qt.QMessageBox.Cancel,
                                         Qt.QMessageBox.Ok)
            if result != Qt.QMessageBox.Ok:
                return
        #After all the checks, we can finally do the action
        try:
            result = modelobj.command_inout(self._command, self._castParameters(self._parameters, self._command, modelobj))
        except Exception, e:
            self.error('Unexpected error when executing command %s of %s: %s'%(self._command, modelobj.getNormalName(), str(e)))
            raise
            self.traceback()
        self.emit(Qt.SIGNAL('commandExecuted'), result)
        self.info('Execution of command %s returned: %s'%(self._command, str(result)))
        return result
    
    
    def _castParameters(self, params=None, command=None, dev=None):
        '''Internal method used to cast the command paramters to the appropriate
        type required for the given command
        
        :param parameters: (sequence) a sequence of parameters. If None is
                           passed, the currently set parameters are used.
        :param command: (str) the command name. If None is passed, the currently
                        set command is used.
        :param dev: (tau.core.TauDevice) the device on which the command is
                    executed. If None is passed, the current model is used.
        
        :return: (sequence or scalar) a sequence of parameters (or a scalar if only one parameter)
        '''
        if params is None: params = self._parameters
        if command is None: command = self._command
        if dev is None: dev = self.getModelObj()
        
        try:
            param_type = dev.command_query(command).in_type
        except Exception, e:
            self.warning('Cannot get parameters info for command %s:%s'%(command, str(e)))
            return params
        if param_type == PyTango.CmdArgType.DevVoid: return None
        if PyTango.is_int_type(param_type, True): cast_type = int
        elif PyTango.is_float_type(param_type, True): cast_type = float
        elif param_type == PyTango.CmdArgType.DevVarStringArray or param_type == PyTango.CmdArgType.DevString:cast_type = str 
        elif param_type == PyTango.CmdArgType.DevVarBooleanArray or param_type == PyTango.CmdArgType.DevBoolean:cast_type = bool
        else:
            self.info('Unsupported parameters type (%s). Casting to "str"'%str(param_type))
            cast_type = str
        if PyTango.is_scalar_type(param_type):
            return cast_type(params[0])
        else:
            return map(cast_type,params)
        
    @Qt.pyqtSignature("setCommand(QString)")
    def setCommand(self, commandName):
        '''sets the command to be executed when the button is clicked
        
        :param commandName: (str or None) the command name
        '''
        self._command = str(commandName)
        self._setText(self.getDisplayValue())
    
    @Qt.pyqtSignature("getCommand()")    
    def getCommand(self):
        '''returns the command name to be executed when the button is clicked
        
        :return: (str or None) the command name
        '''
        return self._command
    
    @Qt.pyqtSignature("resetCommand()")
    def resetCommand(self):
        '''equivalent to self.setCommand(None)'''
        self.setCommand(None)
    
    @Qt.pyqtSignature("setParameters(QStringList)")
    def setParameters(self, parameters):
        '''
        Sets the parameters to be used on command execution.
        
        :param parameters: (sequence) a sequence of parameters. If the
                           elements of the sequence are not of the right type
                           required for the parameter, an automatic conversion
                           will be attempted on execution time. As a special
                           case, if parameters is a string, it will be splitted
                           on whitespace to obtain a sequence of parameters.
        '''
        if isinstance(parameters,(basestring, Qt.QString)): parameters = str(parameters).split()
        self._parameters = parameters
    
    @Qt.pyqtSignature("getParameters()")    
    def getParameters(self):
        '''returns the parameters to be used on command execution
        
        :param parameters: (sequence)
        '''
        return self._parameters
    
    @Qt.pyqtSignature("resetParameters()")
    def resetParameters(self):
        '''Equivalent to setParameters(None)
        '''
        self.setParameters(None)
        
    @Qt.pyqtSignature("setCustomText(QString)")
    def setCustomText(self, customText=None):
        '''Sets a custom text for the button (by default it is the command name)
        
        :param customText: (str or None) the custom text. If None passed, it
                           will use the command name
        '''
        self._customText = customText
        self._setText(self.getDisplayValue())
    
    @Qt.pyqtSignature("getCustomText()")
    def getCustomText(self, customText=None):
        '''Returns the custom text of the buttom, or None if no custom text is
        used
        
        :param customText: (str or None) 
        '''
        return self._customText
    
    @Qt.pyqtSignature("resetCustomText()")
    def resetCustomText(self):
        '''Equivalent to setCustomText(None)'''
        self.setCustomText(None)
        
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    #                      Qt Properties                        #
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    
    Model = Qt.pyqtProperty("QString", TauBaseWidget.getModel, TauBaseWidget.setModel, TauBaseWidget.resetModel)
    
    UseParentModel = Qt.pyqtProperty("bool", TauBaseWidget.getUseParentModel, TauBaseWidget.setUseParentModel, TauBaseWidget.resetUseParentModel)
    
    Command = Qt.pyqtProperty("QString", getCommand, setCommand, resetCommand)
    
    Parameters = Qt.pyqtProperty("QStringList", getParameters, setParameters, resetParameters)  
    
    DangerMessage = Qt.pyqtProperty("QString", TauBaseWidget.getDangerMessage, TauBaseWidget.setDangerMessage, TauBaseWidget.resetDangerMessage)
    
    CustomText = Qt.pyqtProperty("QString", getCustomText, setCustomText, resetCustomText)  
  
#if __name__ == "__main__":
#    import sys
#    from tau.widget.qwt import TauPlot, TauArrayEditor
#    #from tau.widget import TauAttrForm
#    import tau.widget.resources.qrc_taudesigner_icons
#    
#    app = Qt.QApplication(sys.argv)
#    
#    ##Uncomment the following for testing TauLauncherButton
#    w=TauArrayEditor()
#    form = TauLauncherButton(parent=None, designMode=False, widget = w, icon=':/tau.png', text = 'show')
#    form.setModel('bl97/pc/dummy-03/waveform')
#    #form.setModel('bl97/pc/dummy-03')
#    #form.setModel('bl97/pyattributeprocessor/1/a|bl97/pyattributeprocessor/1/b')
#    
#    ##Uncomment the following for testing TauCommandButton
#    ##form = TauCommandButton(parent=None, designMode=False, command = 'Exec', parameters=['2+2'], icon=':/tau.png', text = None)
#    #form = TauCommandButton(parent=None, designMode=False, command = 'Status', parameters=None, icon=':/tau.png', text = None)
#    #form.setModel('bl97/pc/dummy-03')
#    #form.setDangerMessage('Booo scary command!!\n Maybe you should think twice!')
#    #def f(*a):print a
#    #form.connect(form, Qt.SIGNAL('commandExecuted'),f)
#    form.show()
#    
#    sys.exit(app.exec_())
#
#
