
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
combobox.py: combobox tau widgets
"""

# -*- coding: utf-8 -*-

from PyQt4 import Qt
import PyTango
import tau.core
from tau.widget import TauBaseWidget, TauBaseWritableWidget
from tau.core.utils import EventFilters


class TauValueComboBox(Qt.QComboBox, TauBaseWritableWidget):
    '''This widget shows a combobox that offers a limited choice of values that
    can be set on an attribute.'''
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent = None, designMode = False):
        self._previousModelName = None
        self._lastValueByUser = None
        
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TauBaseWritableWidget, name, designMode=designMode)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Helper methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def _setCurrentIndex(self, index):
        bs = self.blockSignals(True)
        try:
            self.setCurrentIndex(index)
        finally:
            self.blockSignals(bs)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QComboBox overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    # The minimum size of the widget (a limit for the user)
    def minimumSizeHint(self):
        return Qt.QSize(20, 20)

    # The default size of the widget
    def sizeHint(self):
        return Qt.QSize(120, 22)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def preAttach(self):
        TauBaseWritableWidget.preAttach(self)
        Qt.QObject.connect(self, Qt.SIGNAL("currentIndexChanged(int)"),
                               self.writeIndexValue)
        Qt.QObject.connect(self, Qt.SIGNAL("applied()"),
                               self.writeValue)

    def postDetach(self):
        TauBaseWritableWidget.postDetach(self)
        Qt.QObject.disconnect(self, Qt.SIGNAL("currentIndexChanged(int)"),
                                  self.writeIndexValue)
        Qt.QObject.disconnect(self, Qt.SIGNAL("applied()"),
                                  self.writeValue)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWritableWidget overwriting / Pending operations
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getValue(self):
        """
            Get the value that the widget is displaying now, not the value
            of the attribute.
        """
        model = self.getModelObj()
        if model is None:
            return None

        new_value = self.itemData(self.currentIndex())
        if new_value is None:
            return None

        if PyTango.is_int_type(model.data_type):
            new_value, ok = new_value.toInt()
        elif PyTango.is_float_type(model.data_type):
            new_value, ok = new_value.toDouble()
        elif model.data_type in [ PyTango.DevString ]:
            new_value, ok = str(new_value.toString()), True
        elif model.data_type in [ PyTango.DevBoolean ]:
            new_value, ok = new_value.toBool(), True
        else:
            return None
        if not ok:
            return None
        return new_value

    def setValue(self, value):
        """
            Set the value for the widget to display, not the value of the
            attribute.
        """
        index = self.findData(Qt.QVariant(value))
        self._setCurrentIndex(index)
    
    def updateStyle(self):
        if self.hasPendingOperations():
            self.setStyleSheet('TauValueComboBox {color: blue; }')
        else:
            self.setStyleSheet('TauValueComboBox {}')
        super(TauValueComboBox, self).updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # signals, gui events... things related to "write" in the end
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSignature("currentIndexChanged(int)")
    def writeIndexValue(self, index):
        self.emitValueChanged()
        if self.getAutoApply():
            self.emit(Qt.SIGNAL("applied()"))

    def keyPressEvent(self, event):
        if event.key() in [Qt.Qt.Key_Return,Qt.Qt.Key_Enter]:
            self.emit(Qt.SIGNAL("applied()"))
            event.accept()
        else:
            return Qt.QComboBox.keyPressEvent(self,event)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauValueComboBox own interface
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def addValueNames(self, names):
        bs = self.blockSignals(True)
        try:
            for k, v in names:
                self.addItem(k, Qt.QVariant(v))

            # Ok, now we should see if the current value matches any
            # of the newly added names. This is kinda a refresh:
            mv = self.getModelValueObj()
            if mv is not None:
                self.setValue(mv.w_value)
        finally:
            self.blockSignals(bs)
            
        self.emitValueChanged()
        
    def getValueString(self, value, default='UNKNOWN(%s)'):
        """Returns the corresponding string in the combobox out of a value.
        """
        item = self.findData(Qt.QVariant(value))
        if item < 0:
            if '%s' in default:
                return default % str(value)
            else:
                return default
        return str(self.itemText(item))
        
    def teachDisplayTranslationToWidget(self, widget, default='UNKNOWN(%s)'):
        """
            Makes a label object change the displayed text by the corresponding
            value of the combobox. This is implemented for the general case
            and may be not what you expect in some cases (as for example, it
            fires a fake periodic event which may be problematic if these
            are being filtered out).
        """
        # We reimplement label.displayValue so that instead of the normal
        # value it displays the string it has associated in the combobox.
        widget.displayValue = lambda v: self.getValueString(v, default)

        # Simulate a first event. Otherwise the displayValue will be
        # the default, not modified by us
        model = widget.getModelObj()
        if model:
            widget.fireEvent( model,
                             tau.core.TauEventType.Periodic,
                             model.getValueObj()
            )

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString",
                                TauBaseWidget.getModel,
                                TauBaseWidget.setModel,
                                TauBaseWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                         TauBaseWidget.getUseParentModel,
                                         TauBaseWidget.setUseParentModel,
                                         TauBaseWidget.resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool",
                                    TauBaseWritableWidget.getAutoApply,
                                    TauBaseWritableWidget.setAutoApply,
                                    TauBaseWritableWidget.resetAutoApply)
    
    forcedApply = Qt.pyqtProperty("bool", TauBaseWritableWidget.getForcedApply,
                                 TauBaseWritableWidget.setForcedApply,
                                 TauBaseWritableWidget.resetForcedApply)



class TauQueryComboBox(Qt.QComboBox, TauBaseWidget):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self._includeNone = False
        self._noneLabel = '<None>'
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TauBaseWidget, name, designMode=designMode)
    
        self.defineStyle()
    
    def defineStyle(self):
        self.setAutoCompletion(True)
        self.setAutoCompletionCaseSensitivity(Qt.Qt.CaseInsensitive)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_type in [tau.core.TauEventType.Change, tau.core.TauEventType.Periodic]:
            self.clear()
            if evt_value:
                if self.getIncludeNone():
                    self.addItem(self.getNoneLabel())
                if isinstance(evt_value,basestring):
                    self.addItem(evt_value)
                else:
                    self.addItems(evt_value)
        else:
            TauBaseWidget.handleEvent(self, evt_src, evt_type, evt_value)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setIncludeNone(self,yesno):
        self._includeNone = yesno
        
    def getIncludeNone(self):
        return self._includeNone
    
    def resetIncludeNone(self):
        self.setIncludeNone(False)

    def setNoneLabel(self,label):
        self._noneLabel = label
        
    def getNoneLabel(self):
        return self._noneLabel
    
    def resetNoneLabel(self):
        self.setNoneLabel('<None>')
        
    includeNone = Qt.pyqtProperty("bool", getIncludeNone, setIncludeNone, 
                                         resetIncludeNone)

    noneLabel = Qt.pyqtProperty("QString", getNoneLabel, setNoneLabel, 
                                         resetNoneLabel)

    useParentModel = Qt.pyqtProperty("bool", TauBaseWidget.getUseParentModel, 
                                         TauBaseWidget.setUseParentModel, 
                                         TauBaseWidget.resetUseParentModel)
   
    model = Qt.pyqtProperty("QString", TauBaseWidget.getModel, 
                                TauBaseWidget.setModel, 
                                TauBaseWidget.resetModel)
    
    
class TauAttrListComboBox(Qt.QComboBox, TauBaseWidget):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, parent = None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TauBaseWidget, name)
        self.insertEventFilter(EventFilters.IGNORE_CONFIG)
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContents)
        self.defineStyle()
    
    def defineStyle(self):
        """ Defines the initial style for the widget """
        self.updateStyle()

    def sizeHint(self):
        return Qt.QComboBox.sizeHint(self)

    def minimumSizeHint(self):
        return Qt.QComboBox.minimumSizeHint(self)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        return tau.core.TauAttribute
            
    def handleEvent(self, evt_src, evt_type, evt_value):
        self.clear()
        if evt_type == tau.core.TauEventType.Error:
                return
        if not (evt_src is None or evt_value is None) :
            attrList = list(evt_value.value)
            attrList.sort()
            self.addItems(attrList)
            self.updateStyle()
    
    def updateStyle(self):
        self.update()
    
    def setModel(self, m):
        if isinstance(m, Qt.QAbstractItemModel):
            Qt.QAbstractItemView.setModel(self, m)
        else:
            TauBaseWidget.setModel(self, m)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TauBaseWidget.getModel, 
                                           TauBaseWidget.setModel,
                                           TauBaseWidget.resetModel)
                                
    useParentModel = Qt.pyqtProperty("bool",
                                         TauBaseWidget.getUseParentModel, 
                                         TauBaseWidget.setUseParentModel,
                                         TauBaseWidget.resetUseParentModel)


#####################################################################
## Testing
#####################################################################
def tauAttrListTest():
    '''tests tauAttrList. Model: an attribute containing a list of strings'''
    model = sys.argv[1]
    a = Qt.QApplication([])
    w = TauAttrListComboBox()
    w.setModel(model) 
    w.show()
    return a.exec_()
        
        
def TauValueComboboxTest():
    '''tests TauValueCombobox '''
    model = sys.argv[1]
    names = [
        ('name0', 0),
        ('name1', 1),
        ('name2', 2),
        ('name3', 3)
    ]
    a = Qt.QApplication([])
    w = TauValueComboBox()
    w.setModel(model)
    w.addValueNames(names)
    #w.setModel(model)
    #w.autoApply = True
    w.show()
    return a.exec_()

if __name__ == '__main__':
    import sys
    #main = TauValueComboboxTest #uncomment to test TauValueCombobox
    main = tauAttrListTest #uncomment to testtauAttrList
    sys.exit(main())
