
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
configuration.py: 
"""


class BaseConfigurableClass:
    '''
    A base class defining the API for configurable objects.
    
    .. note:: One implicit requisite is that a configurable object must also provide a
              `meth:`objectName` method which returns the object name. This is typically
              fullfilled by inheritting from QObject.
    
    .. seealso:: :class:`TauBaseConfigurableClass`
    '''
    def __init__(self):
        self._supportedConfigVersions = ["__UNVERSIONED__"] #the latest element of this list is considered the current version
        self.resetConfigurableItems()
        
    def createConfig(self, allowUnpickable=False):
        '''
        Returns a dictionary containing configuration information about the
        current state of the object and also the configurations of any other
        object which has been registered with
        :meth:`registerConfigurableItem`
        
        This method should be reimplemented in children classes to support
        class-specific configurations
        
        Hint: The following code allows you to serialize the configuration
        dictionary as a string (which you can store as a QSetting, or as a Tango
        Attribute, provided that allowUnpickable==False)::
        
            import pickle
            c = pickle.dumps(widget.createConfig())  #c is a string that can be stored
                
        :param alllowUnpickable:  (bool) if False the returned dict is
                                  guaranteed to be a pickable object. This is
                                  the default and preferred option because it
                                  allows the serialization as a string that can
                                  be directly stored in a QSetting. If True, this
                                  limitation is not enforced, which allows to
                                  use more complex objects as values (but limits
                                  its persistence).
        
        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).
        
        .. seealso: :meth:`applyConfig` , :meth:`registerConfigurableItem`
        '''
        configdict = {"ConfigVersion":self._supportedConfigVersions[-1],
                      "__pickable__": True}
        #store the configurations for all registered configurable items as well
        configdict["__orderedConfigNames__"] = self.__configurableItemNames
        itemcfgs = {}
        for k,v in self.__configurableItems.iteritems():
            itemcfgs[k] = v.createConfig(allowUnpickable=allowUnpickable)
        configdict["__itemConfigurations__"] = itemcfgs
        
        return configdict
    
    def applyConfig(self, configdict, depth=-1):
        """applies the settings stored in a configdict to the current object.
        
        This method should be reimplemented in children classes to support
        class-specific configurations
        
        :param configdict: (dict)
        :param depth: (int)  If depth = 0, applyConfig will only be called 
                      for this object, and not for any other object registered
                      via :meth:`registerConfigurableItem`. If depth > 0,
                      applyConfig will be called recursively as many times as
                      the depth value. If depth < 0 (default), no limit is
                      imposed to recursion (i.e., it will recurse for as deep as
                      there are registered items)
        
        .. seealso:: :meth:`createConfig`
        """
        if not self.checkConfigVersion(configdict): 
            raise ValueError('the given configuration is of unsupported version')
        #delegate restoring the configuration of any registered configurable item
        if depth != 0:
            itemcfgs = configdict["__itemConfigurations__"]
            for key in configdict["__orderedConfigNames__"]: #we use the sorted item names
                cfg = itemcfgs[key]
                if key in self.__configurableItems:
                    self.__configurableItems[key].applyConfig(cfg, depth=depth-1) #delegate, decrementing the depth 
    
    def resetConfigurableItems(self):
        ''' clears the record of configurable items depending of this object
        
        .. seealso:: :meth:`registerConfigurableItem`
        '''
        self.__configurableItemNames = []
        self.__configurableItems = {}
        
    def registerConfigurableItem(self, item, name=None):
        '''
        Registers the given item as a configurable item which depends of this
        Tau widget. Registered items are expected to implement the following
        methods:
          - `createConfig` (provided by BaseConfigurableClass)
          - `applyConfig` (provided by BaseConfigurableClass)
          - `objectName` (provided by, QObjects)
       
        :param item: (object) The object that should be registered.

        .. note:: the registration order will be used when restoring configurations
        
        .. seealso:: :meth:`unregisterConfigurableItem`, :meth:`createConfig`
        '''
        if name is None: name = str(item.objectName())
        if name in self.__configurableItemNames:
            raise ValueError('registerConfigurableItem: An object with name "%s" is already registered'%name) #abort if duplicated names
        self.__configurableItemNames.append(name)
        self.__configurableItems[name] = item
        
    def unregisterConfigurableItem(self, item):
        '''
        unregisters the given item from the configurable items record. It raises
        an exception if the item is not registered
        
        :param item: (object or str) The object that should be unregistered.
                     Alternatively, the object name can be passed as a python
                     string.
        
        .. seealso:: :meth:`registerConfigurableItem`
        '''
        if isinstance(item,str): name = item
        else: name = str(item.objectName())
        self.__configurableItemNames.remove(name)
        self.__configurableItems.pop(name)

      
    def checkConfigVersion(self, configdict, showDialog=False, supportedVersions=None):
        '''
        Check if the version of configdict is supported. By default, the
        BaseConfigurableClass objects have ["__UNVERSIONED__"] as their list of supported
        versions, so unversioned config dicts will be accepted.
        
        :param configdict: (dict) configuration dictionary to check
        :param showDialog: (bool) whether to show a QtWarning dialog if check
                           failed (false by default)
        :param supportedVersions: (sequence<str>, or None) supported version
                                  numbers, if None given, the versions supported
                                  by this widget will be used (i.e., those
                                  defined in self._supportedConfigVersions)
        
        :return: (bool) returns True if the configdict is of the right version
        '''
        if supportedVersions is None: supportedVersions = self._supportedConfigVersions
        version = configdict.get("ConfigVersion","__UNVERSIONED__")
        if version not in supportedVersions:
            msg = 'Unsupported Config Version %s. (Supported: %s)'%(version, repr(supportedVersions))
            self.warning(msg)
            if showDialog: 
                from PyQt4 import Qt
                Qt.QMessageBox.warning(self, "Wrong Configuration Version", msg, Qt.QMessageBox.Ok)
            return False
        return True
    
    def createQConfig(self):
        '''
        returns the current configuration status encoded as a QByteArray. This
        state can therefore be easily stored using QSettings
        
        :return: (QByteArray) (in the current mplementation this is just a
                 pickled configdict encoded as a QByteArray
        
        .. seealso:: :meth:`restoreQConfig`
        '''
        from PyQt4 import Qt
        import cPickle as pickle
        configdict = self.createConfig(allowUnpickable=False)
        return Qt.QByteArray(pickle.dumps(configdict))
        
    def applyQConfig(self, qstate):
        '''
        restores the configuration from a qstate generated by :meth:`getQState`.
        
        :param qstate: (QByteArray)  
        
        .. seealso:: :meth:`createQConfig`
        '''
        if qstate.isNull(): return
        import cPickle as pickle
        configdict = pickle.loads(qstate.data())
        self.applyConfig(configdict)
        
    def saveConfigFile(self, ofile=None):
        """Stores the current configuration on a file
        
        :param ofile: (file or string) file or filename to store the configuration 
        
        :return: (str) file name used
        """
        import cPickle as pickle
        if ofile is None:
            from PyQt4 import Qt
            ofile = Qt.QFileDialog.getSaveFileName( self, 'Save Configuration', '%s.pck'%self.__class__.__name__, 'Configuration File (*.pck)')
            if ofile.isEmpty(): return
        if not isinstance(ofile,file): ofile=open(ofile,'w')
        configdict=self.createConfig(allowUnpickable=False)
        self.info("Saving current settings in '%s'"%ofile.name)
        pickle.dump(configdict, ofile)
        return ofile.name
        
    def loadConfigFile(self, ifile=None):
        """Reads a file stored by :meth:`saveConfig` and applies the settings
        
        :param ifile: (file or string) file or filename from where to read the configuration
        
        :return: (str) file name used
        """
        import cPickle as pickle
        if ifile is None:
            from PyQt4 import Qt
            ifile = Qt.QFileDialog.getOpenFileName( self, 'Load Configuration', '', 'Configuration File (*.pck)')
            if ifile.isEmpty(): return
        if not isinstance(ifile,file): ifile=open(ifile,'r')
        configdict=pickle.load(ifile)
        self.applyConfig(configdict)
        return ifile.name

class TauBaseConfigurableClass(BaseConfigurableClass):
    ''' 
    A specialized :class:`BaseConfigurableClass` that assumes that the object
    is also deriving from TauBaseComponent, and therefore stores Tau-related
    configurations like the model, etc
    '''
    
    def createConfig(self, allowUnpickable=False):
        '''
        extends :meth:`BaseConfigurableClass.createConfig` to store basic
        TauBaseComponent configurations
                
        :param alllowUnpickable:  (bool) 
        
        :return: (dict<str,object>) configurations (which can be loaded with
                 :meth:`applyConfig` 
        
        .. seealso: :meth:`applyConfig` , :meth:`registerConfigurableItem`, 
                    :meth:`BaseConfigurableClass.createConfig`
        '''
        configdict = BaseConfigurableClass.createConfig(self, allowUnpickable=allowUnpickable)
        configdict["model"] = self.getModelName()
        configdict["useParentModel"] = self.getUseParentModel()
        configdict["modifiableByUser"] = self.isModifiableByUser()
        return configdict
    
    def applyConfig(self, configdict, **kwargs):
        """
        extends :meth:`BaseConfigurableClass.applyConfig` to restore basic
        TauBaseComponent configurations
        
        :param configdict: (dict)
        
        .. seealso:: :meth:`BaseConfigurableClass.applyConfig` ,
                     :meth:`createConfig`
        """
        BaseConfigurableClass.applyConfig(self, configdict, **kwargs)
        self.setUseParentModel(configdict.get("useParentModel",False))
        #self.setModel(configdict.get("model",None))  #This is commented out because it causes problems with, at least, TauForms
        self.setModifiableByUser(configdict.get("modifiableByUser",False))
        
        

    