
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
containers.py: tau container widgets
"""

from PyQt4 import Qt

import colors
from taubase import TauBaseWidget, TauBaseComponent

class TauBaseContainer(TauBaseWidget):
    """Base class for the Tau container widgets."""
    
    def __init__(self, name, parent=None, designMode = False):
        name = name or self.__class__.__name__
        
        self.call__init__(TauBaseWidget, name, parent, designMode=designMode)

        self.defineStyle()
        self.designMode = designMode
        
    def defineStyle(self):
        self.updateStyle()
        
    def sizeHint(self):
#        if self.designMode:
#            return Qt.QSize(150, 150)
        return Qt.QWidget.sizeHint(self)
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isReadOnly(self):
        return True
    
    def updateStyle(self):
        if self.getShowQuality():
            self.setAutoFillBackground(True)
            #TODO: get quality/state from model and update accordingly
        else:
            self.setAutoFillBackground(False)
            #TODO: restore colors
        TauBaseWidget.updateStyle(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getPendingOperations(self):
        pending_ops = []
        
        for child in Qt.QObject.children(self):
            if isinstance(child, TauBaseComponent):
                pending_ops += child.getPendingOperations()
        return pending_ops
    
    def resetPendingOperations(self):
        self.debug("Reset changes")
        for child in Qt.QObject.children(self):
            if isinstance(child, TauBaseComponent):
                child.resetPendingOperations()
    
    def hasPendingOperations(self):
        ret = False
        for child in Qt.QObject.children(self):
            if isinstance(child, TauBaseComponent):
                ret |= child.hasPendingOperations()
        return ret

    @Qt.pyqtSignature("setModel(QString)")
    def setModel(self,model):
        TauBaseWidget.setModel(self,model)
    
    def handleEvent(self, evt_src, evt_type, evt_value):
        if not self._setText or not self.getShowText():
            return
        modelObj = self.getModelObj()
        if modelObj:
            txt = modelObj.getDisplayName(complete=False)
        else:
            txt = self.getNoneValue()
        self._setText(txt)


class TauWidget(Qt.QWidget, TauBaseContainer):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TauBaseContainer, name, designMode=designMode)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSignature("applyPendingChanges()")
    def applyPendingChanges(self):
        self.applyPendingOperations()
    
    @Qt.pyqtSignature("resetPendingChanges()")
    def resetPendingChanges(self):
        self.resetPendingOperations()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QString", TauBaseContainer.getModel, 
                                TauBaseContainer.setModel, 
                                TauBaseContainer.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauBaseContainer.getUseParentModel, 
                                         TauBaseContainer.setUseParentModel,
                                         TauBaseContainer.resetUseParentModel)
    
    showQuality = Qt.pyqtProperty("bool", TauBaseContainer.getShowQuality, 
                                      TauBaseContainer.setShowQuality, 
                                      TauBaseContainer.resetShowQuality)

                
class TauFrame(Qt.QFrame, TauBaseContainer):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
        
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QFrame, parent)
        self.call__init__(TauBaseContainer, name, designMode=designMode)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSignature("applyPendingChanges()")
    def applyPendingChanges(self):
        self.applyPendingOperations()
    
    @Qt.pyqtSignature("resetPendingChanges()")
    def resetPendingChanges(self):
        self.resetPendingOperations()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QString", TauBaseContainer.getModel, 
                                TauBaseContainer.setModel, 
                                TauBaseContainer.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauBaseContainer.getUseParentModel, 
                                         TauBaseContainer.setUseParentModel,
                                         TauBaseContainer.resetUseParentModel)
    
    showQuality = Qt.pyqtProperty("bool", TauBaseContainer.getShowQuality, 
                                      TauBaseContainer.setShowQuality, 
                                      TauBaseContainer.resetShowQuality)
    
class TauScrollArea(Qt.QScrollArea, TauBaseContainer):
    __pyqtSignals__ = ("modelChanged(const QString &)",)
        
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QScrollArea, parent)
        self.call__init__(TauBaseContainer, name, designMode=designMode)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getPendingOperations(self):
        widget = self.widget()
        if isinstance(widget, TauBaseComponent):
            return widget.getPendingOperations()
        return []
    
    def resetPendingOperations(self):
        self.debug("Reset changes")
        widget = self.widget()
        if isinstance(widget, TauBaseComponent):
            widget.resetPendingOperations()
    
    def hasPendingOperations(self):
        widget = self.widget()
        if isinstance(widget, TauBaseComponent):
            return widget.hasPendingOperations()
        return False
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSignature("applyPendingChanges()")
    def applyPendingChanges(self):
        self.applyPendingOperations()
    
    @Qt.pyqtSignature("resetPendingChanges()")
    def resetPendingChanges(self):
        self.resetPendingOperations()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QString", TauBaseContainer.getModel, 
                                TauBaseContainer.setModel, 
                                TauBaseContainer.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauBaseContainer.getUseParentModel, 
                                         TauBaseContainer.setUseParentModel,
                                         TauBaseContainer.resetUseParentModel)
    
    showQuality = Qt.pyqtProperty("bool", TauBaseContainer.getShowQuality, 
                                      TauBaseContainer.setShowQuality, 
                                      TauBaseContainer.resetShowQuality)
    
    
class TauGroupBox(Qt.QGroupBox, TauBaseContainer):

    __pyqtSignals__ = ("modelChanged(const QString &)", "pendingOperationsChanged(bool)",)
        
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self._prefix = ''
        self._suffix = ''
        
        self.call__init__wo_kw(Qt.QGroupBox, parent)
        self.call__init__(TauBaseContainer, name, designMode=designMode)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    @Qt.pyqtSignature("pendingOperationsChanged(bool)")
    def pendingOperationsChanged(self, yesno):
        self.debug("emiting pendingOperationsChanged(%s)" % str(yesno))
        self.emit(Qt.SIGNAL('pendingOperationsChanged(bool)'), yesno)

    @Qt.pyqtSignature("applyPendingChanges()")
    def applyPendingChanges(self):
        self.applyPendingOperations()
    
    @Qt.pyqtSignature("resetPendingChanges()")
    def resetPendingChanges(self):
        self.resetPendingOperations()
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
                    
    def getDisplayValue(self):
        v = TauBaseContainer.getDisplayValue(self)
        return "%s%s%s" % ((self._prefix or ''), v, (self._suffix or ''))
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getPrefixText(self):
        return self._prefix
    
    @Qt.pyqtSignature("setPrefixText(QString)")
    def setPrefixText(self,prefix):
        self._prefix = prefix
        import tau.core
        self.fireEvent(evt_type = tau.core.TauEventType.Change)

    def getSuffixText(self):
        return self._suffix
    
    @Qt.pyqtSignature("setSuffixText(QString)")
    def setSuffixText(self,suffix):
        self._suffix = suffix
        import tau.core
        self.fireEvent(evt_type = tau.core.TauEventType.Change)

    model = Qt.pyqtProperty("QString", TauBaseContainer.getModel, 
                                TauBaseContainer.setModel, 
                                TauBaseContainer.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauBaseContainer.getUseParentModel, 
                                         TauBaseContainer.setUseParentModel,
                                         TauBaseContainer.resetUseParentModel)
    
    showQuality = Qt.pyqtProperty("bool", TauBaseContainer.getShowQuality, 
                                      TauBaseContainer.setShowQuality, 
                                      TauBaseContainer.resetShowQuality)

    showText = Qt.pyqtProperty("bool", TauBaseContainer.getShowText, 
                                   TauBaseContainer.setShowText, 
                                   TauBaseContainer.resetShowText)
    
    prefixText = Qt.pyqtProperty("QString", getPrefixText, setPrefixText, 
                                     doc="prefix text (optional)")
    
    suffixText = Qt.pyqtProperty("QString", getSuffixText, setSuffixText, 
                                    doc="suffix text (optional)")

        
