
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
forms.py: 
"""

from PyQt4 import Qt
from datetime import datetime

import tau, PyTango
#tau.setLogLevel(tau.Trace)
from tau.widget import TauWidgetFactory, TauWidget, TauScrollArea, TauFrame
from tau.widget import TauButtonBox, TauValue, TauCommandButton
from tau.widget.dialog import AttrChooser

#class NewScrollArea(Qt.QScrollArea):
#    
#    def __init__(self, parent = None):
#        Qt.QScrollArea.__init__(self, parent)
#        
#    def sizeHint(self):
#        widgetSize = self.widget().sizeHint()
##        print "&&&&&&&&&&&&&&&&&&&&&&&"
##        return Qt.QSize(1000, 1000)
#        return Qt.QSize(widgetSize.width() + 20, widgetSize.height() + 20)
#
#
#def _sizeHint(inst):
#    widgetSize = inst.widget().sizeHint()
##    print "&&&&&&&&&&&&&&&&&&&&&&&"
#    return Qt.QSize(widgetSize.width() + 20, widgetSize.height() + 20)


class ParameterCB(Qt.QComboBox):
    '''A custom combobox'''
    def __init__(self, parent=None):
        Qt.QComboBox.__init__(self, parent)
        self.setEditable(True)
        
    def rememberCurrentText(self):
        '''Adds the current text to the combobox items list (unless it is already there)'''
        text = self.currentText()
        if self.findText(text) < 0:
            self.addItem(text)

class TauForm(TauWidget):
    '''A form widget that gets a list of attributes as a model and displays
    a :class:`TauValue` for each of them. By default it automatically shows
    scrollbars if needed and global Apply and Cancel buttons.
    
    You can also see some code that exemplifies the use of TauForm in :ref:`Tau
    coding examples <examples>` '''
    
    def __init__(self, parent = None,
                 formWidget = None,
                 buttons = None,
                 withButtons = True,
#                 withScrolls = True,  
                 designMode = False):
        TauWidget.__init__(self, parent, designMode)
        
        if buttons == None: buttons = Qt.QDialogButtonBox.Apply| \
                                        Qt.QDialogButtonBox.Reset
        self._model = []
        if formWidget is None: formWidget = TauValue
        self._formWidget = formWidget
        self._withButtons = withButtons
#        self._withScrolls = withScrolls
        
        self.setLayout(Qt.QVBoxLayout())
        
        frame = TauWidget()
        frame.setLayout(Qt.QGridLayout())
        
        self.scrollArea = TauScrollArea(self)
        self.scrollArea.setWidget(frame)
        self.scrollArea.setWidgetResizable(True)
        self.layout().addWidget(self.scrollArea)
        self.__attrChooser = None
        
        self.buttonBox = TauButtonBox(buttons = buttons, parent = self)
        self.layout().addWidget(self.buttonBox)
        
        self._connectButtons()
        self._manageButtonBox()
#        self._manageScrolls()

        
        #Actions (they automatically populate the context menu)
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)
        
        self.chooseAttrsAction = Qt.QAction('Modify Contents', self)
        self.addAction(self.chooseAttrsAction)
        self.connect(self.chooseAttrsAction, Qt.SIGNAL("triggered()"), self.chooseAttrs)        
      
    @Qt.pyqtSignature("modelChanged(const QString &)")
    def parentModelChanged(self, parentmodel_name):
        self.info("Parent model changed to '%s'" % parentmodel_name)
        #print "#########", '|', parentmodel_name,'|', self.getUseParentModel()
        parentmodel_name = str(parentmodel_name)
        if self.getUseParentModel():
            #reset the model of childs
            for obj,model in zip(self.getItems(), self.getModel()):
                obj.setModel('%s/%s'%(parentmodel_name, str(model)))
        else:
            self.debug("received event from parent although not using parent model")
    
    def chooseAttrs(self):
        '''launches an attribute chooser dialog to modify the contents of the form'''
        if self.__attrChooser is None: 
            self.__attrChooser = AttrChooser()
            self.connect(self.__attrChooser, Qt.SIGNAL("UpdateAttrs"), self.setModel)
        self.__attrChooser.updateList(self.getModel())
        self.__attrChooser.show()
        
    def sizeHint(self):
        return Qt.QWidget.sizeHint(self)
        
    def _connectButtons(self):
        Qt.QObject.connect(self.buttonBox, Qt.SIGNAL("applyClicked()"), self.apply)
        Qt.QObject.connect(self.buttonBox, Qt.SIGNAL("resetClicked()"), self.reset)
        
    def getModel(self):
        return self._model
    
    @Qt.pyqtSignature("addModels(QStringList)")
    def addModels(self, modelNames):
        '''Adds models to the existing ones:
        
        :param modelNames:  (sequence<str>) the names of the models to be added
        
        .. seealso:: :meth:`removeModels`
        '''

        self.setModel(self.getModel()+modelNames)
    
    @Qt.pyqtSignature("removeModels(QStringList)")
    def removeModels(self, modelNames):
        '''Removes models from those already in the form.
        
        :param modelNames:  (sequence<str>) the names of the models to be removed
        
        .. seealso:: :meth:`addModels`
        '''
        currentModels = self.getModel()
        for name in modelNames:
            try: currentModels.remove(name)
            except: self.warning("'%s' not in model list"%name)
        self.setModel(currentModels)
    
    def setModelCheck(self, model, check=True):
        self.destroyChildren()
        self._model = model
        self.fillWithChildren()
                
    def resetModel(self):
        self.destroyChildren()
        self._model = Qt.QStringList()
        
    def getFormWidget(self):
        return self._formWidget
        
    def setFormWidget(self, formWidget):
        self._formWidget = formWidget
        
    def resetFormWidget(self):
        self._formWidget = TauValue
        
    def isWithButtons(self):
        return self._withButtons
    
    def setWithButtons(self, trueFalse):
        self._withButtons = trueFalse
        self._manageButtonBox()
        
    def resetWithButtons(self):
        self._withButtons = True
        self._manageButtonBox()
    
    def dragEnterEvent(self, event):
        '''reimplemented to support dropping of modelnames'''
        if self.isModifiableByUser() and event.mimeData().hasFormat("text/plain"):
            event.acceptProposedAction()
            
    def dropEvent(self, event):
        '''reimplemented to support dropping of modelnames'''
        modelname = str(event.mimeData().text())
        try:
            self.addModels([modelname])
            event.acceptProposedAction()
        except:
            pass
    
    def setModifiableByUser(self, modifiable):
        '''
        sets whether the user can change the contents of the form
        (e.g., via Modify Contents in the context menu)
        Reimplemented from :meth:`TauWidget.setModifiableByUser`
        
        :param modifiable: (bool)
        
        .. seealso:: :meth:`TauWidget.setModifiableByUser`
        '''
        TauWidget.setModifiableByUser(self, modifiable)
        self.chooseAttrsAction.setEnabled(modifiable)
        self.setAcceptDrops(modifiable)
        for item in self.getItems():
            try: 
                item.setModifiableByUser(modifiable)
            except:
                pass

        
#    def isWithScrolls(self):
#        return self._withScrolls
#    
#    def setWithScrolls(self, trueFalse):
#        self._withScrolls = trueFalse
#        self._manageScrolls()
#        
#    def resetWithScrolls(self):
#        self._withScrolls = False
#        self._manageScrolls()
        
    def setRegExp(self, regExp):
        pass
    
    def getRegExp(self):
        pass
    
    def destroyChildren(self):
        self.resetConfigurableItems()
        oldFrame = self.scrollArea.widget()
        for child in oldFrame.children():
            if isinstance(child, self.getFormWidget()):
                child.destroy()
                
    def fillWithChildren(self):
        frame = TauWidget()
        frame.setLayout(Qt.QGridLayout())
        
        parent_name = None
        if self.getUseParentModel():
            parent_model = self.getParentModelObj()
            if parent_model:
                parent_name = parent_model.getFullName()
        
        for i,model in enumerate(self.getModel()): 
            model = str(model)
            if parent_name: model = "%s/%s" % (parent_name, model)
            widget = self.getFormWidget()(frame)
            widget.setMinimumHeight(20)
            widget.setModel(model)
            widget.setParent(frame)
            try: widget.setModifiableByUser(self.isModifiableByUser())
            except: pass
            widget.setObjectName("item%i"%i)
            self.registerConfigurableItem(widget)
            
        self.scrollArea.setWidget(frame)
#        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setMinimumWidth(frame.layout().sizeHint().width()+20)

    def getItemByModel(self, model, index=0):
        '''returns the child item with given model. If there is more than one item
        with the same model, the index parameter can be used to distinguish among them
        Please note that his index is only relative to same-model items!'''
        for child in self.scrollArea.widget().children():
            if isinstance(child, self.getFormWidget()):
                #@todo tau model comparision?
                if child.getModel().lower() == model.lower():
                    if index <= 0:
                        return child
                    else:
                        index -= 1
    
    def getItemByIndex(self, index):
        '''returns the child item with at the given index position. '''
        return self.getItems()[index]
    
    def getItems(self):
        '''returns a list of the objects that have been created as childs of the form'''
        klass = self.getFormWidget()
        return [obj for obj in self.scrollArea.widget().children() if isinstance(obj,klass)]
    
    def _manageButtonBox(self):
        if self.isWithButtons():
            self.buttonBox.setVisible(True)
        else:
            self.buttonBox.setVisible(False)
            
#    def _manageScrolls(self):
#        if self.isWithScrolls():
#            self.scrollArea.sizeHint = Qt.QScrollArea.sizeHint
#        else:
#            self.scrollArea.sizeHint = _sizeHint
#            self.scrollArea.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Expanding)
    @Qt.pyqtSignature("apply()")        
    def apply(self):
        self.safeApplyOperations()
                
    @Qt.pyqtSignature("reset()")
    def reset(self):
        self.resetPendingOperations()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QStringList", getModel, 
                                                TauWidget.setModel, 
                                                resetModel)

    useParentModel = Qt.pyqtProperty("bool", 
                                         TauWidget.getUseParentModel, 
                                         TauWidget.setUseParentModel,
                                         TauWidget.resetUseParentModel)
    
    withButtons = Qt.pyqtProperty("bool", isWithButtons, 
                                                setWithButtons,
                                                resetWithButtons)
    
    modifiableByUser = Qt.pyqtProperty("bool", TauWidget.isModifiableByUser, 
                                               TauWidget.setModifiableByUser,
                                               TauWidget.resetModifiableByUser)
      
    

class TauCommandsForm(TauWidget):
    '''A form that shows commands available for a Device Server'''
    def __init__(self, parent = None, designMode = False):
        TauWidget.__init__(self, parent, designMode)
        
        self.setLayout(Qt.QVBoxLayout())
                
        self._splitter = Qt.QSplitter()
        self._splitter.setOrientation(Qt.Qt.Vertical)
        self.layout().addWidget(self._splitter)
        
        self._frame = TauWidget(self)
        self._frame.setLayout(Qt.QGridLayout())
        
        self._scrollArea = TauScrollArea(self)
        self._scrollArea.setWidget(self._frame)
        self._scrollArea.setWidgetResizable(True)
        self._splitter.addWidget(self._scrollArea)
        
        self._outputTE = Qt.QTextEdit()
        self._outputTE.setReadOnly(True)
        self._splitter.addWidget(self._outputTE)
        #self._splitter.moveSplitter(self._splitter.getRange(0)[-1], 0)
        
        self._cmdWidgets = []
        self._paramWidgets = []
        self._viewFilters = []
        self._defaultParameters = []
        self._sortKey = lambda x:x.cmd_name
        
        self._operatorViewFilter = lambda x: x.disp_level == PyTango.DispLevel.OPERATOR
        
        #self.setLayout(Qt.QGridLayout())
        self.connect(self, Qt.SIGNAL('modelChanged(const QString &)'),self._updateCommandWidgets )
        
    def createConfig(self, allowUnpickable=False):
        '''
        extending  :meth:`TauBaseWidget.createConfig`                
        :param alllowUnpickable:  (bool) 
        
        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).
        
        .. seealso: :meth:`TauBaseWidget.createConfig`, :meth:`applyConfig` 
        '''
        #get the basic config
        configdict = TauWidget.createConfig(self, allowUnpickable=allowUnpickable)
        #store the splitter config
        configdict['splitter/state'] = self._splitter.saveState().data()        
        return configdict
    
    def applyConfig(self, configdict, **kwargs):
        """extending :meth:`TauBaseWidget.applyConfig` to restore the splitter config
     
        :param configdict: (dict)
        
        .. seealso:: :meth:`TauBaseWidget.applyConfig`, :meth:`createConfig`
        """
        #first do the basic stuff...
        TauWidget.applyConfig(self, configdict, **kwargs)
        #restore the splitter config
        self._splitter.restoreState(Qt.QByteArray(configdict['splitter/state']))
        
    def getModelClass(self):
        '''see :meth:`TauBaseComponent.getModelClass`'''
        return tau.core.TauDevice
        
    def _updateCommandWidgets(self, *args):
        '''
        Inserts command buttons and parameter widgets in the layout, according to
        the commands from the model
        '''
        #self.debug('In TauCommandsForm._updateCommandWidgets())')
        dev = self.getModelObj()
        if dev is None:
            self.error('Cannot connect to device')
            return
        commands = sorted(dev.command_list_query(), key=self._sortKey)
        
        for f in self.getViewFilters():
            commands = filter(f, commands)
        
        self._clearFrame()
        
        layout = self._frame.layout()

        for row,c in enumerate(commands):
            self.debug('Adding button for command %s'%c.cmd_name)
            button = TauCommandButton(command=c.cmd_name,text=c.cmd_name)
            #button.setUseParentModel(True)
            button.setModel(self.getModel())
            layout.addWidget(button, row, 0)
            self._cmdWidgets.append(button)
            self.connect(button, Qt.SIGNAL('commandExecuted'), self._onCommandExecuted)
            
            if c.in_type != PyTango.CmdArgType.DevVoid:
                self.debug('Adding arguments for command %s'%c.cmd_name)
                pwidget = ParameterCB()
                if c.cmd_name.lower() in self._defaultParameters:
                    for par in self._defaultParameters.get(c.cmd_name.lower(),[]): pwidget.addItem(par)
                    pwidget.setEditable(False)
                self.connect(pwidget, Qt.SIGNAL('editTextChanged (const QString&)'),button.setParameters)
                self.connect(pwidget, Qt.SIGNAL('activated (int)'), button.setFocus)
                self.connect(button, Qt.SIGNAL('commandExecuted'), pwidget.rememberCurrentText)
                layout.addWidget(pwidget, row, 1)
                self._paramWidgets.append(pwidget)
    
    def _clearFrame(self):
        '''destroys all widgets in the scroll area frame'''
        self._frame = TauWidget(self)
        self._frame.setLayout(Qt.QGridLayout())
        self._frame.setModel(self.getModelName()) # hack because useParentModel does not work!
        self._scrollArea.setWidget(self._frame)
        self._cmdWidgets = []
        self._paramWidgets = []
        
    def _onCommandExecuted(self, result):
        '''Slot called when the command is executed, to manage the output
        
        :param result: return value from the command. The type depends on the command
        '''
        timestamp = datetime.now()
        cmdbutton = self.sender()
        output = '<i>%s</i><br>'%timestamp.strftime('%Y-%m-%d %H:%M:%S') +\
                 '<b>Command:</b> %s<br>'%cmdbutton.getCommand() +\
                 '<b>Pars:</b> %s<br>'%repr(cmdbutton.getParameters()) +\
                 '<b>Return Value:</b><br>%s'%str(result)
        self._outputTE.append(output)
        separator = '<table width=\"100%\"><tr><td><hr /></td></tr></table>' # ugly workaround for bug in html rendering in Qt
        self._outputTE.append(separator)                                     # see http://lists.trolltech.com/qt-interest/2008-04/thread00224-0.html
        #self._outputTE.append('%s'%timestamp)
        #self._outputTE.append('<b>Command:</b> "%s"'%cmdbutton.getCommand())
        #self._outputTE.append('<b>Pars:</b> %s'%repr(cmdbutton.getParameters()))
        #self._outputTE.append('<b>Return Value:</b><br>%s<hr>'%str(result))
        
    def setSortKey(self, sortkey):
        '''sets the method used to sort the commands (cmd_name by default)
        
        :param sortkey: (callable) a function that takes a :class:`CommandInfo`
                        as argument and returns a key to use for sorting purposes
                        (e.g. the default sortKey is ``lambda x:x.cmd_name``)
        '''
        self._sortKey = sortkey
        self._updateCommandWidgets()

    def setDefaultParameters(self, params):
        '''sets the values that will appear by default in the parameters combo box
        :param params: { 'cmd_name': ['param1, param2', 'param1b, param2b' ] }
        '''
        self._defaultParameters = dict((k.lower(),v) for k,v in params.items())
        self._updateCommandWidgets()

    def setViewFilters(self, filterlist):
        '''sets the filters to be applied when displaying the commands
        
        :param filterlist: (sequence<callable>) a sequence of command filters.
                           All filters will be applied to each command to decide
                           whether to display it or not.
                           for a command to be plotted, the following condition
                           must be true for all filters:
                           bool(filter(command))==True
        '''
        self._viewFilters = filterlist
        self._updateCommandWidgets()
    
    def getViewFilters(self):
        '''returns the filters used in deciding which commands are displayed
        
        :return: (sequence<callable>) a sequence of filters
        '''
        return self._viewFilters
    
    @Qt.pyqtSignature("setCommand(bool)")
    def setExpertView(self,expert):
        '''sets the expert view mode
        
        :param expert: (bool) If expert is True, commands won't be filtered. If
        it is False, commands with  display level Expert won't be shown
        '''
        currentfilters = self.getViewFilters()
        if expert:
            if self._operatorViewFilter in currentfilters:
                currentfilters.remove(self._operatorViewFilter)
        else:
            if self._operatorViewFilter not in currentfilters:
                currentfilters.insert(0, self._operatorViewFilter)
        self.setViewFilters(currentfilters)
        self._expertView = expert

class TauAttrForm(TauWidget):
    '''A form that displays the attributes of a Device Server'''
    def __init__(self, parent = None, designMode = False):
        TauWidget.__init__(self, parent, designMode)
        
        self._viewFilters = []
        self._operatorViewFilter = lambda x: x.disp_level == PyTango.DispLevel.OPERATOR
        
        self.setLayout(Qt.QVBoxLayout())
   
        self._form = TauForm(parent = self)
        self.layout().addWidget(self._form)
        self.registerConfigurableItem(self._form)
        
        self.connect(self, Qt.SIGNAL('modelChanged(const QString &)'),self._updateAttrWidgets )
        
        self._sortKey = lambda x:x.name

    def setSortKey(self, sortkey):
        '''sets the key function used to sort the attributes in the form 
        
        :param sortkey: (callable) a function that takes a :class:`AttributeInfo`
                        as argument and returns a key to use for sorting purposes
                        (e.g. the default sortKey is ``lambda x:x.name``)
        
        '''
        self._sortKey = sortkey
        self._updateAttrWidgets()        
        
    def getModelClass(self):
        '''see :meth:`TauBaseComponent.getModelClass`'''
        return tau.core.TauDevice
    
    def _updateAttrWidgets(self):
        '''Populates the form with TauValues for each of the attributes shown
        '''
        dev = self.getModelObj()
        if dev is None:
            self.error('Cannot connect to device')
            return
        attrlist = sorted(dev.attribute_list_query(), key=self._sortKey)
        for f in self.getViewFilters():
            attrlist = filter(f, attrlist)
        attrnames = []
        devname = self.getModelName()
        for a in attrlist:
            attrnames.append("%s/%s"%(devname,a.name)) #ugly hack . But setUseParentModel does not work well
        self.debug('Filling with attribute list: %s'%("; ".join(attrnames))) 
        self._form.setModel(attrnames)
        
    def setViewFilters(self, filterlist):
        '''sets the filters to be applied when displaying the attributes
        
        :param filterlist: (sequence<callable>) a sequence of attr filters. All
                           filters will be applied to each attribute name to
                           decide whether to display it or not. for an attribute
                           to be plotted, the following condition must be true
                           for all filters: ``bool(filter(attrname))==True``
        '''
        self._viewFilters = filterlist
        self._updateAttrWidgets()
    
    def getViewFilters(self):
        '''returns the filters used in deciding which attributes are displayed
        
        :return: (sequence<callable>) a sequence of filters
        '''
        return self._viewFilters
    
    @Qt.pyqtSignature("setExpertView(bool)")
    def setExpertView(self, expert):
        '''sets the expert view mode
        
        :param expert: (bool) If expert is True, attributes won't be filtered. If
                       it is False, attributes with display level Expert won't
                       be shown
        '''
        currentfilters = self.getViewFilters()
        if expert:
            if self._operatorViewFilter in currentfilters:
                currentfilters.remove(self._operatorViewFilter)
        else:
            if self._operatorViewFilter not in currentfilters:
                currentfilters.insert(0, self._operatorViewFilter)
        self.setViewFilters(currentfilters)
        self._expertView = expert
    
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QString", TauWidget.getModel, 
                                       TauWidget.setModel, 
                                       TauWidget.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauWidget.getUseParentModel, 
                                         TauWidget.setUseParentModel,
                                         TauWidget.resetUseParentModel)



def test1():
    '''tests tauForm'''
    import sys
    if len(sys.argv)>1: models=sys.argv[1:]
    else: models = None
    app = Qt.QApplication(sys.argv)
    if models is None:
        models = ['lab/op/ccd-01/CurrentlySavingMovie',
                  'lab/op/ccd-01/MovieRemainingTime',
                  'lab/op/ccd-01/Image',
                  'lab/op/ccd-01/ExposureTime', 
                  'lab/op/ccd-01/FrameRate', 
                  'lab/op/ccd-01/ResultingFrameRate', 
                  'lab/op/ccd-01/BitDepth', 
                  'lab/op/ccd-01/SensorWidth', 
                  'lab/op/ccd-01/SensorHeight', 
                  'lab/op/ccd-01/ImageCounter', 
                  'lab/op/ccd-01/BlackLevel', 
                  'lab/op/ccd-01/Gain', 
                  'lab/op/ccd-01/TriggerMode', 
                  'lab/op/ccd-01/TriggerLine', 
                  'lab/op/ccd-01/TriggerActivation', 
                  'lab/op/ccd-01/InternalFrameRate', 
                  'lab/op/ccd-01/Overruns', 
                  'lab/op/ccd-01/InternalAcquisitionBuffers', 
                  'lab/op/ccd-01/State', 
                  'lab/op/ccd-01/Status']
    dialog = TauForm()
    dialog.setModel(models)
    for i,tv in enumerate(dialog.getItems()):
        tv.setDangerMessage("Booooo scaring %d!!!"%i)
    dialog.show()

    sys.exit(app.exec_())

def test2():
    '''tests tauAttrForm'''
    import sys
    if len(sys.argv)>1: model=sys.argv[1]
    else: model = None
    
    if model is None: model = 'bl97/pc/dummy-01'
    app = Qt.QApplication(sys.argv)
    dialog = TauAttrForm()
    dialog.setModel(model)
    dialog.show()

    sys.exit(app.exec_())
    
def test3():
    '''tests tauCommandsForm'''
    import sys
    if len(sys.argv)>1: model=sys.argv[1]
    else: model = None
    
    if model is None: model = 'bl97/pc/dummy-01'
    app = Qt.QApplication(sys.argv)
    dialog = TauCommandsForm()
    dialog.setModel(model)
    dialog.show()

    sys.exit(app.exec_()) 

def main():
    #test1()
    test2()
    #test3()
    
if __name__ == "__main__":
    main() 