
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
labels.py: 
"""

__docformat__ = 'restructuredtext'

from PyQt4 import QtCore, QtGui, Qt
import tau

import colors
import taubase

from tau.core.utils import DebugIt

class TauValueLabel(QtGui.QLabel, taubase.TauBaseWidget):
    """A widget that represents the read value of an attribute.

    The following features are present:
    
        * The background color changes with the attribute quality according to the :ref:`color-guide`
        * The attribute value is displayed according to the attribute format
        * Supports all data types and formats (although it is designed mainly for SCALAR format)
        * A tooltip is automatically generated displaying additional attribute information
        * If the size of the widget is too small to display the entire value, the text changes to
          a symbolic link that opens a dialog with the attribute value when clicked
    
    This widget emits the following signals (excluding signals of the super classes):
    
    **modelChanged**:
        - *Signature*: ``modelChanged(const QString &) -> None``
        - *Description*: the signal is emmited when the model for this widget 
          changes either by direct model manipulation, change of the useParentModel 
          and change of the parent model (if useParentModel is set to True)
    
    Example::
    
        import sys
        from PyQt4 import Qt
        import tau.widget

        app = Qt.QApplication(sys.argv)

        w = tau.widget.TauValueLabel()
        w.setModel('sys/tautest/1/position')

        w.setVisible(True)
        sys.exit(app.exec_())

    .. image:: /_static/label01.png
        :align: center

    """
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, parent = None, designMode = False, background = 'quality'):
        name = self.__class__.__name__

        self.call__init__wo_kw(QtGui.QLabel, parent)
        self.call__init__(taubase.TauBaseWidget, name, designMode=designMode)

        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        self.setTextInteractionFlags (Qt.Qt.TextSelectableByMouse|Qt.Qt.LinksAccessibleByMouse)
        self.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self._charWidth = self.font().pointSize()
        self._showState = False
        self._showValueStateAsBackground = False
        self.defineBackground(background)
        self.connect(self, Qt.SIGNAL("linkActivated (const QString&)"), self.showFullValueDialog)
        self._text = ''
    
    def defineBackground(self, background):
        background = str(background).lower()
        self.setShowQuality(background == 'quality')
        self.setShowState(background == 'state')
        self.setShowValueStateAsBackground(background == 'value_state')
        self.__background = background;
        
    # The minimum size of the widget (a limit for the user)
    def minimumSizeHint(self):
        return QtCore.QSize(20, 20)
        
    # The default size of the widget
    def sizeHint(self):
        return QtCore.QSize(60, 22)
        #return QtCore.QSize(QtGui.QLabel.sizeHint(self).width(), 24)
    
    @Qt.pyqtSignature("setShowState(bool)")
    def setShowState(self, showState):
        '''Whether or not to show the device state as background color.
        Note: obviously, setShowState(True) forces setShowQuality(False) !'''
        if showState == self._showState:
            return
        if showState and self.getShowQuality():
            self.setShowQuality(False)
        self._showState = showState
        self.updateStyle()
    
    def getShowState(self):
        """either or not to show the state as a background color"""
        return self._showState

    def resetShowState(self):
        self.setShowState(False)
    
    @Qt.pyqtSignature("setShowValueStateAsBackground(bool)")
    def setShowValueStateAsBackground(self, showState):
        '''Whether or not to show the value as background color.
        The attribute value must be a state!
        Note: obviously, forces setShowState(False) and setShowQuality(False)
    '''
        if showState:
            self.setShowQuality(False)
            self.setShowState(False)
        self._showValueStateAsBackground = showState

    def getShowValueStateAsBackground(self):
        """either or not to show the state as a background color"""
        return self._showValueStateAsBackground

    def resetShowValueStateAsBackground(self):
        self.setShowValueStateAsBackground(False)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        return tau.core.TauAttribute

    def isReadOnly(self):
        return True
    
    def updateStyle(self, extra=''):
        if self.getShowQuality():
            quality = getattr(self.getModelValueObj(), 'quality', None)
            ss = colors.QT_ATTRIBUTE_QUALITY_PALETTE.qtStyleSheet(quality)
        elif self.getShowState():
            try: state = self.getModelObj().getParent().getState()
            except Exception,e:
                state = None
            ss = colors.QT_DEVICE_STATE_PALETTE.qtStyleSheet(state)
        elif self.getShowValueStateAsBackground():
            value = getattr(self.getModelValueObj(), 'value', None)
            ss = colors.QT_DEVICE_STATE_PALETTE.qtStyleSheet(value)
        else:
            ss = "background-color: transparent; color:black;"
        ss = """TauValueLabel {
            border-style: outset;
            border-width: 2px;
            border-color: rgba(255,255,255,128);
            %s %s}""" % (ss, extra)
        self.setStyleSheet(ss)
        taubase.TauBaseWidget.updateStyle(self)

    def validateDisplay(self):
        '''Checks that the display fits in the widget and sets it to "..." if it does not'''
        trimmedtext = "<a href='...'>...</a>"
        self._text = self.text()
        if QtGui.QLabel.sizeHint(self).width() > self.size().width():
            self.setToolTip(self.getFormatedToolTip(showValue=True) )
            self._setText(trimmedtext)
            self.updateStyle()
            return False
        return True
    
    def getFormatedToolTip(self,cache=True, showValue=False):
        ret = taubase.TauBaseWidget.getFormatedToolTip(self,cache=cache)
        if showValue:
            ret = u"<p><b>Value:</b> %s</p><hr>%s"%(self._text, ret)
        return Qt.QString(ret)
    
    def showFullValueDialog(self,*args):
        Qt.QMessageBox.about(self,  "Full text", self._text)
    
    def handleEvent(self, evt_src, evt_type, evt_value):
        '''reimplemented to check that the display fits the size each time that the value changes'''
        taubase.TauBaseWidget.handleEvent(self, evt_src, evt_type, evt_value)
        self.validateDisplay()
    
    def resizeEvent(self,event):
        #We recheck the Display every time we resize
        self.updateStyle()
        self.setText(self._text) #We set the original (untrimmed) text to force a validation from scratch.
        self.validateDisplay()
        QtGui.QLabel.resizeEvent(self,event)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    #: This property holds the unique URI string representing the model name 
    #: with which this widget will get its data from. The convention used for 
    #: the string can be found :ref:`here <model-concept>`.
    #: 
    #: In case the property :attr:`useParentModel` is set to True, the model 
    #: text must start with a '/' followed by the attribute name.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TauBaseWidget.getModel`
    #:     * :meth:`TauBaseWidget.setModel`
    #:     * :meth:`TauBaseWidget.resetModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    model = QtCore.pyqtProperty("QString", taubase.TauBaseWidget.getModel, 
                                taubase.TauBaseWidget.setModel, 
                                taubase.TauBaseWidget.resetModel)
    
    #: This property holds whether or not this widget should search in the 
    #: widget hierarchy for a model prefix in a parent widget.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TauBaseWidget.getUseParentModel`
    #:     * :meth:`TauBaseWidget.setUseParentModel`
    #:     * :meth:`TauBaseWidget.resetUseParentModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    useParentModel = QtCore.pyqtProperty("bool", 
                                         taubase.TauBaseWidget.getUseParentModel, 
                                         taubase.TauBaseWidget.setUseParentModel,
                                         taubase.TauBaseWidget.resetUseParentModel)

    #: This property holds whether or not this widget should fill the background
    #: with the attribute quality/device state according to the 
    #: :ref:`color-guide`
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TauBaseWidget.getShowQuality`
    #:     * :meth:`TauBaseWidget.setShowQuality`
    #:     * :meth:`TauBaseWidget.resetShowQuality`
    #:
    #: .. seealso:: :ref:`color-guide`
    showQuality = QtCore.pyqtProperty("bool", 
                                      taubase.TauBaseWidget.getShowQuality,
                                      taubase.TauBaseWidget.setShowQuality,
                                      taubase.TauBaseWidget.resetShowQuality)
    
    #: This property holds whether or not this widget should display the current
    #: value of the model as text. Setting this to False is useful when you just
    #: want to represent the quality/state as background color in a small space
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TauBaseWidget.getShowText`
    #:     * :meth:`TauBaseWidget.setShowText`
    #:     * :meth:`TauBaseWidget.resetShowText`
    showText = QtCore.pyqtProperty("bool",taubase.TauBaseWidget.getShowText,
                                   taubase.TauBaseWidget.setShowText, 
                                   taubase.TauBaseWidget.resetShowText)
    
    #: This property holds the contents of a the popup menu as an XML string.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TauBaseWidget.getTauPopupMenu`
    #:     * :meth:`TauBaseWidget.setTauPopupMenu`
    #:     * :meth:`TauBaseWidget.resetTauPopupMenu`
    #:
    #: .. seealso:: :ref:`popupmenu-tutorial`
    tauPopupMenu = QtCore.pyqtProperty("QString",
                                       taubase.TauBaseWidget.getTauPopupMenu,
                                       taubase.TauBaseWidget.setTauPopupMenu,
                                       taubase.TauBaseWidget.resetTauPopupMenu)


class TauConfigLabel(QtGui.QLabel, taubase.TauBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent = None, designMode = False):
        name = "TauConfigLabel"
        self._prefix = ''
        self._suffix = ''
        self._configParam = None
        self._extra_style = ''
        self.call__init__wo_kw(QtGui.QLabel, parent)
        self.call__init__(taubase.TauBaseWidget, name, designMode=designMode)
        
        self.setAlignment(QtCore.Qt.AlignVCenter)
        
    def defineStyle(self):
        
        palette = self.palette()
        
        # define a default background color in case the model is not connected
        bg_brush, fg_brush = colors.QT_DEVICE_STATE_PALETTE.qbrush(None)
        palette.setBrush(QtGui.QPalette.Window,bg_brush)
        palette.setBrush(QtGui.QPalette.Base,bg_brush)
        palette.setBrush(QtGui.QPalette.WindowText,fg_brush)
        self.setPalette(palette) # mandatory since PyQt 4.5
        self.updateStyle()

    def sizeHint(self):
        return QtCore.QSize(60, 24)

    def getConfigurationAttributeValue(self):
        """Helper method to get the proper parameter value from the configuration"""
        model = self.getModelObj()
        if model is None:
            return self.getNoneValue()
        try:
            val = getattr(model, str(self._configParam))
            try:
                no_val = getattr(model, "no_" + str(self._configParam))
                if val.lower() == no_val.lower():
                    val = self.getNoneValue()
            except:
                pass
        except:
            self.debug("Invalid configuration parameter %s" % self._configParam)
            val = self.getNoneValue()
        return val

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        return tau.core.TauConfiguration

    def getDisplayValue(self):
        cfg_value = self.getConfigurationAttributeValue()
        cfg_value = (self._prefix or '') + (cfg_value or '') + (self._suffix or '')
        return cfg_value

    def getFormatedToolTip(self,cache=True):
        obj = self.getModelObj()
        if obj is None:
            return Qt.QString.fromLatin1(self.getNoneValue())
        p = obj.getParentObj()
        if p is None:
            return Qt.QString.fromLatin1(self.getNoneValue())
        return Qt.QString.fromLatin1("<HTML>" + self._configParam + ' for ' + p.getDisplayName())
    
    def isReadOnly(self):
        return True
    
    def setStyleSheet(self, ss):
        ss = str(ss)
        self._extra_style = ss
        if not self.getModelValueObj():
            ss = """TauConfigLabel { 
                background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                              stop: 0.00 rgb(240,240,240),
                              stop: 0.33 rgb(200,200,200),
                              stop: 0.66 rgb(240,240,240),
                              stop: 1.00 rgb(200,200,200));
                color: gray;
            %s}""" % (ss)
        else:
            ss = """TauConfigLabel { %s }""" % ss
        QtGui.QLabel.setStyleSheet(self, ss)
        
    def updateStyle(self):
        self.setStyleSheet(self._extra_style)
        self.updatePendingOpsStyle()
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    @QtCore.pyqtSignature("setModel(QString)")
    def setModel(self, model):
        model = str(model)
        try:
            self._configParam = model[model.rfind('=')+1:].lower()
        except:
            self._configParam = ''
        taubase.TauBaseWidget.setModel(self,model)
    
    def getPrefixText(self):
        return self._prefix
    
    @QtCore.pyqtSignature("setPrefixText(QString)")
    def setPrefixText(self,prefix):
        self._prefix = prefix
        self.fireEvent(evt_type = tau.core.TauEventType.Change)

    def resetPrefixText(self):
        self.setPrefixText('')
        
    def getSuffixText(self):
        return self._suffix
    
    @QtCore.pyqtSignature("setSuffixText(QString)")
    def setSuffixText(self,suffix):
        self._suffix = suffix
        self.fireEvent(evt_type = tau.core.TauEventType.Change)
    
    def resetSuffixText(self):
        self.setSuffixText('')
        
    model = QtCore.pyqtProperty("QString", taubase.TauBaseWidget.getModel, 
                                setModel, taubase.TauBaseWidget.resetModel)
                                
    useParentModel = QtCore.pyqtProperty("bool",
                                         taubase.TauBaseWidget.getUseParentModel, 
                                         taubase.TauBaseWidget.setUseParentModel,
                                         taubase.TauBaseWidget.resetUseParentModel)
                                         
    prefixText = QtCore.pyqtProperty("QString", getPrefixText, setPrefixText, 
                                     resetPrefixText, doc="prefix text (optional)")
                                     
    suffixText = QtCore.pyqtProperty("QString", getSuffixText, setSuffixText, 
                                     resetSuffixText, doc="suffix text (optional)")

class TauStateLabel(TauValueLabel):
    
    def __init__(self, parent = None, designMode = False):
        self.call__init__(TauValueLabel, parent=parent, designMode=designMode, background = 'value_state')
        self.setAlignment(QtCore.Qt.AlignCenter)
    
    def getFormatedToolTip(self, cache=True):
        """ The tooltip should refer to the device and not the state attribute.
            That is why this method is being rewritten
        """
        if self.modelObj is None:
            return Qt.QString.fromLatin1(self.getNoneValue())
        parent = self.modelObj.getParentObj()
        if parent is None:
            return Qt.QString.fromLatin1(self.getNoneValue())
        return Qt.QString.fromLatin1(self.toolTipObjToStr( parent.getDisplayDescrObj() ))
    
    @QtCore.pyqtSignature("setModel(QString)")
    def setModel(self,model):
        """ Added to allow setModel(device_name); assuming that State attribute
            is the desired model. To inherit the parent model but with a non
            'State' named attribute, use slash: '/TheAttribute'
        """
        try:
            if str(model).count('/')<3 and str(model)[0]!='/':
                model = str(model)+'/State'
        except Exception:
            pass
        taubase.TauBaseWidget.setModel(self,model)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = QtCore.pyqtProperty("QString", taubase.TauBaseWidget.getModel, 
                                setModel, 
                                taubase.TauBaseWidget.resetModel)

def main():
    import sys
    app = QtGui.QApplication([])
    tau.setLogLevel(tau.Trace)
    
    if len(sys.argv) == 2:  name = sys.argv[1]
    else: name = 'motor/bl99_simumotctrl1/1/state'
    
    if name.lower().endswith('state'):
        w = TauStateLabel()
    elif name.lower().count('?configuration'):
        w = TauConfigLabel()
    else:
        w = TauValueLabel()
    
    w.setModel(name)
    w.show()
    app.exec_()

if __name__ == '__main__':
    main()