
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
Led.py:  A led widget for PyQT4
"""

import sys
from PyQt4 import Qt

import tau.core.utils as utils
from tau.widget.resources import qrc_extra_icons_leds_images24
from tau.widget.resources import qrc_extra_icons_leds_images48

LedColor  = utils.Enumeration("LedColor",  ["BLUE", "GREEN", "RED", "YELLOW", "ORANGE"])
LedStatus = utils.Enumeration("LedStatus", ["ON", "OFF"])
LedSize   = utils.Enumeration("LedSize",   [("SMALL", 24), ("LARGE", 48)])


class Led(Qt.QLabel):
    ledDirPattern = ":leds/images%(size)d"

    def __init__(self, parent = None, ledsize = LedSize.SMALL, ledcolor = LedColor.GREEN):

        Qt.QLabel.__init__(self, parent)
        
        self.ledsize = ledsize
        self.status = LedStatus.OFF
        self.ledcolor = ledcolor
        
        self.setObjectName("Led")
        sizePolicy = Qt.QSizePolicy(Qt.QSizePolicy.Policy(0),Qt.QSizePolicy.Policy(0))
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)
        
        # no need to call changeColor. changeSize calls it for us
        self.changeSize(ledsize)
        
        self.retranslateUi(self)
        Qt.QMetaObject.connectSlotsByName(self)

    def toLedName(self,status,color):
        name = "led" + LedColor.whatis(color).lower()
        if status == LedStatus.OFF:
            name += LedStatus.whatis(status).lower()
        return name
    
    def toCompleteLedName(self,size,status,color):
        ret = Led.ledDirPattern % { "size" : size} 
        ret += "/" + self.toLedName(status,color) + ".png"
        return ret

    def tr(self, string):
        return Qt.QApplication.translate("Led", string, None, Qt.QApplication.UnicodeUTF8)

    def retranslateUi(self, Led):
        Led.setWindowTitle(self.tr("Form"))
    
    def on(self):
        if self.status == LedStatus.OFF:
            self.status = LedStatus.ON
            self.setPixmap(self.onled)
    
    def off(self):
        if self.status == LedStatus.ON:
            self.status = LedStatus.OFF
            self.setPixmap(self.offled)  
    
    def changeSize(self, ledSize):
        self.ledsize = ledSize
        self.resize(ledSize,ledSize)
        self.setMinimumSize(Qt.QSize(ledSize,ledSize))
        self.setMaximumSize(Qt.QSize(ledSize,ledSize))
        
        # force a call to led color to change the place from which the images are fetched
        self.changeColor(self.ledcolor)
        
    def changeColor(self, ledColor):
        self.ledcolor = ledColor
        
        off_pixmap_name = self.toCompleteLedName(self.ledsize, LedStatus.OFF, ledColor)
        self.offled = Qt.QPixmap(off_pixmap_name)
        on_pixmap_name = self.toCompleteLedName(self.ledsize, LedStatus.ON, ledColor)
        self.onled = Qt.QPixmap(on_pixmap_name)
        
        if self.status == LedStatus.OFF:
            self.setPixmap(self.offled)
        else:
            self.setPixmap(self.onled)
    
    def _setProblemsBackground(self, doIt=True):
        if doIt:
            self.setAutoFillBackground(True)
            bg_brush = Qt.QBrush()
            bg_brush.setStyle(Qt.Qt.BDiagPattern)
            palette = self.palette()
            palette.setBrush(Qt.QPalette.Window, bg_brush)
            palette.setBrush(Qt.QPalette.Base, bg_brush)
        else:
            self.setAutoFillBackground(False)

def main():
    app = Qt.QApplication(sys.argv)
    
    w = Qt.QWidget()
    l = Qt.QGridLayout()
    w.setLayout(l)
    
    for i in xrange(5):
        led = Led()
        led.off()
        led.changeColor(i)
        l.addWidget(led, 0, i)

    for i in xrange(5):
        led = Led()
        led.on()
        led.changeColor(i)
        l.addWidget(led, 1, i)

    for i in xrange(5):
        led = Led()
        led.off()
        led.changeColor(i)
        led.changeSize(LedSize.LARGE)
        l.addWidget(led, 2, i)

    for i in xrange(5):
        led = Led()
        led.on()
        led.changeColor(i)
        led.changeSize(LedSize.LARGE)
        l.addWidget(led, 3, i)
        
    w.setVisible(True)
    app.exec_()
    
if __name__ == "__main__":
    main()