#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
Simple macro execution widgets
"""

from PyQt4 import Qt

import tau.core
import tau.widget

class MacroListModel(Qt.QAbstractListModel):
    
    def __init__(self, ms):
        self._ms = ms
        Qt.QAbstractListModel.__init__(self)
        
    def data(self, index, role=Qt.Qt.DisplayRole):
        if index.isValid():
            macro_name = self._ms.getMacroStrList()[index.row()]
            return Qt.QVariant(macro_name)
        return Qt.QVariant()
    
    def rowCount(self, parent=Qt.QModelIndex()):
        s = len(self._ms.getMacroStrList())
        return s


class MacroPanel(Qt.QWidget, tau.widget.TauBaseWidget):
    
    def __init__(self, qt_parent=None, designMode=False, macro_server=None):
        Qt.QWidget.__init__(self, qt_parent)
        tau.widget.TauBaseWidget.__init__(self, self.__class__.__name__, designMode=designMode)
        self._runningMacro = None
        layout = Qt.QGridLayout()
        self.setLayout(layout)
        self._macrolineEdit = Qt.QLineEdit(self)
        self._runButton = Qt.QPushButton("Run", self)
        self._progressBar = Qt.QProgressBar(self)
        layout.addWidget(self._macrolineEdit, 0, 0)
        layout.addWidget(self._runButton, 0, 1)
        layout.addWidget(self._progressBar, 0, 2)
        self.connect(self._runButton , Qt.SIGNAL("clicked()"), self._runMacro)
    
    def getModelClass(self):
        return tau.core.TauDevice
    
    def postAttach(self):
        tau.widget.TauBaseWidget.postAttach(self)
        door = self.getModelObj()
        if door is None: return
        door.connect(door, Qt.SIGNAL("macroStatusUpdated"), self._macroStatusUpdated)
        self._updateCompleter(door)
    
    def preDetach(self):
        tau.widget.TauBaseWidget.preDetach(self)
        door = self.getModelObj()
        if door is None: return
        door.disconnect(door, Qt.SIGNAL("macroStatusUpdated"), self._macroStatusUpdated)
    
    def _updateCompleter(self, door):
        completer = None
        if door is not None:
            self._macro_list_model = MacroListModel(door.macro_server)
            completer = Qt.QCompleter(self._macro_list_model, self._macrolineEdit)
            completer.setCaseSensitivity(Qt.Qt.CaseInsensitive)
            completer.setCompletionMode(Qt.QCompleter.InlineCompletion)
        self._macrolineEdit.setCompleter(completer)
        self._progressBar.reset()
    
    def _runMacro(self):
        macro_parameters = str(self._macrolineEdit.text()).split()
        door = self.getModelObj()
        if door is not None:
            macro_name = macro_parameters[0]
            macro_parameters = macro_parameters[1:]
            self._progressBar.reset()
            self._progressBar.setFormat("%s (%%p%%)" % macro_name)
            door.runMacro(macro_name, macro_parameters)
    
    def _macroStatusUpdated(self, macro):
        pb = self._progressBar
        if macro is None:
            pb.reset()
            return
        pb.setRange(*macro.range)
        pb.setValue(macro.step)
    
def main():
    import tau.widget.tango.macroserver
    import sys
    
    tau.widget.tango.macroserver.registerExtensions()
    
    app = Qt.QApplication([])
    panel = MacroPanel()
    panel.setModel("door/bl99/1")
    panel.show()
    sys.exit(app.exec_())
    
if __name__ == "__main__": main()