
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
mainwindow.py: a main window implementation with many added features by default
"""

from PyQt4 import Qt

from containers import TauBaseContainer
#from tauLogo import TauLogo

from logger import TauLoggerWidget

from tau.widget.resources import qrc_tango_icons_places

import cPickle as pickle

class TauMainWindow(Qt.QMainWindow, TauBaseContainer):
    '''A Tau-aware QMainWindow with several customizations:
    
        - It takes care of (re)storing its geometry and state
        - It provides a statusBar (@TODO)
        - It provides basic Tau menus and actions:
            The following Menus are already Provided:
                -Help (self.helpMenu)
                    - About
                -View (self.viewMenu)
                
                -Tau (self.tauMenu)
            The following actions are already provided:
                -Help-->About
        - It incorporates a TauLogo (@TODO)
    '''
    __pyqtSignals__ = ("modelChanged(const QString &)",)
        
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QMainWindow, parent)
        self.call__init__(TauBaseContainer, name, designMode=designMode)
        
        self.__tangoHost = ""
        
        self.__settings = None
        
        #create a few common Application-wide actions
        #Qt.QTimer.singleShot(0, self.__createActions())
        self.__createActions()
        
        
#        self.tauLogo = TauLogo()

        #logger dock widget
        self.__loggerDW = Qt.QDockWidget("Tau logs", self)
        self.__loggerDW.setWidget(TauLoggerWidget())
        self.__loggerDW.setObjectName("loggerDW")
        self.addDockWidget(Qt.Qt.BottomDockWidgetArea, self.__loggerDW)

        
        #Create Menus
        #View menu
        self.viewMenu = self.menuBar().addMenu("View")
        self.viewMenu.addAction(self.__loggerDW.toggleViewAction())
        
        #Tau Menu
        self.tauMenu = self.menuBar().addMenu("Tau")
        self.tauMenu.addAction(self.changeTangoHostAction)
        
        #Help Menu
        self.helpMenu = self.menuBar().addMenu("Help")
        self.helpMenu.addAction(self.helpAboutAction)
        
        
    def basicTauToolbar(self):
        '''returns a QToolBar with few basic buttons (most important, the logo)
        
        :return: (QToolBar)
        '''
        tb = Qt.QToolBar('Tau Toolbar')
#        tb.addAction(self.changeTangoHostAction)
#        tb.addWidget(self.tauLogo)
        tb.addAction(Qt.QIcon(":/logo.png"),Qt.qApp.organizationName())
        tb.setIconSize(Qt.QSize(50,50))
        return tb

    def __createActions(self):
        '''initializes the application-wide actions'''
        self.changeTangoHostAction =  Qt.QAction(Qt.QIcon(":/places/network-server.svg"),'Change Tango Host...', self)
        self.changeTangoHostAction.setShortcut(Qt.QKeySequence("Ctrl+P"))
        self.connect(self.changeTangoHostAction, Qt.SIGNAL("triggered()"), self._onChangeTangoHostAction)
        
        self.helpAboutAction = Qt.QAction('About...', self)
        #self.helpUserManual = Qt.QAction('')
    
    
    def setQSettings(self, settings):
        '''sets the main window settings object
        
        :param settings: (QSettings or None)
        
        .. seealso:: :meth:`getQSettings`
        '''
        self.__settings = settings
        
    def resetQSettings(self):
        '''equivalent to setQSettings(None) '''
        self.setQSettings(None)
        
    def getQSettings(self):
        '''Returns the main window settings object.
        If the settings were set to None (which is the default unless the
        settings are explicitly set with :meth:`setQSettings`), the returned
        QSettings object will comply with the Tau defaults for storing
        application settings (i.e., it uses Ini format and userSpace)
        
        :return: (QSettings) a tau-specific QSettings object
        '''
        if self.__settings is None:
            #using the ALBA-Controls Coding Convention on how to store application settings
            format=Qt.QSettings.IniFormat
            scope = Qt.QSettings.UserScope
            appname = Qt.QApplication.applicationName()
            orgname = Qt.QApplication.organizationName()
            self.__settings = Qt.QSettings(format, scope, orgname, appname)
            ##Note: for Qt v>=4.4, the five previous lines could have been substituted by the following two: 
            #self.__settings = Qt.QSettings() #this uses QCoreApplication.applicationName() and QCoreApplication.organizationName()
            #self.__settings.setDefaultFormat(Qt.QSettings.IniFormat)
        return self.__settings
        
    def loadSettings(self):
        '''restores the application settings previously saved with saveSettings.
        
        This method should be called explicitly from derived classes after all
        initialization is done
        '''
        settings = self.getQSettings()
        self.restoreState(settings.value("MainWindow/State").toByteArray())
        self.restoreGeometry(settings.value("MainWindow/Geometry").toByteArray())
        #self.setModel(settings.value("MainWindow/model").toString())
        #self.setTangoHost(settings.value('TangoHost', Qt.QVariant('')).toString())
        
        #restore the Tau config
        self.applyQConfig(settings.value('TauConfig').toByteArray())
        self.info('MainWindow settings restored')
        
    
    def saveSettings(self):
        '''saves the application settings (so that they can be restored with loadSettings)
        '''
        #main window geometry
        settings = self.getQSettings()
        settings.setValue("MainWindow/State",Qt.QVariant(self.saveState()))
        settings.setValue("MainWindow/Geometry", Qt.QVariant(self.saveGeometry()))
        #settings.setValue("MainWindow/model",Qt.QVariant(Qt.QString(self.getModelName())))
        #settings.setValue('TangoHost',Qt.QVariant(Qt.QString(self.getTangoHost())))
        
        #store the config dict
        settings.setValue("TauConfig", Qt.QVariant(self.createQConfig()))
        
        self.info('MainWindow settings saved in "%s"'%settings.fileName())
    
    def closeEvent(self,event):
        '''This event handler receives widget close events'''
        self.saveSettings() #save current window state before closing
        
    def _onChangeTangoHostAction(self):
        '''
        slot called when the Change Tango Host is triggered. It prompts for a
        Tango host name and calls :meth:`setTangoHost`
        '''
        host, valid = Qt.QInputDialog.getText  ( self, 'Change Tango Host', 'New Tango Host', Qt.QLineEdit.Normal,  str(self.getTangoHost()))
        if valid:
            self.setTangoHost(str(host))
        
    def setTangoHost(self, host):
        self.__tangoHost = host
        
    def getTangoHost(self):
        return self.__tangoHost
    
    def resetTangoHost(self):
        self.setTangoHost(None)
        
        
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Public slots for apply/restore changes
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSignature("applyPendingChanges()")
    def applyPendingChanges(self):
        self.applyPendingOperations()
    
    @Qt.pyqtSignature("resetPendingChanges()")
    def resetPendingChanges(self):
        self.resetPendingOperations()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QString", TauBaseContainer.getModel, 
                                TauBaseContainer.setModel, 
                                TauBaseContainer.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool", 
                                         TauBaseContainer.getUseParentModel, 
                                         TauBaseContainer.setUseParentModel,
                                         TauBaseContainer.resetUseParentModel)
    
    showQuality = Qt.pyqtProperty("bool", TauBaseContainer.getShowQuality, 
                                      TauBaseContainer.setShowQuality, 
                                      TauBaseContainer.resetShowQuality)
    
    tangoHost = Qt.pyqtProperty("QString", getTangoHost, 
                                        setTangoHost, 
                                        resetTangoHost)

#---------

if __name__ == "__main__":
    
    import sys
    app = Qt.QApplication(sys.argv)
    app.setApplicationName('TauMainWindow-test')
    app.setOrganizationName('ALBA')
    
    form = TauMainWindow()
    form.loadSettings()
    
    form.setCentralWidget(Qt.QMdiArea()) #just for testing
    
    
    form.show()
    
    sys.exit(app.exec_())
    