
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
qub.py: tau wrapping of Qub image visualization widget
"""
import weakref

from PyQt4 import Qt
from Qub.Widget.DataDisplay.QubDataImageDisplay import QubDataImageDisplay
from Qub.Widget.QubActionSet import QubToolbarToggleButtonAction
from Qub.Widget.QubActionSet import QubOpenDialogAction
import tau.widget

try:
    import EdfFile
except ImportError:
    EdfFile = None

class _EDFFileSavePlug:
    def setData(self,data) :
        if data is not None:
            self._data = weakref.ref(data)
        else:
            self._data = None
            
class TauQubDataImageDisplay(QubDataImageDisplay, tau.widget.TauBaseWidget):
    
    def __init__(self, parent=None, **kwargs):
        designMode = kwargs.get('designMode', False)
        kwargs['forcePopupSubWindow'] = True
        kwargs['parent'] = parent
        self.call__init__(QubDataImageDisplay, **kwargs)
        name = self.__class__.__name__
        self._image_attr_name = ''
        self.call__init__(tau.widget.TauBaseWidget, name, designMode=designMode)
        self.statusBar().setSizeGripEnabled(False)
        
        self.__addActions()
        
    def __addActions(self):
        updateAction = QubToolbarToggleButtonAction(name = 'update', iconName = 'update',
                                                    initState = True, group='image',
                                                    tooltip='Start/Stop image refresh')
        Qt.QObject.connect(updateAction, Qt.SIGNAL('StateChanged'), self.__setRefresh)
        self.addAction(updateAction)
        # a bug in Qub forces to set the state after adding the action
        updateAction.setState(True)
        self._updateAction = updateAction
        
        if EdfFile:
            saveAction = QubOpenDialogAction(name='save edf', iconName='save',
                                             label='Save data', group='admin',
                                             tooltip='Save image data in EDF format')
            
            saveDialog = Qt.QFileDialog(self, 'Save image as EDF')
            saveDialog.setAcceptMode(Qt.QFileDialog.AcceptSave)
            saveDialog.setDefaultSuffix('edf')
            nameFilters = list(saveDialog.nameFilters())
            nameFilters.insert(0, 'EDF Files (*.edf)')
            saveDialog.setNameFilters(nameFilters)
            Qt.QObject.connect(saveDialog, Qt.SIGNAL('fileSelected(const QString &)'), 
                               self.__saveEDF)
            saveAction.setDialog(saveDialog)
            self._saveDataAction = saveAction
            
            plug = _EDFFileSavePlug()
            self.addDataAction(saveAction, plug)
            self._saveEDFPlug = plug
        
    def __saveEDF(self, fileSelected):
        if not fileSelected: return
        fileSelected = str(fileSelected)
        data = self._saveEDFPlug._data()
        edf_file = EdfFile.EdfFile(fileSelected)
        edf_file.WriteImage({}, data)

    def __setRefresh(self, onOff):
        self._refreshEnabled = onOff
    
    def _save_dialog_new(self,openDialogAction,aQubGraphicsView):
        QubDataImageDisplay._save_dialog_new(self,openDialogAction,aQubGraphicsView)
    
    def isReadOnly(self):
        return True
    
    def getModelClass(self):
        return tau.core.TauDevice
    
    def handleEvent(self, evt_src, evt_type, evt_value):
        if self._updateAction.state():
            if evt_type in (tau.core.TauEventType.Change, tau.core.TauEventType.Periodic):
                data = self.getModelObj().getImageData()
                if data:
                    data = data[self._image_attr_name][1]
                    self.setData(data.value)
                    self.setInfo({
                        'name'     : data.name,
                        'width'    : data.dim_x,
                        'height'   : data.dim_y,
                        'quality'  : data.quality,
                        'type'     : data.type,
                        'timestamp': data.time
                    })

    def sizeHint(self):
        return Qt.QSize(640, 480)
    
    def setModel(self, model):
        dev_name, self._image_attr_name = model.rsplit('/', 1)
        obj = self.getTauFactory().getDevice(dev_name)
        if obj is not None:
            obj.addImageAttrName(self._image_attr_name)        
        tau.widget.TauBaseWidget.setModel(self, dev_name)
        self.setWindowTitle(self.getModel())
    
    def getModel(self):
        return '%s/%s' % (tau.widget.TauBaseWidget.getModel(self), self._image_attr_name)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", tau.widget.TauBaseWidget.getModel,
                            setModel,
                            tau.widget.TauBaseWidget.resetModel)
    
    useParentModel = Qt.pyqtProperty("bool",
                                     tau.widget.TauBaseWidget.getUseParentModel,
                                     tau.widget.TauBaseWidget.setUseParentModel,
                                     tau.widget.TauBaseWidget.resetUseParentModel)

def main():
    import sys
    import tau.core.tango.img
    tau.core.tango.img.registerExtensions()

    app = Qt.QApplication(sys.argv)
    models=sys.argv[1:]
    panel = Qt.QWidget()
    l = Qt.QHBoxLayout()
    panel.setLayout(l)
    if not models:
        print "must give at least one image device name"
        return
    for model in models:
        w = TauQubDataImageDisplay()
        w.setModel(model)
        l.addWidget(w)
  
    panel.setVisible(True)
    sys.exit(app.exec_())
    
if __name__ == "__main__":
    main()
