
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
tautrend.py: Generic trend widget for Tau
"""


import os, copy
from datetime import datetime
import numpy
from PyQt4 import Qt, Qwt5

import tau.core
from tau.core.utils import CaselessDict, CaselessList
from tau.widget import TauBaseComponent
from tau.widget.qwt import TauPlot, TauCurveSignalProvider
import PyTango

try:
    from PyTangoArchiving import getArchivedTrendValues
except:
    #print 'ARCHIVING NOT AVAILABLE!'
    getArchivedTrendValues = (lambda trend,model,start,stop: [])


class TauTrendsSet(Qt.QObject, TauBaseComponent):
    """A collection of TauCurves generated from a Tau Attribute.
    
    If the attribute is a scalar, The Trend Set consists of only one curve
    representing the evolution of the value of the attribute. If the attribute
    is a SPECTRUM, as many curves as the length of the spectrum are created,
    each representing the evolution of the value of a component of the array.
       
    When an event is received, all curves belonging to a TauTrendSet are updated.
    
    TauTrendSet objects can be considered as containers of (sorted) curves. As
    such, the curves contained by them can be accessed with item notation, as in
    the following example::
    
        attrname = 'a/b/c/d'       #consider this attribute is a SPECTRUM of 3 elements
        ts=TauTrendSet(attrname)
        ...                        # wait for a Tau Event arriving so that the curves are created
        n_curves = len(ts)         #ncurves will be 3 (assuming the event already arrived)
        curve0 = ts[0]             #you can access the curve by index
        curve1 = ts['a/b/c/d[1]']  #and also by name

    
    Note that internally each curve is treated as a RawData curve (i.e., it is
    not aware of events by itself, but it relies on the TauTrendSet object to
    update its values)
    
    """
    def __init__(self, name, parent = None, curves=None):
        self._parent = parent
        self._signalGen = TauCurveSignalProvider(parent)
        self.call__init__(TauBaseComponent, self.__class__.__name__)
        self._history = []
        self.__xBuffer = None
        self.__yBuffer = None
        try: self.__maxBufferSize = self._parent.getMaxDataBufferSize()
        except: self.__maxBufferSize = 1048576 #(1M= 2**20)
        if curves is None:
            self._curves = {}
            self._orderedCurveNames = []
        else:
            self._curves = curves
            self._orderedCurveNames = curves.keys()
        self.setModel(name)
        
    def __getitem__(self, key):
        if isinstance(key, int):
            key = self._orderedCurveNames[key]
        return self._curves[key]
        
    def __len__(self):
        return len(self._orderedCurveNames)
        
    def __contains__(self, k):
        return self._curves.__contains__(k)
        
    def index(self, curveName):
        '''Returns the index in the trend for the given curve name. It gives an
        exception if the curve is not in the set.
        
        :param curveName: (str) the curvename to find
        
        :return: (int) The index associated to the given curve in the TrendSet
        '''
        return self.getCurveNames().index(curveName)
    
    def setTitleText(self, basetitle):
        '''Sets the title text of the trends this trendset. The name will be
        constructed by appending "[%i]" to the basetitle, where %i is the index
        position of the trend in the trendset. As a particular case, nothing is
        appended if the trendset consists of only one trend.
        
        :param basetitle: (str) The title text to use as a base for constructing
                          the title of each trend belonging to this trendset. It
                          may contain placeholders as those used in
                          :meth:`TauCurve.setTitleText`
        
        .. seealso:: :meth:`TauCurve.setTitleText`
        '''
        self._titleText = basetitle
        titles = self.compileTitles(basetitle)
        for t,(n,c) in zip(titles, self.getCurves()):
            c.setTitleText(t)
    
    def compileTitles(self, basetitle):
        '''Return a list of titles. Each title corresponds to a trend of the
        trendset (ordered). Substitution of known placeholders is performed.
        
        :param basetitle: (str) A string to be used as base title. It may
                          contain any of the following placeholders (which will
                          be substituted by their corresponding value):
            
            - <label> the attribute label (default)
            - <model> the model name
            - <attr_name> attribute name
            - <attr_full_name> full attribute name
            - <dev_alias> device alias
            - <dev_name> device name
            - <dev_full_name> full device name
            - <current_title> The current title
            - <trend_index> The index of the trend in the trendset
            - <[trend_index]> Same as: `"[<trend_index>]" if Ntrends>1 else ""`
        
        :return: (string_list) a list of title strings that correspond to the
                 list of trends in the set. 
        
        .. seealso:: :meth:`compileBaseTile`
        '''
        basetitle = self.compileBaseTile(basetitle)
        ntrends = len(self._curves)
        if '<trend_index>' in basetitle:
            ret = [basetitle.replace('<trend_index>', "%i"%i) for i in xrange(ntrends)]
        else:
            ret = [basetitle]*ntrends
        return ret
    
    def compileBaseTile(self, basetitle):
        '''Return a base tile for a trend in whichs substitution of known
        placeholders has been performed.
        
        :param basetitle: (str) String on which the substitutions will be
                          performed. The following placeholders are supported:
            
            - <label> the attribute label (default)
            - <model> the model name
            - <attr_name> attribute name
            - <attr_full_name> full attribute name
            - <dev_alias> device alias
            - <dev_name> device name
            - <dev_full_name> full device name
            - <current_title> The current title
            - <[trend_index]> Same as: `"[<trend_index>]" if Ntrends>1 else ""`
        
        **Note** that <trend_index> itself is not substituted!
        
        :return: (str) the compiled base title.
        
        .. seealso:: :meth:`compileTitles`
        '''
        attr = self.getModelObj()
        basetitle = basetitle.replace('<current_title>',self._titleText)
        basetitle = basetitle.replace('<model>',self.getModel())
        if isinstance(attr, tau.core.TauAttribute):
            basetitle = basetitle.replace('<label>',attr.label or '---')
            basetitle = basetitle.replace('<attr_name>',attr.name or '---')
            basetitle = basetitle.replace('<attr_full_name>',attr.getFullName() or '---')
            basetitle = basetitle.replace('<dev_alias>',attr.dev_alias or '---')
            basetitle = basetitle.replace('<dev_name>',attr.dev_name or '---')
        
        dev = attr.getParentObj()
        if dev is not None:
            basetitle = basetitle.replace('<dev_full_name>',dev.getFullName() or '---')

        if len(self._curves)==1: basetitle = basetitle.replace('<[trend_index]>','')
        else: basetitle = basetitle.replace('<[trend_index]>','[<trend_index>]')    
            
        return basetitle
        
    def addCurve(self, name, curve):
        '''add a curve (with the given name) to the internal curves dictionary of this TauTrendSet
        
        :param name: (str) the name of the curve
        :param curve: (TauCurve) the curve object to be added
        '''  
        self._curves[name] = curve
        self._orderedCurveNames.append(name)
        
    def getCurves(self):
        '''returns an iterator of (curveName,curveObject) tuples associated to
        this TauTrendSet. The curves will always be returned in the order they
        were added to the set
        
        :return: (iterator<str,TauCurve>)
        '''
        return iter([(n,self._curves[n]) for n in self._orderedCurveNames])
    
    def getCurveNames(self):
        '''returns a list of the names of the curves associated to this
        TauTrendSet. The curve names will always be returned in the order they
        were added to the set
        
        :return: (list<str>) the names of the curves
        '''
        return self._orderedCurveNames
        
    def getModelClass(self):
        '''see :meth:`TauBaseComponent.getModelClass`'''
        return tau.core.TauAttribute

    def registerDataChanged(self,listener,meth):
        '''see :meth:`TauBaseComponent.registerDataChanged`'''
        listener.connect(self._signalGen, self._signalGen.newDataChangedSignal(), meth)
        
    def unregisterDataChanged(self,listener,meth):
        '''see :meth:`TauBaseComponent.unregisterDataChanged`'''
        listener.disconnect(self._signalGen, self._signalGen.newDataChangedSignal(), meth)
    
    def preAttach(self):
        """reimplemented because this cannot use signals(they are not QObjects).
        see :meth:`TauBaseComponent.preAttach`"""
        Qt.QObject.connect(self._signalGen, Qt.SIGNAL('tauEvent'), self.filterEvent)

    def preDetach(self):
        """reimplemented because this cannot use signals(they are not QObjects)
        see :meth:`TauBaseComponent.preDetach`"""
        Qt.QObject.disconnect(self._signalGen, Qt.SIGNAL('tauEvent'), self.filterEvent)
        
    def fireEvent(self,  evt_src, type, evt_value):
        """reimplemented because this cannot use signals(they are not QObjects).
        see :meth:`TauBaseComponent.fireEvent`"""
        self._signalGen.emit(Qt.SIGNAL('tauEvent'),  evt_src, type, evt_value)
    
    def _updateHistory(self, value):
        '''Update the history data buffers using the latest value from the event
        
        :param value: (PyTango.DeviceAttribute) the value from the event
        
        :return: (tuple<numpy.ndarray, numpy.ndarray>) Tuple of two arrays
                 containing the X data and Y data, respectively, from the
                 history buffers:
                 - The dtype of both arrays is "double".
                 - The X array will contain timestamps if the parent is in
                   XisTime mode, and a sequential event number otherwise.
                 - The Y array may be two-dimensional if the attribute value is
                   a spectrum
                   
        **Example of return values:** Consider history of 10 events for a
        SPECTRUM attribute with dim_x=8. Then the return value will be (X,Y)
        where X.shape=(10,) and Y.shape=(10,8); X.dtype = Y.dtype = <dtype('float64')>
        '''
        
        if self.__xBuffer is None:
            self.__xBuffer = tau.core.utils.ArrayBuffer(numpy.zeros(128, dtype='d'), maxSize=self.__maxBufferSize )
        if self.__yBuffer is None:
            self.__yBuffer = tau.core.utils.ArrayBuffer(numpy.zeros((128, value.dim_x),dtype='d'), maxSize=self.__maxBufferSize )
        
        self.__yBuffer.append(value.value)
        
        if self._parent.xIsTime:
            #add the timestamp to the x buffer
            self.__xBuffer.append(value.time.totime())
            ##Adding archiving values
            if self._parent.getUseArchiving():
                if self._parent.getXDynScale() or not self._parent.axisAutoScale(Qwt5.QwtPlot.xBottom): #Do not open a mysql connection for autoscaled plots
                    startdate = self._parent.axisScaleDiv(Qwt5.QwtPlot.xBottom).lowerBound()
                    stopdate = self.__xBuffer[0]
                    try:
                        archived = getArchivedTrendValues(self,model,startdate,stopdate)
                        del(archived[:-self.__xBuffer.remainingSize()]) #limit the archived values according to the maximum size of the buffer
                        t = numpy.zeros(len(archived), dtype=float)
                        y = numpy.zeros((len(archived), value.dim_x), dtype=self.__yBuffer.dtype)
                        for i,v in enumerate(archived):
                            t[i]=v.time.totime()
                            y[i]=v.value
                        self.__xBuffer.extendLeft(t)
                        self.__yBuffer.extendLeft(y)
                    except Exception,e:
                        self.trace('%s: reading from archiving failed: %s'%(datetime.now().isoformat('_'),str(e)))     
        else:
            #add the event number to the x buffer
            try:
                self.__xBuffer.append(1.+self.__xBuffer[-1]) 
            except IndexError: #this will happen when the x buffer is empty
                self.__xBuffer.append(0) 
        
        return self.__xBuffer.contents(), self.__yBuffer.contents()

    
    def handleEvent(self, evt_src, evt_type, evt_value):
        ''' processes Change (and Periodic) Tau Events: updates the data of all
        curves in the set according to the value of the attribute.
        
        For documentation about the parameters of this method, see
        :meth:`TauBaseComponent.handleEvent`'''
        
        if evt_type == tau.core.TauEventType.Config:
            self.setTitleText(self._parent.getDefaultCurvesTitle())
            return
        
        if evt_type == tau.core.TauEventType.Error:
            return
        
        model = evt_src if evt_src is not None else self.getModelObj()
        if model is None: return
        
        value = evt_value if isinstance(evt_value, PyTango.DeviceAttribute) else self.getModelValueObj()
        if value is None or value.value is None: return
        
        #Check that the data dimensions are consistent with what was plotted before
        ntrends = value.dim_x
        if ntrends != len(self._curves):
            #clean previous curves
            for subname in self.getCurveNames():
                self._parent.detachRawData(subname)
            self._curves = {}
            self._orderedCurveNames = []
            #clean history Buffers
            self.__xBuffer = None
            self.__yBuffer = None
            #create as many curves as the dim_x of the given model and add them to the TrendSet
            name = self.getModelName()
            rawdata = {'x':numpy.zeros(0), 'y':numpy.zeros(0)}
            for i in xrange(ntrends):
                subname = "%s[%i]"%(name,i)
                self._parent.attachRawData(rawdata,id=subname)
                self.addCurve(subname, self._parent.curves[subname])
            self.setTitleText(self._parent.getDefaultCurvesTitle())
            self._parent.autoShowYAxes()

        #get the data from the event
        self._xValues, self._yValues = self._updateHistory(value)

        #assign xvalues and yvalues to each of the curves in self._curves
        for i,(n,c) in enumerate(self.getCurves()):
            c._xValues, c._yValues = self._xValues, self._yValues[:,i]
            
            #update min/max markers
            if c.isVisible():
                if c._showMaxPeak:
                    try: maxpoint = [c._xValues[c._yValues.argmax()],c._yValues.max()]
                    except: maxpoint = [0, 0]
                    c._maxPeakMarker.setValue(*maxpoint)
                    label = c._maxPeakMarker.label()
                    label.setText("Max. " + n + " " + repr(maxpoint[1]) + ' at x = ' + repr(maxpoint[0]))
                    c._maxPeakMarker.setLabel(label)
                if c._showMinPeak:
                    try: minpoint = [c._xValues[c._yValues.argmin()],c._yValues.min()]
                    except: minpoint = [0,0]
                    c._minPeakMarker.setValue(*minpoint)
                    label = c._minPeakMarker.label()
                    label.setText("Max. " + n + " " + repr(minpoint[1]) + ' at x = ' + repr(minpoint[0]))
                    c._minPeakMarker.setLabel(label)

        self._signalGen.generateDataChanged(Qt.QString(self.getModel()),evt_type)
        
    def isReadOnly(self):
        return True
    
    def setMaxDataBufferSize(self, maxSize):
        '''sets the maximum number of events that are stored in the internal
        buffers of the trend. Note that this sets the maximum amount of memory
        used by the data in this trend set to:
        
            ~(1+ntrends)*2*8*maxSize bytes
        
        (the data is stored as float64, and two copies of it are kept: one at
        the x and y buffers and another at the QwtPlotCurve.data)
        
        :param maxSize: (int) the maximum limit
        '''
        self.__xBuffer.setMaxSize(maxSize)
        self.__yBuffer.setMaxSize(maxSize)
        self.__maxBufferSize = maxSize
           
class TauTrend(TauPlot):
    '''
    A :class:`TauPlot` -derived widget specialised in plotting trends (i.e.,
    evolution of parameters).

    .. image:: /_static/tautrend01.png
       :align: center

    TauTrend inherits all the features from TauPlot (zooming,
    exporting/importing, data inspection,...) and also provides some specific
    features (e.g. fixed-range X scale mode and Archiving support).
    
    For an overview of the features from an user point of view, see the
    :ref:`TauTrend User's Interface Guide <trend_ui>`.
    
    You can also see some code that exemplifies the use of TauTrend in :ref:`the
    TauTrend coding examples <examples_tautrend>`
    
    Note: if you pass a model that is a Tango SPECTRUM attribute (instead of a
    scalar), TauTrend will interpret it as a collection of scalar values and
    will plot a separate trend line for each.


    
    .. seealso:: :class:`TauPlot`,
                 :ref:`TauTrend User's Interface Guide <trend_ui>`, 
                 :ref:`The TauTrend coding examples <examples_tautrend>`
    '''
    def __init__(self, parent = None, designMode = False):
        TauPlot.__init__(self, parent = parent, designMode = designMode)
        self.trendSets = CaselessDict()
        self._supportedConfigVersions = ["ttc-1"]
        #Use a rotated labels x timescale by default
        self.setXIsTime(True)
        rotation = -45
        alignment = self.getDefaultAxisLabelsAlignment(Qwt5.QwtPlot.xBottom, rotation)
        self.setAxisLabelRotation(Qwt5.QwtPlot.xBottom, rotation)
        self.setAxisLabelAlignment(Qwt5.QwtPlot.xBottom,alignment)
        
        self._xDynScaleSupported = True
        self._useArchiving = True
        self.setDefaultCurvesTitle('<label><[trend_index]>')
        self._maxDataBufferSize = 1048576 #(=2**20, i.e., 1M events))
    
    def updateCurves(self, names):
        '''Defines the curves that need to be plotted. For a TauTrend, the
        models can refer to:
        
        - PyTango.SCALARS: they are to be plotted in a trend
        - PyTango.SPECTRUM: each element of the spectrum is considered
          independently as in a)
        
        Note that passing an attribute for X values makes no sense in this case
        
        Internally, every curve is grouped in a TauTrendSet. For each SPECTRUM
        attribute, a TrendSet is created, containing as many curves as the
        lenght of the spectrum For eacha SCALAR attribute, a TrendSet containing
        just one curve is created.
        
        :param names: (sequence<str>) a sequence of model names
        
        .. note:: Adding/removing a model will add/remove a whole set. No
                  sub-set adding/removing is allowed.
                  Still, each curve will be independent regarding its
                  properties, and can be hidden/shown independently.
        
        .. seealso:: :meth:`TauPlot.updateCurves`
        '''
        self.curves_lock.acquire()
        try:
            # For it to work properly, 'names' must be a CaselessList, just as
            # self.trendSets is a CaselessDict
            del_sets = [ name for name in self.trendSets.keys() if name not in names]
            
            for name in names:
                name = str(name)
                if "|" in name: raise ValueError('composed ("X|Y") models are not supported by TauTrend')
                
                #create a new TrendSet if not already there
                if not self.trendSets.has_key(name):
                    tset = self.trendSets[name] = TauTrendsSet(name, parent=self)
                    tset.registerDataChanged(self, self.curveDataChanged)
            # Trend Sets to be removed
            for name in del_sets:
                name = str(name)
                tset = self.trendSets.pop(name)
                tset.unregisterDataChanged(self, self.curveDataChanged)
                for subname in tset.getCurveNames():
                    self.detachRawData(subname)
            if del_sets:        
                self.autoShowYAxes()
                
            # legend
            self.showLegend(len(self.curves) > 1, forever=False)
            self.replot()
            
        finally:
            self.curves_lock.release()

    def getTrendSetNames(self):
        '''returns the names of all TrendSets attached to this TauTrend.
        
        :return: (list<str>) a copy of self.trendSets.keys()
        '''
        self.curves_lock.acquire()
        try:
            ret = copy.deepcopy(self.trendSets.keys())
        finally:
            self.curves_lock.release()
        return ret
    
    def getTrendSet(self, name):
        '''gets a trend set object by name.
        
        **Important**: Note that the TrendSet object is not thread safe.
        Therefore, if you access it you must do it protected by the
        TauTrend.curves_lock reentrant lock.
        
        :param name: (str) the trend set name
        
        :return: (TauTrendSet) the trend set object corresponding to name
        '''
        self.curves_lock.acquire()
        try:
            ret = self.trendSets.get(str(name))
        finally:
            self.curves_lock.release()
        return ret
    
    def changeCurvesTitlesDialog(self, curveNamesList=None):
        '''Shows a dialog to set the curves titles (it will change the current
        curves titles and the default curves titles)
                
        :param curveNamesList: (string_sequence or string_iterator) names of the
                               curves to which the title will be changed (if
                               None given , it will apply to all the TrendsSets
                               and it will also be used as default for newly
                               created ones)
        
        :return: (caselessDict<str,QString> or None) The return value will be
                 `None` if `curveNamesList` is None. Otherwise it will be a
                 dictionary with key=curvename and value=newtitle. 
        
        .. seealso:: :meth:`setCurvesTitle`, :meth:`setDefaultCurvesTitle`
        '''
        newTitlesDict = None
        titletext, ok = Qt.QInputDialog.getText( self,
                                                'New Title for Curves',
                                                'New text to be used for the curves.'\
                                                'You can use any of the following placeholders:\n'\
                                                '<label>, <model>, <attr_name>, <attr_full_name>,'\
                                                '<dev_alias>, <dev_name>, <dev_full_name>,' \
                                                '<current_title>, <trend_index>, <[trend_index]>',
                                                Qt.QLineEdit.Normal,
                                                self._defaultCurvesTitle)
        if ok:
            titletext = str(titletext)
            if curveNamesList is None:
                self.setDefaultCurvesTitle(titletext)
                self.setTrendSetsTitles(titletext)
            else:
                self.curves_lock.acquire()
                try:
                    newTitlesDict = CaselessDict() 
                    for curveName in curveNamesList:
                        curvetitle = titletext
                        for ts in self.trendSets.itervalues():
                            if curveName in ts:
                                curvetitle = ts.compileBaseTile(curvetitle)
                                curvetitle = curvetitle.replace('<trend_index>', "%i"%ts.index(curveName))
                                break
                        curve = self.curves.get(curveName)
                        curve.setTitleText(curvetitle)
                        newTitlesDict[curveName] = curve.title().text()
                    self.updateLegend(self.legend())
                    return newTitlesDict
                finally:
                    self.curves_lock.release()
        return newTitlesDict
    
    def setTrendSetsTitles(self, basetitle, setNames=None):
        '''Calls setTitleText(basetitle) for each Trend Set set in setNames
        
        :param basetitle: (str) the base title
        :param setNames: (sequence<str> or iterator<str>) names of the sets to be changed
        
        See: TauTrendsSet.setTitleText
        '''
        self.curves_lock.acquire()
        try:
            if setNames is None: setNames = self.trendSets.iterkeys()
            for tname in setNames:
                if tname in self.trendSets:
                    self.trendSets[tname].setTitleText(basetitle)
        finally:
            self.curves_lock.release()
        self.updateLegend(self.legend())

    @Qt.pyqtSignature("dataChanged(const QString &)")
    def curveDataChanged(self, name):
        '''slot that is called whenever a curve emits a dataChanged signal
        
        :emits: "dataChanged(const QString &)"
        
        :param name: (str) curve name
        '''
        name=str(name)
        self.curves_lock.acquire()
        try:
            for n,curve in self.trendSets[name].getCurves():
                curve.setData(curve._xValues,curve._yValues)
            #self._zoomer.setZoomBase()
            if self.getXDynScale() and len(curve._xValues)>0: #keep the scale width constant, but translate it to get the last value
                max= curve._xValues[-1]
                min=max-self.getXAxisRange() 
                self.setAxisScale(Qwt5.QwtPlot.xBottom, min, max)
        finally:
            self.curves_lock.release()
        self.emit(Qt.SIGNAL("dataChanged(const QString &)"), Qt.QString(name))
        self.replot()
    
    def setPaused(self, paused = True):
        '''Pauses itself and other listeners (e.g. the trendsets) depending on it
        
        .. seealso:: :meth:`TauBaseComponent.setPaused`
        '''
        for ts in self.trendSets.itervalues():
            ts.setPaused(paused)
        self._isPaused = paused
    
    def createConfig(self, names=None):
        '''Returns a pickable dictionary containing all relevant information
        about the current plot.
        For Tango attributes it stores the attribute name and the curve properties
        For raw data curves, it stores the data as well.
        
        Hint: The following code allows you to serialize the configuration
        dictionary as a string (which you can store as a QSetting, or as a Tango
        Attribute)::
        
            import pickle
            c = pickle.dumps(tauplot.createConfig())  #c is a string that can be stored
                
        :param names:  (sequence<str>) a sequence of TrendSet names for which the
                       configuration will be stored (all by default).
        
        :return: (dict) configurations (which can be loaded with applyConfig)
        '''
        configdict = TauPlot.createConfig(self, curvenames=names) #use the superclass configdict as a starting point
        if names is None: names = CaselessList(self.trendSets.keys())
        configdict.pop("TangoCurves") #delete the TangoCurves key since it is meaningless in a TauTrend
        tsetsdict = CaselessDict()
        rawdatadict = CaselessDict(configdict["RawData"])
        self.curves_lock.acquire()
        try:
            for tsname,ts in self.trendSets.iteritems():
                if tsname in names:
                    tsetsdict[tsname] = tsname #store a dict containing just model names (key and value are the same)
                for cname in CaselessList(ts.getCurveNames()):
                    rawdatadict.pop(cname)#clean the rawdatadict of rawdata curves that come from trendsets (but we keep the properties!)                
        finally:
            self.curves_lock.release()
        configdict["TrendSets"] = tsetsdict
        configdict["RawData"] = rawdatadict
        return configdict
        
    def applyConfig(self, configdict, **kwargs):
        """applies the settings stored in a configdict to the current plot.
        
        :param configdict: (dict)
        
        .. seealso:: :meth:`createConfig`
        """
        if not self.checkConfigVersion(configdict): return
        #attach the curves
        for rd in configdict["RawData"].values(): self.attachRawData(rd)
        self.addModels(configdict["TrendSets"].values())
        #set curve properties
        self.setCurveAppearanceProperties(configdict["CurveProp"])
        self.updateLegend(force=True)
        #set the axes
        self.applyAxesConfig(configdict["Axes"])
        #set other misc configurations
        self.applyMiscConfig(configdict["Misc"])
        
    def setEventFilters(self, filters=None, tsetnames=None):
        '''propagates a list of tau filters to the curves given by curvenames.
        See :meth:`TauBaseComponent.setEventFilters`
        '''
        if tsetnames is None: tsetnames=self.trendSets.keys()
        self.curves_lock.acquire()
        try:
            for name in tsetnames:
                self.trendSets[name].setEventFilters(filters)
        finally:
            self.curves_lock.release()
            
    def setUseArchiving(self, enable):
        '''enables/disables looking up in the archiver for data stored before
        the Trend was started
        
        :param enable: (bool) if True, archiving values will be used if available
        '''
        self._useArchiving = enable
        self.replot()

    def getUseArchiving(self):
        '''whether TauTrend is looking for data in the archiver when needed
        
        :return: (bool)
        
        .. seealso:: :meth:`setUseArchiving`
        '''
        return self._useArchiving

    def resetUseArchiving(self):
        '''Same as setUseArchiving(True)'''
        self.setUseArchiving(True)   
    
    def setMaxDataBufferSize(self, maxSize):
        '''sets the maximum number of events that can be plotted in the trends
        
        :param maxSize: (int) the maximum limit
        
        .. seealso:: :meth:`TauTrendSet.setMaxDataBufferSize`
        '''
        self.curves_lock.acquire()
        try:
            for ts in self.trendSets.itervalues():
                ts.setMaxDataBufferSize(maxSize)
        finally:
            self.curves_lock.release()
        self._maxDataBufferSize = maxSize
        
    def getMaxDataBufferSize(self):
        '''returns the maximum number of events that can be plotted in the trend
        
        :return: (int)
        '''
        return self._maxDataBufferSize
            
    def resetMaxDataBufferSize(self):
        '''Same as setUseArchiving(1048576)  (i.e. 1M of events)'''
        self.setMaxDataBufferSize(1048576)
                 
    useArchiving = Qt.pyqtProperty("bool", getUseArchiving, setUseArchiving, resetUseArchiving)
    maxDataBufferSize = Qt.pyqtProperty("int", getMaxDataBufferSize, setMaxDataBufferSize, resetMaxDataBufferSize)
    


if __name__ == "__main__":
    import sys
    app = Qt.QApplication(sys.argv)
    
    args=sys.argv[1:]
    
    KLASS = TauTrend
    SHOW = True
    EXPORT = None
    CONFIG = None
    MODELS = []
    XISTIME = True
    
    ## MANAGING ARGUMENTS 
    #----------------------------------------------    
    for a in args:
        if a == "-xt":  #argument "-xt" means interprete X values as time
            XISTIME = True
        elif a == "-xe":  #argument "-xe" means interprete X values as events
            XISTIME = False
        elif a.startswith('-config='): #argument "-conf=FILENAME" launches TauTrend/Plot with a predefined config file
            CONFIG = a.split('=')[-1]
        elif a.startswith('-export='): #instead of showing a window the plot will be exported to a file
            EXPORT = a.split('=')[-1]
            SHOW = False
        elif a.startswith('-'): #whatever other argument starting by "-"
            print "\n Usage: \n%s [-xe|-xt] [-config=configfilename] [-export=exportfilename] [model1 [model2] ...]\n"%sys.argv[0]
            sys.exit(1)
        else: #anything that is not a parameter is interpreted as a model
            MODELS.append(a)
    #----------------------------------------------
    
    form =  KLASS()    
    form.setXIsTime(XISTIME)
    if CONFIG is not None: form.loadConfig(CONFIG)
    #form.setDefaultCurvesTitle("<dev_full_name><[trend_index]>")
    if MODELS: form.setModel(MODELS)  
    if EXPORT is not None:
        curves = dict.fromkeys(form.trendSets.keys(),0)        
        def exportIfAllCurves(curve,trend=form,counters=curves):
            curve = str(curve)
            print '*'*10 + ' %s: Event received for %s  '%(datetime.now().isoformat(),curve) +'*'*10        
            if curve in counters:
                counters[curve]+=1
                if all(counters.values()):
                    trend.exportPdf(EXPORT)
                    print '*'*10 + ' %s: Exported to : %s  '%(datetime.now().isoformat(),EXPORT) +'*'*10        
                    trend.close()
            return
        if not curves: form.close()
        else:
            for ts in form.trendSets.values():
                Qt.QObject.connect(ts._signalGen, ts._signalGen.newDataChangedSignal(), exportIfAllCurves)    
    
    if SHOW:
        form.show()
        #if no models are passed, show the data import dialog
        if len(MODELS) == 0 and CONFIG is None:
            form.showDataImportDlg()
    sys.exit(app.exec_())
