
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
scalaredit.py: 
"""

import sys
from PyQt4 import QtCore, QtGui, Qt
import PyTango
import tau.core
import edit
import colors
from tau.widget import TauBaseWidget

class TauBaseWritableWidget(TauBaseWidget):
    
    def __init__(self, name, tau_parent=None, designMode = False):
        self.call__init__(TauBaseWidget, name, parent=tau_parent, designMode=designMode)
        
        self._lastValue = None
        
        # Overwrite not to show quality by default
        self._showQuality = False
        
        # Don't do auto-apply by default
        self._autoApply = False
        
        # Don't force a writing to attribute when there are not pending operations
        self._forcedApply = False
        
        self.connect(self, Qt.SIGNAL('valueChanged'), self.updatePendingOperations)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        return tau.core.TauAttribute

    def isReadOnly(self):
        return False
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setAutoApply(self, auto):
        self._autoApply = auto
        
    def getAutoApply(self):
        return self._autoApply

    def resetAutoApply(self):
        self.setAutoApply(False)
        
    def setForcedApply(self, forced):
        self._forcedApply = forced
        
    def getForcedApply(self):
        return self._forcedApply

    def resetForcedApply(self):
        self.setForcedApply(False)
        
    def valueChanged(self, *args):
        '''Subclasses should connect some particular signal to this method for
        indicating that something has changed.
        e.g., a QLineEdit should connect its "textChanged" signal...'''
        self.emitValueChanged()
        if self._autoApply:
            self.writeValue()

    def writeValue(self, forceApply=False):
        '''Writes the value to the attribute, either by applying pending
        operations or (if the ForcedApply flag is True), it wites directly when
        no operations are pending'''
        
        if self.hasPendingOperations():
            self.safeApplyOperations()
            return
        
        #maybe we want to force an apply even if there are no pending ops...
        kmods = Qt.QCoreApplication.instance().keyboardModifiers()
        controlpressed = bool(kmods&Qt.Qt.ControlModifier)
        if self.getForcedApply() or forceApply or controlpressed:
            self.forceApply()
            
    def forceApply(self):
        '''It (re)applies the value regardless of pending operations.
        WARNING: USE WITH CARE. In most cases what you need is to make sure
        that pending operations are properly created, not calling this method'''
        try:
            v = self.getValue()
            op = tau.core.WriteAttrOperation(self.getModelObj(), v, 
                                            self.getOperationCallbacks())
            op.setDangerMessage(self.getDangerMessage())
            self.safeApplyOperations([op])
            self.info('Force-Applied value = %s'%str(v))
        except:
            self.error('Unexpected exception in forceApply')
            self.traceback()
        
    def handleEvent(self, src, evt_type, evt_value):
        if evt_type in (tau.core.TauEventType.Change, tau.core.TauEventType.Periodic):
            self.emitValueChanged()
    
    def postAttach(self):
        TauBaseWidget.postAttach(self)
        if self.isAttached():
            try:
                v = self.getModelValueObj().w_value
            except:
                v = None
            self.setValue(v)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def resetPendingOperations(self):
        if self.isAttached():
            try:
                v = self.getModelValueObj().w_value
            except:
                v = None
            self.setValue(v)
        TauBaseWidget.resetPendingOperations(self)
        self.updateStyle()

    def updatePendingOperations(self):
        model = self.getModelObj()
        try:
            model_value = model.getValueObj().w_value
            wigdet_value = self.getValue()
            if model.areStrValuesEqual(model_value, wigdet_value):
                self._operations = []
            else:
                operation = tau.core.WriteAttrOperation(model, wigdet_value, 
                                                            self.getOperationCallbacks())
                operation.setDangerMessage(self.getDangerMessage())
                self._operations = [operation]
        except:
            self._operations = []
        self.updateStyle()
        
    def getOperationCallbacks(self):
        return []

    def getValue(self):
        raise RuntimeError("Not allowed to call TauBaseWritableWidget.getValue()")

    def setValue(self, v):
        raise RuntimeError("Not allowed to call TauBaseWritableWidget.setValue()")

    def updateStyle(self):
        toolTip = self.getFormatedToolTip()
        if self.hasPendingOperations():
            v_str = str(self.getValue())
            model_v_str = getattr(self.getModelValueObj(),'w_value', '-----')
            toolTip += '<hr/>Displayed value (%s) differs from applied value (%s)' % (v_str, model_v_str)
        self.setToolTip(toolTip)

    def _updateValidator(self, evt_value):
        #re-set the validator ranges if applicable
        if evt_value is None: return
        v = self.validator()
        if isinstance(v, Qt.QIntValidator):
            bottom = evt_value.min_value
            top = evt_value.max_value
            bottom = int(bottom) if bottom != tau.core.TauConfiguration.no_min_value else -sys.maxint
            top = int(top) if top != tau.core.TauConfiguration.no_max_value else sys.maxint
            v.setRange(bottom, top)
            self.debug("Validator range set to %i-%i"%(bottom,top))
        elif isinstance(v, Qt.QDoubleValidator):
            bottom = evt_value.min_value
            top = evt_value.max_value
            bottom = float(bottom) if bottom != tau.core.TauConfiguration.no_min_value else -float("inf") 
            top = float(top) if top != tau.core.TauConfiguration.no_max_value else float("inf")
            v.setBottom(bottom)
            v.setTop(top)
            self.debug("Validator range set to %f-%f"%(bottom,top))        


class TauWheelEdit(edit.WheelEdit, TauBaseWritableWidget):
    __pyqtSignals__ = ('modelChanged(const QString &)',)

    def __init__(self, qt_parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(edit.WheelEdit, qt_parent)
        self.call__init__(TauBaseWritableWidget, name, designMode=designMode)
        self.connect(self, Qt.SIGNAL(edit.WheelEdit.NumberChangedStr), self.valueChanged)
        self.connect(self, Qt.SIGNAL('returnPressed()'), self.writeValue)
        self.connect(self, Qt.SIGNAL('valueChanged'), self.updatePendingOperations)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_type == tau.core.TauEventType.Config and not evt_value is None:
            f = evt_value.format.lower()
            if f[-1] not in ('d', 'f', 'g'):
                raise ValueError("'%s' format unsupported" % f)
            f = f.replace('g','f')
            if 'd' == f[-1]:
                dec_nb = 0
                try: total = int(f[1:-1])
                except: total = 6
            else:
                f = map(int, f[1:-1].split('.', 1))
                total = f[0]
                if len(f)>1:
                    dec_nb = f[1]
                else:
                    dec_nb = 2
            int_nb = total-dec_nb
            self.setDigitCount(int_nb=int_nb, dec_nb=dec_nb)
            try: self.setMinValue(float(evt_value.min_value))
            except: pass
            try: self.setMaxValue(float(evt_value.max_value))
            except: pass
        TauBaseWritableWidget.handleEvent(self, evt_src, evt_type, evt_value)
    
    def updateStyle(self):
        TauBaseWritableWidget.updateStyle(self)
        if self.hasPendingOperations():
            self.setStyleSheet('WheelEdit {border: 2px solid; border-radius: 4px; border-color: blue}')
        else:
            self.setStyleSheet('WheelEdit {border: 2px solid; border-radius: 4px; border-color: rgba(0,0,255,0)}')
            
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = QtCore.pyqtProperty("QString", TauBaseWidget.getModel,
                                TauBaseWidget.setModel,
                                TauBaseWidget.resetModel)
                                
    useParentModel = QtCore.pyqtProperty("bool", 
                                         TauBaseWidget.getUseParentModel,
                                         TauBaseWidget.setUseParentModel,
                                         TauBaseWidget.resetUseParentModel)

    autoApply = QtCore.pyqtProperty("bool", TauBaseWritableWidget.getAutoApply,
                                    TauBaseWritableWidget.setAutoApply,
                                    TauBaseWritableWidget.resetAutoApply)
    
    forcedApply = QtCore.pyqtProperty("bool", TauBaseWritableWidget.getForcedApply,
                                    TauBaseWritableWidget.setForcedApply,
                                    TauBaseWritableWidget.resetForcedApply)


class TauValueLineEdit(QtGui.QLineEdit, TauBaseWritableWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, qt_parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(QtGui.QLineEdit, qt_parent)
        self.call__init__(TauBaseWritableWidget, name, designMode=designMode)
        self._enableWheelEvent = False
        self.__minAlarm = -float("inf")
        self.__maxAlarm = float("inf")
        self.__minLimit = -float("inf")
        self.__maxLimit = float("inf")

        self.setAlignment(QtCore.Qt.AlignRight)
        self.setValidator(None)
        
        self.connect(self, Qt.SIGNAL('textChanged(const QString &)'), self.valueChanged)
        self.connect(self, Qt.SIGNAL('returnPressed()'), self.writeValue)
        self.connect(self, Qt.SIGNAL('valueChanged'), self.updatePendingOperations)
        
    def _updateValidator(self, attrinfo):
        '''This method sets a validator depending on the data type
        attrinfo is an AttributeInfoEx object'''
        if PyTango.is_int_type(attrinfo.data_type):
            validator = Qt.QIntValidator(self) #initial range is -2147483648 to 2147483647 (and cannot be set larger)
            if validator.bottom() < self.__minLimit < validator.top(): 
                validator.setBottom(int(self.__minLimit))
            if validator.bottom() < self.__maxLimit < validator.top():
                validator.setTop(int(self.__maxLimit))
            self.setValidator(validator)
            self.debug("IntValidator set with limits=[%d,%d]"%(validator.bottom(), validator.top()))
        elif PyTango.is_float_type(attrinfo.data_type):
            validator= Qt.QDoubleValidator(self)
            validator.setBottom(self.__minLimit)
            validator.setTop(self.__maxLimit)
            decimalDigits = self.__decimalDigits(attrinfo.format)
            if decimalDigits is not None:
                validator.setDecimals(decimalDigits)
            self.setValidator(validator)
            self.debug("DoubleValidator set with limits=[%f,%f]"%(self.__minLimit, self.__maxLimit))
        else: #@TODO Other validators can be configured for other types (e.g. with string lengths, tango names,...)
            self.setValidator(None)
            self.debug("Validator disabled")
    
    def __decimalDigits(self, fmt):
        '''returns the number of decimal digits from a format string
        (or None if they are not defined)''' 
        try:
            if fmt[-1].lower() in ['f','g'] and '.' in fmt:
                return int(fmt[:-1].split('.')[-1])
            else:
                return None
        except:
            return None
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWritableWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_type == tau.core.TauEventType.Config:
            if evt_value.min_alarm != tau.core.TauConfiguration.no_min_alarm:
                self.__minAlarm = float(evt_value.min_alarm)
            else:
                self.__minAlarm = -float("inf")
            if evt_value.max_alarm != tau.core.TauConfiguration.no_max_alarm:
                self.__maxAlarm = float(evt_value.max_alarm)
            else:
                self.__maxAlarm = float("inf")
            if evt_value.min_value != tau.core.TauConfiguration.no_min_value:
                self.__minLimit = float(evt_value.min_value)
            else:
                self.__minLimit = -float("inf")
            if evt_value.max_value != tau.core.TauConfiguration.no_max_value:
                self.__maxLimit = float(evt_value.max_value)
            else:
                self.__maxLimit = float("inf")
                
            self._updateValidator(evt_value)
        TauBaseWritableWidget.handleEvent(self, evt_src, evt_type, evt_value)
    
    def _inAlarm(self, v):
        try: return not(self.__minAlarm < float(v) < self.__maxAlarm)
        except: return False #this will return false for non-numerical values
    
    def _outOfRange(self, v):
        validator = self.validator()
        if validator:
            return validator.validate(Qt.QString(str(v)),0)[0] != validator.Acceptable
        else: #fallback, only for numeric types (returns False for other types)
            try: return not(self.__minLimit <= float(v) <=  self.__maxLimit)
            except: return False 
        
    def updateStyle(self):
        TauBaseWritableWidget.updateStyle(self)
        color, weight = 'black', 'normal' #default case: the value is in normal range with no pending changes
        v = self.getValue()
        if self._outOfRange(v): #the value is invalid and can't be applied
            color = 'gray'
        elif self._inAlarm(v): #the value is valid but in alarm range...
            color = 'orange'
            if self.hasPendingOperations(): #...and some change is pending
                weight = 'bold'
        elif self.hasPendingOperations(): #the value is in valid range with pending changes
            color, weight= 'blue','bold'
        self.setStyleSheet('TauValueLineEdit {color: %s; font-weight: %s}'%(color,weight))

    def wheelEvent(self, evt):
        if not self.getEnableWheelEvent() or QtGui.QLineEdit.isReadOnly(self):
            return QtGui.QLineEdit.wheelEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return QtGui.QLineEdit.wheelEvent(self, evt)

        evt.accept()
        numDegrees = evt.delta() / 8
        numSteps = numDegrees / 15
        modifiers = evt.modifiers()
        if modifiers & Qt.Qt.ControlModifier:
            numSteps *= 10
        elif (modifiers & Qt.Qt.AltModifier) and model.isFloat():
            numSteps *= .1
        self._stepBy(numSteps)

    def keyPressEvent(self, evt):
        if QtGui.QLineEdit.isReadOnly(self):
            return QtGui.QLineEdit.keyPressEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return QtGui.QLineEdit.keyPressEvent(self, evt)
            
        if evt.key() == Qt.Qt.Key_Up:     numSteps = 1
        elif evt.key() == Qt.Qt.Key_Down: numSteps = -1
        else: return QtGui.QLineEdit.keyPressEvent(self, evt)
        
        evt.accept()
        modifiers = evt.modifiers()
        if modifiers & Qt.Qt.ControlModifier:
            numSteps *= 10
        elif (modifiers & Qt.Qt.AltModifier) and model.isFloat():
            numSteps *= .1
        self._stepBy(numSteps)
    
    def _stepBy(self, v):
        self.setValue(self.getValue() + v)

    def setValue(self, v):
        model = self.getModelObj()
        if model is None:
            v_str = str(v)
        else:
            v_str = str(model.displayValue(v))
        v_str = v_str.strip()
        self.setText(v_str)
    
    def getValue(self):
        v_qstr = self.text()
        model = self.getModelObj()
        try:
            return model.encode(v_qstr)
        except:
            return None
    
    def setEnableWheelEvent(self, b):
        self._enableWheelEvent = b

    def getEnableWheelEvent(self):
        return self._enableWheelEvent
    
    def resetEnableWheelEvent(self):
        self.setEnableWheelEvent(False)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = QtCore.pyqtProperty("QString", TauBaseWidget.getModel,
                                TauBaseWidget.setModel,
                                TauBaseWidget.resetModel)
                                
    useParentModel = QtCore.pyqtProperty("bool", 
                                         TauBaseWidget.getUseParentModel,
                                         TauBaseWidget.setUseParentModel,
                                         TauBaseWidget.resetUseParentModel)

    autoApply = QtCore.pyqtProperty("bool", TauBaseWritableWidget.getAutoApply,
                                    TauBaseWritableWidget.setAutoApply,
                                    TauBaseWritableWidget.resetAutoApply)
    
    forcedApply = QtCore.pyqtProperty("bool", TauBaseWritableWidget.getForcedApply,
                                      TauBaseWritableWidget.setForcedApply,
                                      TauBaseWritableWidget.resetForcedApply)
    
    enableWheelEvent = QtCore.pyqtProperty("bool", getEnableWheelEvent,
                                           setEnableWheelEvent,
                                           resetEnableWheelEvent)

class TauValueSpinBox(QtGui.QAbstractSpinBox):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, qt_parent = None, designMode = False):
        QtGui.QAbstractSpinBox.__init__(self, qt_parent)
        
        # Overwrite not to show quality by default
        self._showQuality = False
        
        self._singleStep = 1.0
        
        self.setLineEdit(TauValueLineEdit(designMode=designMode))
        self.setAccelerated(True)

    def __getattr__(self, name):
        return getattr(self.lineEdit(), name)

    # The minimum size of the widget (a limit for the user)
    def minimumSizeHint(self):
        return QtCore.QSize(20, 20)
    
    # The default size of the widget
    def sizeHint(self):
        return QtCore.QSize(80, 24)
        
    def setValue(self, v):
        self.lineEdit().setValue(v)
    
    def getValue(self):
        return self.lineEdit().getValue()
    
    def keyPressEvent(self, event):
        if event.key() in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            self.lineEdit().writeValue()
            event.accept()
        else:
            QtGui.QAbstractSpinBox.keyPressEvent(self,event)
            event.ignore()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory overload from QAbstractSpinBox
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def stepBy(self, steps):
        #validator = self.lineEdit().validator()
        self.setValue(self.getValue() + self.getSingleStep()*steps)
    
    def stepEnabled(self):
        le, curr_val, ss = self.lineEdit(), self.getValue(), self.getSingleStep()
        ret = QtGui.QAbstractSpinBox.StepEnabled(QtGui.QAbstractSpinBox.StepNone)
        
        if curr_val == None:
            return ret
        
        if not le._outOfRange(curr_val + ss):
            ret |= QtGui.QAbstractSpinBox.StepUpEnabled
        if not le._outOfRange(curr_val - ss):
            ret |= QtGui.QAbstractSpinBox.StepDownEnabled
        return ret
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Model related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def setModel(self, model):
        self.lineEdit().setModel(model)
    
    def getModel(self):
        return self.lineEdit().getModel()
    
    def resetModel(self):
        return self.lineEdit().resetModel()

    def setUseParentModel(self, model):
        self.lineEdit().setUseParentModel(model)
    
    def getUseParentModel(self):
        return self.lineEdit().getUseParentModel()
    
    def resetUseParentModel(self):
        return self.lineEdit().resetUseParentModel()

    def setAutoApply(self, model):
        self.lineEdit().setAutoApply(model)
    
    def getAutoApply(self):
        return self.lineEdit().getAutoApply()
    
    def resetAutoApply(self):
        return self.lineEdit().resetAutoApply()
        
    def setForcedApply(self, model):
        self.lineEdit().setAutoApply(model)
    
    def getForcedApply(self):
        return self.lineEdit().getAutoApply()
    
    def resetForcedApply(self):
        return self.lineEdit().resetAutoApply()
    
    def getSingleStep(self):
        return self._singleStep
    
    def setSingleStep(self, step):
        self._singleStep = step
    
    def resetSingleStep(self):
        self.setSingleStep(1.0)
    
    singleStep = QtCore.pyqtProperty("double", getSingleStep, setSingleStep,
                                     resetSingleStep)
    
    model = QtCore.pyqtProperty("QString", getModel, setModel, resetModel)
                                
    useParentModel = QtCore.pyqtProperty("bool", getUseParentModel,
                                         setUseParentModel, resetUseParentModel)

    autoApply = QtCore.pyqtProperty("bool", getAutoApply, setAutoApply,
                                    resetAutoApply)

    forcedApply = QtCore.pyqtProperty("bool", getForcedApply, setForcedApply,
                                    resetForcedApply)


## TauConfigLineEdit has been marked as obsolete because it is not adapted to new 
## event handling and it is not clear wether anyone is interested in keeping it
## If you want to use it, contact cpascual@cells.es or tcoutinho@cells.es
class TauConfigLineEdit(QtGui.QLineEdit, TauBaseWidget):

    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, qt_parent = None, designMode = False):
        
        raise DeprecationWarning('TauConfigLineEdit is deprecated. It does not integrate well with latest tau features')
        
        name = self.__class__.__name__
        self._lastValueByUser = ''
        self.call__init__wo_kw(QtGui.QLineEdit, qt_parent)
        self.call__init__(TauBaseWidget, name, designMode=designMode)
        
        self.setObjectName(name)
        
        self._dftShadow = self.palette().shadow()
        self._dftHighlight = self.palette().highlight()
        self._dftButton = self.palette().button()
        
        self.defineStyle()
        
        self.connect(self, Qt.SIGNAL('textChanged(QString)'),self.emitValueChanged)
        
    def defineStyle(self):
        self.setAlignment(QtCore.Qt.AlignRight)
        self.updateStyle()
        
    # The minimum size of the widget (a limit for the user)
    def minimumSizeHint(self):
        return QtCore.QSize(20, 20)
    
    # The default size of the widget
    def sizeHint(self):
        return QtCore.QSize(60, 24)
                
    def focusInEvent(self,focusEvent):
        QtGui.QLineEdit.focusInEvent(self,focusEvent)
        self.updateStyle()

    def focusOutEvent(self,focusEvent):
        QtGui.QLineEdit.focusOutEvent(self,focusEvent)
        self.updateStyle()
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        return tau.core.TauConfiguration
    
    def preAttach(self):
        TauBaseWidget.preAttach(self)
        QtCore.QObject.connect(self, QtCore.SIGNAL("returnPressed()"), 
                               self.writeTextValue)
        QtCore.QObject.connect(self, QtCore.SIGNAL("textChanged(QString)"), 
                               self.userChangedValue)

    def postAttach(self):
        TauBaseWidget.postAttach(self)

        # Set an initial value if necessary
        if self.isAttached() and not self.isReadOnly():
            value_str = self.getDisplayValue()
            self._lastValueByUser = value_str
            self.emit(QtCore.SIGNAL('valueChangedDueToEvent(QString)'), 
                      QtCore.QString(value_str))

    def postDetach(self):
        TauBaseWidget.postDetach(self)
        QtCore.QObject.disconnect(self, QtCore.SIGNAL("returnPressed()"), 
                                  self.writeTextValue)
        QtCore.QObject.disconnect(self, QtCore.SIGNAL("textChanged(QString)"), 
                                  self.userChangedValue)

    def getConfigurationAttributeValue(self):
        """Helper method to get the proper parameter value from the configuration"""
        model = self.getModelObj()
        if model is None:
            return None
        try:
            return getattr(model, str(self._configParam))
        except:
            self.debug("Invalid configuration parameter %s" % self._configParam)
            return None
        
    def getDisplayValue(self):
        cfg_value = self.getConfigurationAttributeValue()
        if cfg_value is None:
            cfg_value = self.getNoneValue()
        else:
            cfg_value = str(cfg_value)
        return cfg_value

    def isValueChangedByUser(self):
        localValue = str(self.text())
        return not localValue == self._lastValueByUser

    def updateStyle(self):
        p = self.palette()
        
        p.setBrush(QtGui.QPalette.Base, QtCore.Qt.white)
        p.setBrush(QtGui.QPalette.Text,QtCore.Qt.black)
        
        if not self.isReadOnly() and self.isValueChangedByUser():
            if self.hasFocus():
                p.setBrush(QtGui.QPalette.Highlight, QtCore.Qt.blue)
            else:
                p.setBrush(QtGui.QPalette.Shadow, QtCore.Qt.blue)
                p.setBrush(QtGui.QPalette.Button, QtCore.Qt.blue)
        else:
            p.setBrush(QtGui.QPalette.Shadow, self._dftShadow)
            p.setBrush(QtGui.QPalette.Button, self._dftButton)
            p.setBrush(QtGui.QPalette.Highlight, self._dftHighlight)
        self.updatePendingOpsStyle()
        self.update()


    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    @QtCore.pyqtSignature("setModel(QString)")
    def setModel(self, model):
        model = str(model)
        try:
            self._configParam = model[model.rfind('=')+1:].lower()
        except:
            self._configParam = ''
        TauBaseWidget.setModel(self,model)   

    model = QtCore.pyqtProperty("QString", TauBaseWidget.getModel, 
                                setModel, 
                                TauBaseWidget.resetModel)
                                
    useParentModel = QtCore.pyqtProperty("bool", 
                                         TauBaseWidget.getUseParentModel, 
                                         TauBaseWidget.setUseParentModel, 
                                         TauBaseWidget.resetUseParentModel)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT signal handlers
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @QtCore.pyqtSignature("textChanged(QString)")
    def userChangedValue(self,qstr):
        self.updateStyle()

    
    @QtCore.pyqtSignature("returnPressed(QString)")
    def writeTextValue(self):
        model = self.getModelObj()
        if not self.isReadOnly() and model:
            try:
                new_value = str(self.text())
                model.setParam(self._configParam, new_value)
                self._lastValueByUser = new_value
                self.updateStyle()
            except PyTango.DevFailed, e:
                ret = QtGui.QMessageBox.warning(self, "Write error",e[0]['desc'],
                                                QtGui.QMessageBox.Ok)


class TauValueCheckBox(QtGui.QCheckBox, TauBaseWritableWidget):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, qt_parent = None, designMode = False):
        name = "TauValueCheckBox" 
        self.call__init__wo_kw(QtGui.QCheckBox, qt_parent)
        self.call__init__(TauBaseWritableWidget, name, designMode=designMode)

        self.setObjectName(name)
        self.updateStyle()
        self.connect(self, Qt.SIGNAL('stateChanged(int)'),self.valueChanged)
            
    def keyPressEvent(self, event):
        if event.key() in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            self.writeValue()
            event.accept()
        else:
            QtGui.QCheckBox.keyPressEvent(self,event)
            event.ignore()

    def minimumSizeHint(self):
        return QtCore.QSize(20, 20)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TauBaseWritableWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def updateStyle(self):
        TauBaseWritableWidget.updateStyle(self)
        # Show text only if it is not specifically hidden
        if self._showText:
            try:
                self.setText(str(self.getModelObj().getConfig().getLabel()))
            except:
                self.setText('----')
        else:
            self.setText('')
        #Update pending operations style
        if self.hasPendingOperations():
            if self.text().trimmed().isEmpty(): self.setText("!")
            self.setStyleSheet('TauValueCheckBox {color: blue;}')
        else:
            if str(self.text()) == "!": self.setText(" ")
            self.setStyleSheet('TauValueCheckBox {}')
        self.update()
    
    def setValue(self, v):
        self.setChecked(bool(v))
    
    def getValue(self):
        return self.isChecked()
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = QtCore.pyqtProperty("QString",
                                TauBaseWidget.getModel,
                                TauBaseWidget.setModel, 
                                TauBaseWidget.resetModel)

    showText = QtCore.pyqtProperty( "bool",
                                    TauBaseWidget.getShowText,
                                    TauBaseWidget.setShowText,
                                    TauBaseWidget.resetShowText)
                                
    useParentModel = QtCore.pyqtProperty("bool", 
                                         TauBaseWidget.getUseParentModel, 
                                         TauBaseWidget.setUseParentModel,
                                         TauBaseWidget.resetUseParentModel)

    autoApply = QtCore.pyqtProperty("bool",
                                    TauBaseWritableWidget.getAutoApply,
                                    TauBaseWritableWidget.setAutoApply,
                                    TauBaseWritableWidget.resetAutoApply)
    
    forcedApply = QtCore.pyqtProperty("bool", TauBaseWritableWidget.getForcedApply,
                                    TauBaseWritableWidget.setForcedApply,
                                    TauBaseWritableWidget.resetForcedApply)

def main():
    attr_name = sys.argv[1]
    a = QtGui.QApplication([])
    panel = QtGui.QWidget()
    l = QtGui.QGridLayout()
    w1 = TauValueLineEdit()
    w2 = TauWheelEdit()
    w3 = TauValueSpinBox()
    w4 = TauValueCheckBox()
    w1.setModel(attr_name)
    w2.setModel(attr_name)
    w3.setModel(attr_name)
    
    l.addWidget(w1,0,0)
    l.addWidget(w2,1,0)
    l.addWidget(w3,2,0)
    
    if len (sys.argv) > 2:
        attr_name2 = sys.argv[2]
        w4.setModel(attr_name2)
        l.addWidget(w4,3,0)
        w4.setAutoApply(True)
        w4.setDangerMessage('AAAAAYYYY')
    
    panel.setLayout(l)
    panel.setVisible(True)
    return a.exec_()

if __name__ == "__main__":
    sys.exit(main())
