
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
jdraw.py: 
"""

import os,traceback
from PyQt4 import Qt, QtCore, QtGui

LINESTYLE_JDW2QT = { 0: Qt.Qt.SolidLine,
                     1: Qt.Qt.DotLine,
                     2: Qt.Qt.DashLine,
                     3: Qt.Qt.DashLine,
                     4: Qt.Qt.DashDotLine }

FILLSTYLE_JDW2QT = { 0: Qt.Qt.NoBrush,
                     1: Qt.Qt.SolidPattern,
                     2: Qt.Qt.BDiagPattern,
                     3: Qt.Qt.FDiagPattern,
                     4: Qt.Qt.DiagCrossPattern,
                     5: Qt.Qt.BDiagPattern,
                     6: Qt.Qt.FDiagPattern,
                     7: Qt.Qt.Dense5Pattern,
                     8: Qt.Qt.Dense7Pattern,
                     9: Qt.Qt.Dense7Pattern,
                     10:Qt.Qt.Dense7Pattern,
                     11:Qt.Qt.RadialGradientPattern }

import tau.core.utils as utils

TEXTHINT_JDW2QT = utils.CaselessDict({ 'helvetica'  : QtGui.QFont.Helvetica,
                                       'serif'      : QtGui.QFont.Serif,
                                       'sansserif'  : QtGui.QFont.SansSerif,
                                       'courier'    : QtGui.QFont.Courier,
                                       'Monospaced' : QtGui.QFont.Courier,
                                       'times'      : QtGui.QFont.Times,
                                       ''           : QtGui.QFont.AnyStyle,})

import tau.core
import tau.widget
from tau.widget import AbstractGraphicsFactory

class JDrawTauGraphicsFactory(utils.Singleton, AbstractGraphicsFactory, utils.Logger):
    
    def __init__(self,parent):
        self.myparent=parent
        self.call__init__wo_kw(AbstractGraphicsFactory, parent)
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args):
        """Singleton instance initialization."""
        self.call__init__(utils.Logger, self.__class__.__name__) 
        self.call__init__(tau.widget.AbstractGraphicsFactory)
        
    def getSceneObj(self,items):   
        scene = tau.widget.TauGraphicsScene(self.myparent)
        for item in items:
            try:
                if isinstance(item, QtGui.QWidget):
                    scene.addWidget(item)
                elif isinstance(item, QtGui.QGraphicsItem):
                    scene.addItem(item)
            except:
                self.warning("Unable to add item %s to scene" % str(item))
                self.traceback()
        return scene
    
    def getObj(self,name,params):
        method_name = 'get' + name.lstrip('JD') + 'Obj'
        try:
            method = getattr(self, method_name)
            return method(params)
        except:
            print traceback.format_exc()
            pass
        return None
    
    def getRectangleObj(self,params):
        item = self.getGraphicsItem('Rectangle',params)
        x1, y1, x2, y2 = params.get('summit')
        width = x2 - x1
        height = y2 - y1
        item.setRect(x1,y1,width,height)
        
        return item

    def getRoundRectangleObj(self,params):
        item = self.getGraphicsItem('RoundRectangle',params)
        x1, y1, x2, y2 = params.get('summit')
        width = x2 - x1
        height = y2 - y1
        
        item.setRect(x1,y1,width,height)
        
        return item
        
    def getLineObj(self,params):
        item = self.getGraphicsItem('Line',params)
        x1,y1,x2,y2 = params.get('summit')
        item.setLine(x1,y1,x2,y2)
        
        return item
    
    def getEllipseObj(self,params):
        item = self.getGraphicsItem('Ellipse',params)
        
        x1, y1, x2, y2 = params.get('summit')
        width = x2 - x1
        height = y2 - y1
        item.setRect(x1,y1,width,height)

        return item        

    def getPolylineObj(self,params):
        item = self.getGraphicsItem('Polyline',params)
        
        polygon = Qt.QPolygonF()
        p = params.get('summit')
        for i in xrange(0,len(p),2):
            polygon.append(Qt.QPointF(p[i],p[i+1]))
        item.setPolygon(polygon)

        return item
    
    def getLabelObj(self,params):
        item = self.getGraphicsItem('Label',params)
        
        s = params.get('summit')
        x1, y1 = s[0], s[1]
        item.setPos(x1,y1)

        txt = params.get('text')
        if txt:
            item.setPlainText(QtCore.QString(txt))
        fnt = params.get('font')
        if fnt:
            family,style,size = fnt
            f = QtGui.QFont(family, size, QtGui.QFont.Light, False)
            f.setStyleHint(TEXTHINT_JDW2QT.get(family, QtGui.QFont.AnyStyle))
            f.setStyleStrategy(QtGui.QFont.PreferMatch)
            if style == 1:
                f.setWeight(QtGui.QFont.DemiBold)
            elif style == 2:
                f.setItalic(True)
            elif style == 3:
                f.setWeight(QtGui.QFont.DemiBold)
                f.setItalic(True)
            #TODO: Improve code in order to be able to set a suitable font
            item.setFont(f)
        return item        
    
    def getGroupObj(self,params):
        item = self.getGraphicsItem('Group',params)
        s = params.get('summit')
        x1, y1 = s[0], s[1]
        item.setPos(x1,y1)
        children = params.get('children')
        if children:
            for child in children:
                if child:
                    item.addToGroup(child)

        return item
    
    def getSwingObjectObj(self,params):
        item = self.getGraphicsItem('SwingObject', params)
        s = params.get('summit')
        x1, y1 = s[0], s[1]
        item.setPos(x1,y1)
        ext = params.get('extensions')
        return item
    
    def getImageObj(self,params):
        item = self.getGraphicsItem('Image',params)
        s = params.get('summit')
        x1, y1 , x2, y2 = s
        item.setPos(x1,y1)
        
        fname = params.get('file_name')
        if fname:
            print 'Opening JDImage(%s) = x,y,w,h=%f,%f,%f,%f' % (fname,x1,y1,x2-x1,y2-y1)
            fname = os.path.realpath(fname)
            pixmap = QtGui.QPixmap(fname)
            item.setPixmap(pixmap.scaled(x2-x1,y2-y1))
            #item.scale(float(w)/pixmap.width(), float(h)/pixmap.height())
        return item
        
    def set_common_params(self,item,params):
        if not item:
            return

        visibilitymapper = params.get('visibilitymapper')
        if not visibilitymapper is None:
            mapping_type = visibilitymapper['mapping_type']
            mode = visibilitymapper['mode']
            default = visibilitymapper['default']
            map = visibilitymapper['map']
            item._default = default
            item._map = map      

        visible = params.get('visible')
        if not visible is None:
            item.setVisible(visible)

        name = params.get('name')
        extensions = params.get('extensions')
        if extensions:
            item._extensions = extensions

        try:
            getattr(item,'setPen')
            fg = params.get("foreground", (0,0,0))
            pen = Qt.QPen(Qt.QColor(fg[0],fg[1],fg[2]))
            pen.setWidth(params.get("lineWidth", 1))
            pen.setStyle(LINESTYLE_JDW2QT[params.get("lineStyle", 0)])
            item.setPen(pen)
        except:
            pass

        fillStyle = FILLSTYLE_JDW2QT[params.get('fillStyle', 0)]

        try:
            getattr(item,'setBrush')
            brush = Qt.QBrush() 
            if fillStyle == Qt.Qt.RadialGradientPattern:
                x1, y1, x2, y2 = params.get('summit')
                w, h = (x2-x1)/2.0, (y2-y1)/2.0
                gradient = QtGui.QLinearGradient(params.get('gradX1',0)+w,
                                                 params.get('gradY1',0)+h,
                                                 params.get('gradX2',0)+w,
                                                 params.get('gradY2',0)+h)
                c = params.get('gradC1',(0,0,0))
                gradient.setColorAt(0,Qt.QColor(c[0],c[1],c[2]))
                c = params.get('gradC2',(255,255,255))
                gradient.setColorAt(1,Qt.QColor(c[0],c[1],c[2])) 
                brush = Qt.QBrush(gradient)
            else:
                brush.setStyle(fillStyle)
            
            bg = params.get('background',(255,255,255))
            brush.setColor(Qt.QColor(bg[0],bg[1],bg[2]))
            item.setBrush(brush)
        except:
            pass

        name = params.get('name')

        if name and isinstance(item, tau.widget.TauGraphicsItem):
            #print 'JDrawTauGraphicsFactory.set_common_params(%s): calling setModel(%s)'%(item,name)
            item.setModel(name)

if __name__ == "__main__":
    import sys
    app = Qt.QApplication([])
    gui = tau.widget.TauJDrawSynopticsView()
    gui.setModel(sys.argv[1])
    gui.show()
    sys.exit(app.exec_())