
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
taubase.py: 
"""

import traceback
import inspect
import sys

import PyTango

from PyQt4 import Qt

import utils
import tau
import tau.core
from tau.core.utils import EventFilters
from configuration import TauBaseConfigurableClass
import actionfactory
import traceback

DefaultNoneValue = "-----"

TTANGO_TO_TQT = {
    str                        : 'QString',
    int                        : 'int',
    long                       : 'long',
    float                      : 'float',
    bool                       : 'bool',
    list                       : 'QList',
    tuple                      : 'QList',
    dict                       : 'QMap',
    PyTango.ArgType.DevDouble          : 'float',
    PyTango.ArgType.DevFloat           : 'float',
    PyTango.ArgType.DevLong            : 'int',
    PyTango.ArgType.DevShort           : 'int',
    PyTango.ArgType.DevBoolean         : 'bool',
    PyTango.ArgType.DevUChar           : 'int',
    PyTango.ArgType.DevState           : 'int',
    PyTango.ArgType.DevString          : 'QString',
    PyTango.ArgType.DevULong           : 'int',
    PyTango.ArgType.DevLong64          : 'long',
    PyTango.ArgType.DevULong64         : 'long',
    PyTango.ArgType.DevUShort          : 'int',
    PyTango.ArgType.DevVarBooleanArray : 'QList',
    PyTango.ArgType.DevVarCharArray    : 'QList',
    PyTango.ArgType.DevVarDoubleArray  : 'QList',
    PyTango.ArgType.DevVarFloatArray   : 'QList',
    PyTango.ArgType.DevVarLongArray    : 'QList',
    PyTango.ArgType.DevVarLong64Array  : 'QList',
    PyTango.ArgType.DevVarShortArray   : 'QList',
    PyTango.ArgType.DevVarStringArray  : 'QList',
    PyTango.ArgType.DevVarULongArray   : 'QList',
    PyTango.ArgType.DevVarUShortArray  : 'QList',
}   

class TauBaseComponent(tau.core.TauListener, TauBaseConfigurableClass):
    """A generic Tau component.
    
    **Important**: Any class which inherits from TauBaseComponent is expected to
    also inherit from QObject (or from a QObject derived class) """
    
    def __init__(self, name, parent=None):
        """Initialization of TauBaseComponent"""
        utils.initTauQtLogger()
        self.modelObj = None
        self.modelName = ''
        self.noneValue = DefaultNoneValue
        self.call__init__(tau.core.TauListener, name, parent)
        
        TauBaseConfigurableClass.__init__(self)
        
        self.tauMenu = None
        self.tauMenuData = ''
        
        # attributes storing property values
        self._localModelName = ''
        self._useParentModel = False
        self._showQuality = True
        self._showText = True
        self._attached = False
        self._dangerMessage = None
        self._isDangerous = False
        self._forceDangerousOperations = False
        self._eventFilters = []
        self._isPaused = False
        self._operations = []
        self._modifiableByUser = False

        if parent != None and hasattr(parent, "_exception_listener"):
            self._exception_listener = parent._exception_listener
        else:
            self._exception_listener = set([tau.core.TauExceptionListener()])

    def deleteLater(self):
        '''Reimplements the Qt.QObject deleteLater method to ensure that the
        this object stops listening its model.'''
        self.setUseParentModel(False)
        self.resetModel()
        Qt.QObject.deleteLater(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Utility methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    
    def getTauManager(self):
        """Returns the the tau manager singleton. This is just a helper method.
        It is the equivalent of doing::
            
            import tau
            manager = tau.Manager()
        
        :return: (tau.core.TauManager) the TauManager
        """
        return tau.Manager()
    
    def getTauFactory(self, scheme=None):
        """Returns the the tau factory singleton for the given scheme.
        This is just a helper method. It is the equivalent of doing::

            import tau
            factory = tau.Factory(scheme)
        
        :param scheme: (str or None) the scheme. None defaults to 'tango'.
        
        :return: (tau.core.TauFactory) the TauFactory
        """
        return tau.Factory(scheme)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Popup menu behavior
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def contextMenuEvent(self, event):
        """Handle the popup menu event
        
        :param event: the popup menu event
        """
        if not self.tauMenu is None:
            self.tauMenu.exec_(event.globalPos())
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to be implemented in subclass implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    def updateStyle(self):
        """Method called when the component detects an event that triggers a 
        change in the style.
        Default implementation doesn't do anything. Overwrite when necessary
        """
        pass

    def getParentTauComponent(self):
        """ Returns a parent Tau component or None if no parent 
        :class:`tau.core.TauBaseComponent` is found.
        
        :raises: RuntimeError
        """
        raise RuntimeError("Not allowed to call TauBaseComponent::getParentTauComponent()")
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Event handling chain
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def eventReceived(self, evt_src, evt_type, evt_value):
        """The basic implementation of the event handling chain is as follows:
               - eventReceived just calls :meth:`fireEvent` which emits a "tauEvent"
                 PyQt signal that is connected (by :meth:`preAttach`) to the
                 :meth:`filterEvent` method.
               - After filtering, :meth:`handleEvent` is invoked with the resulting
                 filtered event
        
        .. note::
            in the earlier steps of the chain (i.e., in :meth:`eventReceived`/:meth:`fireEvent`),
            the code is executed in a Python thread, while from eventFilter
            ahead, the code is executed in a Qt thread.
            When writing widgets, one should normally work on the Qt thread
            (i.e. reimplementing :meth:`handleEvent`)
            
        :param evt_src: (object) object that triggered the event
        :param evt_type: (tau.core.TauEventType) type of event
        :param evt_value: (object) event value
        """
        self.fireEvent(evt_src, evt_type, evt_value)
        
    def fireEvent(self, evt_src = None, evt_type = None, evt_value = None):
        """Emits a "tauEvent" signal.
        It is unlikely that you may need to reimplement this method in subclasses.
        Consider reimplementing :meth:`eventReceived` or :meth:`handleEvent` 
        instead depending on whether you need to execute code in the python 
        or Qt threads, respectively

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (tau.core.TauEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        try: self.emit(Qt.SIGNAL('tauEvent'),  evt_src, evt_type, evt_value)
        except: pass #self.error('%s.fireEvent(...) failed!'%type(self))
        
    @Qt.pyqtSignature("tauEvent")
    def filterEvent(self, evt_src=-1, evt_type=-1, evt_value=-1):
        """The event is processed by each and all filters in strict order
        unless one of them returns None (in which case the event is discarded)
        
        :param evt_src: (object) object that triggered the event
        :param evt_type: (tau.core.TauEventType) type of event
        :param evt_value: (object) event value
        """
        
        r = evt_src, evt_type, evt_value
        
        if r == (-1,-1,-1):
            # @todo In an ideal world the signature of this method should be
            # (evt_src, evt_type, evt_value). However there's a bug in PyQt:
            # If a signal is disconnected between the moment it is emitted and
            # the moment the slot is called, then the slot is called without
            # parameters (!?). We added this default values to detect if
            # this is the case without printing an error message each time.
            # If this gets fixed, we should remove this line.
            return
        
        for f in self._eventFilters:
            r = f(*r)
            if r is None: return
        self.handleEvent(*r)

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Event handling. Default implementation does nothing.
        Reimplement as necessary
        
        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (tau.core.TauEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        pass
        
    def setEventFilters(self, filters = None):
        """sets the tau event filters list.
        The filters are run in order, using each output to feed the next filter.
        A filter must be a function that accepts 3 arguments ``(evt_src, evt_type, evt_value)``
        If the event is to be ignored, the filter must return None.
        If the event is  not to be ignored, filter must return a
        ``(evt_src, evt_type, evt_value)`` tuple which may (or not) differ from the input.
        
        For a library of common filters, see tau/utils/eventfilters.py
        
        :param filters: (sequence) a sequence of filters
        
        *Note*: If you are setting a filter that applies a transformation on
        the parameters, you may want to generate a fake event to force the last
        value to be filtered as well. This can be done as in this example::
        
            TauBaseComponent.fireEvent( TauBaseComponent.getModelObj(),
                                        tau.core.TauEventType.Periodic,
                                        TauBaseComponent.getModelObj().getValueObj())
        
        See also: insertEventFilter
        """
        if filters is None: filters = []
        self._eventFilters = list(filters)
        
    def getEventFilters(self):
        """Returns the list of event filters for this widget
        
        :return: (sequence<callable>) the event filters
        """
        return self._eventFilters
    
    def insertEventFilter(self, filter, index=-1):
        """insert a filter in a given position
        
        :param filter: (callable(evt_src, evt_type, evt_value)) a filter
        :param index: (int) index to place the filter (default = -1 meaning place at the end)
        
        See also: setEventFilters
        """
        self._eventFilters.insert(index, filter)
    
    def setPaused(self, paused = True):
        """Toggles the pause mode.
        
        :param paused: (bool) whether or not to pause (default = True)
        """
        if paused == self._isPaused: return #nothing to do
        if paused:  #pausing
            self.insertEventFilter(EventFilters.IGNORE_ALL, 0)
            self.debug('paused')
        else:   #unpausing
            try:
                self._eventFilters.remove(EventFilters.IGNORE_ALL)
                self.debug('Unpaused')  
            except ValueError:
                self.warning('Unpause failed')
        self._isPaused = paused
        
    def isPaused(self):
        """Return the current pause state
        
        :return: (bool) wheater or not the widget is paused
        """
        return self._isPaused
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Model class methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModelClass(self):
        """Return the class object for the widget.
        Default behavior is to do a 'best effort' to determine which model 
        type corresponds to the current model name.
        Overwrite as necessary.
        
        :return: (class TauModel or None) The class object corresponding to the type 
                 of Tau model this widget handles or None if no valid class is found.
        """
        return self.findModelClass()
    
    def findModelClass(self):
        """Do a "best effort" to determine which model type corresponds to the 
        given model name.
        
        :return: (class TauModel or None) The class object corresponding to the type
                 of Tau model this widget handles or None if no valid class is found.
        """
        if self.getUseParentModel():
            return self._findRelativeModelClass(self.getModel())
        else:
            return self._findAbsoluteModelClass(self.getModel())
        
    def _findAbsoluteModelClass(self, absolute_name):
        obj = tau.Manager().findObject(absolute_name)
        if obj is None:
            return None
        else:
            return obj.__class__
        
    def _findRelativeModelClass(self, relative_name):
        parent_widget = self.getParentTauComponent()
        if parent_widget is None:
            return None
        
        parent_obj = parent_widget.getModelObj()
        if parent_obj is None:
            return None
        
        if relative_name is None or len(relative_name) == 0:
            return parent_widget.getModelClass()
        
        obj = parent_obj.getChildObj(relative_name)
        if obj is None:
            return None
        else:
            return obj.__class__

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Model related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def setModelName(self, modelName, parent=None):
        """This method will detach from the previous tau model (if any), it 
        will set the new model to the given modelName and it will attach 
        this component to the new tau model.
        
        :param modelName: (str) the new tau model name (according to the tau convention)
        :param parent: (TauBaseComponent) the parent or None (default) if this 
                       component does not have a parent Tau component
        """
        modelName = str(modelName)
        if parent:
            modelClass = self.getModelClass()
            if not modelClass is None:
                parent_model = self.getParentModelObj()
                modelName = modelClass.buildModelName(parent_model, modelName)
        self._detach()
        self.modelName = modelName
        self._attach()
    
    def getModelName(self):
        """Returns the current model name.
            
        :return: (str) the model name
        """
        return self.modelName

    def getParentModelName(self):
        """Returns the parent model name or an empty string if the component
        has no parent
            
        :return: (str) the parent model name
        """
        try:
            p = self.getParentTauComponent()
            if p is None:
                return ''
        except:
            return ''
        return p.getModelName()

    def getParentModelObj(self):
        """Returns the parent model object or None if the component has no 
        parent or if the parent's model is None
            
        :return: (tau.core.TauModel or None) the parent tau model object
        """
        try:
            p = self.getParentTauComponent()
            if p is None:
                return None
        except:
            return None
        return p.getModelObj()
    
    def getModelObj(self):
        """Returns the tau model obj associated with this component or None if 
        no tau model is associated.
            
        :return: (tau.core.TauModel or None) the tau model object
        """
        return self.modelObj
    
    def getModelValueObj(self,cache=True):
        """Returns the tango obj value associated with this component or None 
        if no tau model is associated.
            
        :param cache: (bool) if set to True (default) use the cache value. If set to 
                      False will force a connection to the server.
            
        :return: (PyTango.DeviceAttribute) the tango value object.
        """
        if self.modelObj is None:
            return None
        return self.modelObj.getValueObj(cache=cache)
        
    def getFormatedToolTip(self,cache=True):
        """Returns a string with contents to be displayed in a tooltip.
            
        :param cache: (bool) if set to True (default) use the cache value. If set to 
                      False will force a connection to the server.
            
        :return: (str) a tooltip
        """
        if self.modelObj is None:
            return Qt.QString.fromLatin1(self.getNoneValue())
        obj = self.modelObj.getDisplayDescrObj()
        return Qt.QString.fromLatin1(self.toolTipObjToStr(obj))
    
    def toolTipObjToStr(self, toolTipObj):
        """Converts a python dict to a tooltip string.
        
        :param toolTipObj: (dict) a python object
        
        :return: (str) a tooltip
        """
        if toolTipObj is None:
            return self.getNoneValue()
        ret = '<TABLE width="500" border="0" cellpadding="1" cellspacing="0">'
        
        for id,value in toolTipObj:
            ret += '<TR><TD WIDTH="80" ALIGN="RIGHT" VALIGN="MIDDLE"><B>%s:</B></TD><TD>%s</TD></TR>' % (id.capitalize(), value)
        ret += '</TABLE>'
        return ret
    
    def displayValue(self, v):
        """Returns a string representation of the given value
        
        :param v: (object) the value to be translated to string
        
        :return: (str) a string representing the given value
        """
        if self.modelObj is None:
            return str(v)
        ret = self.modelObj.displayValue(v)
        if ret is None: ret = self.getNoneValue()
        return ret
        
    def getDisplayValue(self, cache=True):
        """Returns a string representation of the model value associated with
        this component.
            
        :param cache: (bool) if set to True (default) use the cache value. If set to 
                      False will force a connection to the server.

        :return: (str) a string representation of the model value.
        """        
        if self.modelObj is None:
            return self.getNoneValue()
        
        ret = self.modelObj.getDisplayValue(cache)
        if ret is None:
            return self.getNoneValue()
        return ret

    def setNoneValue(self, v):
        """Sets the new string representation when no model or no model value exists.
            
        :param v: (str) the string representation for an invalid value
        """
        self.noneValue = v
        
    def getNoneValue(self):
        """Returns the current string representation when no valid model or model value exists.
            
        :return: (str) a string representation for an invalid value
        """
        return self.noneValue
    
    def isChangeable(self):
        """Tells if this component value can be changed by the user. Default implementation 
        will return True if and only if:
        
            - this component is attached to a valid tau model and
            - the tau model is writable and
            - this component is not read-only
            
        :return: (bool) True if this component value can be changed by the user or False otherwise
        """
        res = False
        if not self.modelObj is None:
            res = self.modelObj.isWritable()
        res = res and not self.isReadOnly()
        return res
    
    def isReadOnly(self):
        """Determines if this component is read-only or not in the sense that the
        user can interact with it. Default implementation returns True.
        
        Override when necessary.
        
        :return: (bool) whether or not this component is read-only
        """
    
    def isAttached(self):
        """Determines if this component is attached to the tau model.
           
        :return: (bool) True if the component is attached or False otherwise.
        """
        return self._attached
    
    def preAttach(self):
        """Called inside self.attach() before actual attach is performed.
        Default implementation just emits a signal.
           
        Override when necessary.
        """
        try: Qt.QObject.connect(self, Qt.SIGNAL('tauEvent'), self.filterEvent)
        except: pass #self.error("In %s.preAttach() ... failed!" % str(type(self)))

    def postAttach(self):
        """Called inside self.attach() after actual attach is performed.
        Default implementation does not do anything.
           
        Override when necessary.
        """
        pass
    
    def preDetach(self):
        """Called inside self.detach() before actual deattach is performed.
        Default implementation just disconnects a signal.
           
        Override when necessary.
        """
        try: Qt.QObject.disconnect(self, Qt.SIGNAL('tauEvent'), self.filterEvent)
        except: pass #self.error("In %s.preDetach() ... failed!" % str(type(self)))

        
    def postDetach(self):
        """Called inside self.detach() after actual deattach is performed.
        Default implementation does not do anything.
        
        Override when necessary.
        """
        pass
    
    def _attach(self):
        """Attaches the component to the tau model.
        In general it should not be necessary to overwrite this method in a subclass.
        
        :return: (bool) True if success in attachment or False otherwise.
        """
        if self.isAttached():
            return self._attached
        
        self.preAttach()
        
        cls = self.getModelClass()
        
        if cls is None:
            self._attached = False
            #self.trace("Failed to attach: Model class not found")
        else:
            try:
                self.modelObj = tau.Manager().getObject(cls, self.modelName)
                if not self.modelObj is None:
                    self.modelObj.addListener(self)
                    self._attached = True
                    self.changeLogName(self.log_name + "." + self.modelName)
            except Exception,e:
                self.modelObj = None
                self._attached = False
                self.debug("Exception occured while trying to attach '%s'" % self.modelName)
                
        self.postAttach()
        return self._attached
    
    def _detach(self):
        """Detaches the component from the tau model"""
        self.preDetach()
        
        if self.isAttached():
            m = self.getModelObj()
            if not m is None:
                m.removeListener(self)
            
            pos = self.log_name.find('.')
            if pos >= 0:
                new_log_name = self.log_name[:self.log_name.rfind('.')]
                self.changeLogName(new_log_name)
            self.modelObj = None
            self._attached = False
            self.fireEvent(m, tau.core.TauEventType.Change, None)

        self.postDetach()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def applyPendingOperations(self, ops=None):
        """Applies operations without caring about danger messages.
        Use :meth:`TauBaseWidget.safeApplyOperation` if you want to warn the
        user before applying
        
        :param ops: (sequence<tau.core.TauOperation> or None) list of operations to apply. 
                    If None is given (default) the component fetches the pending operations
        """
        self.debug("Apply changes")
        if ops is None: ops = self.getPendingOperations()
        self.getTauManager().applyPendingOperations(ops)
        
    def hasPendingOperations(self):
        """Returns if the component has pending operations
        
        :return: (bool) True if there are pending operations or False otherwise
        """
        return len(self.getPendingOperations()) > 0
   
    def getPendingOperations(self):
        """Returns the sequence of pending operations
        
        :return:  (sequence<tau.core.TauOperation>) a list of pending operations
        """
        return self._operations
    
    def resetPendingOperations(self):
        """Clears the list of pending operations"""
        self._operations = []
    
    def setDangerMessage(self, dangerMessage=None):
        """Sets the danger message when applying an operation. If dangerMessage is None, 
        the apply operation is considered safe
        
        :param dangerMessage: (str or None) the danger message. If None is given (default)
                              the apply operation is considered safe
        """
        self._dangerMessage = dangerMessage
        self._isDangerous = dangerMessage is not None
    
    def getDangerMessage(self):
        """Returns the current apply danger message or None if the apply operation is safe
        
        :return: (str or None) the apply danger message
        """
        return self._dangerMessage
        
    def resetDangerMessage(self):
        """Clears the danger message. After this method is executed the apply operation
        for this component will be considered safe."""
        self.setDangerMessage(None)
        
    def isDangerous(self):
        """Returns if the apply operation for this component is dangerous
        
        :return: (bool) wheter or not the apply operation for this component is dangerous
        """
        return self._isDangerous
        
    def setForceDangerousOperations(self, yesno):
        """Forces/clears the dangerous operations
        
        :param yesno: (bool) force or not the dangerous operations"""
        self._forceDangerousOperations = yesno
        
    def getForceDangerousOperations(self):
        """Returns if apply dangerous operations is forced
        
        :return: (bool) wheter or not apply dangerous operations is forced
        """
        return self._forceDangerousOperations
        
    def resetForceDangerousOperations(self):
        """Clears forcing apply dangerous operations"""
        self.setForceDangerousOperations(False)
        
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Standard Qt properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getModel(self):
        """Returns the model name for this component.
            
        :return: (str) the model name.
        """
        return self._localModelName
    
    @Qt.pyqtSignature("setModel(QString)")
    def setModel(self,model):
        """Sets/unsets the model name for this component
        
        :param model: (str) the new model name"""
        self.setModelCheck(model)
        self.updateStyle()

    def setModelCheck(self,model,check=True):
        """Sets the component tau model. Setting the check argument to True
        (default) will check if the current model is equal to the given argument.
        If so then nothing is done. This should be the standard way to call this
        method since it will avoid recursion.
        
        :param model: (str) the new model name
        :param check: (bool) whether or not to check against the actual model name""" 
        model = str(model)
        if check == True and model == self._localModelName:
            return
        
        self._localModelName = model
        
#        # if in offline mode don't bother trying to register
#        opMode = tau.core.TauManager().getOperationMode()
#        if opMode == tau.core.OperationMode.OFFLINE:
#            return

        parent_widget = None
        try:
            # if this widget has a buddy, check to see if it is a valid TauWidget
            buddy_func = getattr(self,'buddy')
            buddy_widget = buddy_func()
            if buddy_widget and isinstance(buddy_widget,taubase.TauBaseComponent):
                parent_widget = buddy_widget
            elif self.getUseParentModel():
                parent_widget = self.getParentTauComponent()
        except:
            if self.getUseParentModel():
                parent_widget = self.getParentTauComponent()
        self.setModelName(model, parent_widget)
        #self.fireEvent(self.getModelObj(), tau.core.TauEventType.Change, self.getModelValueObj())
    
    def resetModel(self):
        """Sets the model name to the empty string"""
        self.setModel('')
    
    def getUseParentModel(self):
        """Returns whether this component is using the parent model
        
        :return: (bool) True if using parent model or False otherwise
        """
        return getattr(self, '_useParentModel', False)
    
    @Qt.pyqtSignature("setUseParentModel(bool)")
    def setUseParentModel(self, yesno):
        """Sets/unsets using the parent model
        
        :param yesno: (bool) whether or not to use parent model
        """
        if yesno == self._useParentModel:
            return
        self._useParentModel = yesno
        # force a recalculation of the model
        self.setModelCheck(self.getModel(),False)

    def resetUseParentModel(self):
        """Resets the usage of parent model to False"""
        self.setUseParentModel(False)
        self.updateStyle()

    @Qt.pyqtSignature("setShowQuality(bool)")
    def setShowQuality(self,showQuality):
        """Sets/unsets the show quality property
        
        :param showQuality: (bool) whether or not to show the quality
        """
        if showQuality == self._showQuality:
            return
        self._showQuality = showQuality
        self.updateStyle()

    def getShowQuality(self):
        """Returns if showing the quality as a background color
        
        :return: (bool) True if showing the quality or False otherwise
        """
        return self._showQuality

    def resetShowQuality(self):
        """Resets the show quality to False"""
        self.setShowQuality(False)

    @Qt.pyqtSignature("setShowText(bool)")
    def setShowText(self, showText):
        """Sets/unsets showing the display value of the model
        
        :param showText: (bool) whether or not to show the display value
        """
        if showText == self._showText:
            return
        self._showText = showText
        self.fireEvent(self.getModelObj(),tau.core.TauEventType.Change, self.getModelValueObj())
        self.updateStyle()

    def getShowText(self):
        """Returns if showing the display value
        
        :return: (bool) True if showing the display value or False otherwise
        """
        return self._showText

    def resetShowText(self):
        """Resets the showing of the display value to True"""
        self.setShowText(True)
        
    @Qt.pyqtSignature("setTauPopupMenu(QString)")
    def setTauPopupMenu(self, menuData):
        """Sets/unsets the tau popup menu
        
        :param menuData: (str) an xml representing the popup menu"""
        self.tauMenuData = str(menuData)
        factory = actionfactory.ActionFactory()
        self.tauMenu = factory.getNewMenu(self, self.tauMenuData)
    
    def getTauPopupMenu(self):
        """Returns an xml string representing the current tau popup menu
        
        :return: (str) an xml string representing the current tau popup menu
        """
        return self.tauMenuData
  
    def resetTauPopupMenu(self):
        """Resets the tau popup menu to empty"""
        self.tauMenuData = ''
        
    def isModifiableByUser(self):
        '''whether the user can change the contents of the form
        
        :return: (bool) True is the user is allowed to modify the look&feel'''
        return self._modifiableByUser
    
    def setModifiableByUser(self, modifiable):
        '''
        sets whether the user is allowed to modify the look&feel
        
        :param modifiable: (bool)
        '''
        self._modifiableByUser = modifiable
            
        
    def resetModifiableByUser(self):
        '''Equivalent to setModifiableByUser(True)'''
        self.setModifiableByUser(False)
        
    
class TauBaseWidget(TauBaseComponent):
    """ The base class for all Qt Tau widgets.
    
    **Important**: Any class which inherits from TauBaseWidget is expected to
    also inherit from QWidget (or from a QWidget derived class) """
    
    ModelChangedSignal = 'modelChanged(const QString &)'
    
    def __init__(self, name, parent=None, designMode = False):
        self.call__init__(TauBaseComponent, name, parent)
        self._setText = self._findSetTextMethod()
        self._disconnect_on_hide = False
    
    ## It makes the GUI to hang... If this needs implementing, we should
    ## reimplement it using the Qt parent class, not QWidget...
    #def destroy(self):
    #    '''Reimplements the Qt.QWidget destroy method to ensure that this object
    #    stops listening its model.'''
    #    self.setUseParentModel(False)
    #    self.resetModel()
    #    Qt.QWidget.destroy(self)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Helper methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def getQtClass(self, bases=None):
        """Returns the parent Qt class for this widget
        
        :param bases: (sequence<class> or None) the list of class objects. If None
                      is given (default) it uses the object base classes from __bases__
        
        :return: (QWidget class) the QWidget class object
        """
        bases = bases or self.__class__.__bases__
        for klass in bases:
            is_taubasewidget = issubclass(klass, TauBaseWidget)
            if issubclass(klass, Qt.QWidget):
                if is_taubasewidget:
                    return self.getQtClass(klass.__bases__)
                return klass
            elif is_taubasewidget:
                return self.getQtClass(klass.__bases__)
        return None
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Qt properties from TauBaseComponent that need to be overwritten
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    @Qt.pyqtSignature("setUseParentModel(bool)")
    def setUseParentModel(self, yesno):
        """Sets/unsets using the parent model
        
        :param yesno: (bool) whether or not to use parent model
        """
        is_same = yesno == self._useParentModel
        if not is_same:
            self._updateUseParentModel(yesno)
        TauBaseComponent.setUseParentModel(self, yesno)

    def _updateUseParentModel(self, yesno):
        parent_widget = self.getParentTauComponent()
        if parent_widget:
            if yesno:
                self.connect(parent_widget, 
                             Qt.SIGNAL(TauBaseWidget.ModelChangedSignal), 
                             self.parentModelChanged)
            else:
                self.disconnect(parent_widget,
                                Qt.SIGNAL(TauBaseWidget.ModelChangedSignal), 
                                self.parentModelChanged)
        
    def setModelCheck(self,model,check=True):
        """Sets the component tau model. Setting the check argument to True
        (default) will check if the current model is equal to the given argument.
        If so then nothing is done. This should be the standard way to call this
        method since it will avoid recursion.
        
        :param model: (str) the new model name
        :param check: (bool) whether or not to check against the actual model name
        """ 
        model = str(model)
        send_signal = check == False or model != self._localModelName
        TauBaseComponent.setModelCheck(self, model, check)
        
        if send_signal:
            # emit a signal informing the child widgets that the model has changed
            self.emit(Qt.SIGNAL(TauBaseWidget.ModelChangedSignal), model)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Default Qt signal handlers. Overwrite them as necessary
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def changeEvent(self, evt):
        """overwrites QWidget.changeEvent(self, evt) to handle the ParentChangeEvent 
        in case this widget is using the parent model. Always calls the QWidget.changeEvent 
        in order not to loose events
        """
        if self.getUseParentModel():
            evt_type = evt.type()
            if evt_type == Qt.QEvent.ParentChange:
                # disconnect from old parent
                if self._parentTauComponent:
                    self.disconnect(self._parentTauComponent,
                                    Qt.SIGNAL(TauBaseWidget.ModelChangedSignal),
                                    self.parentModelChanged)
                self._updateUseParentModel(True)
                self.setModelCheck(self.getModel(), False)
        self.getQtClass().changeEvent(self, evt)
    
    @Qt.pyqtSignature("parentModelChanged(const QString &)")
    def parentModelChanged(self, parentmodel_name):
        """Invoked when the parent model changes
        
        :param parentmodel_name: (str) the new name of the parent model
        """
        self.debug("Parent model changed to '%s'" % parentmodel_name)
        parentmodel_name = str(parentmodel_name)
        if self.getUseParentModel():
            # force an update of the interpretation of the model property
            self.setModelCheck(self.getModel(),False)
        else:
            self.debug("received event from parent although not using parent model")

    def handleEvent(self, evt_src, evt_type, evt_value):
        """very basic and generalistic handling of events.
        
        Override when necessary.
        
        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (tau.core.TauEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        #Update the text shown by the widget
        if self._setText:
            text = ''
            if self.getShowText():
                if isinstance(evt_src, tau.core.TauAttribute):
                    if evt_type in (tau.core.TauEventType.Change, tau.core.TauEventType.Periodic):
                        text = self.displayValue(evt_value.value)
                    elif evt_type == tau.core.TauEventType.Error:
                        text = self.getNoneValue()
                    elif evt_type == tau.core.TauEventType.Config:
                        text = self.getDisplayValue() 
                else:
                    text = self.getDisplayValue()
            self._setText(text)
        
        #update tooltip
        self.setToolTip(self.getFormatedToolTip())
        
        #TODO: update whatsThis
        
        #update appearance
        self.updateStyle()
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory overwrite from TauBaseComponent
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def updateStyle(self):
        """Updates the widget style. Default implementation just calls QWidget.update()
        
        Override when necessary.
        """
        self.update()

    def getParentTauComponent(self):
        """Returns the first tau component in the widget hierarchy or None if no
        tau component is found
        
        :return: (TauBaseComponent or None) the parent tau base component
        """
        p = self.parentWidget()
        while p and not isinstance(p,TauBaseWidget):
            p = p.parentWidget()
        self._parentTauComponent = p
        return p

    def setDisconnectOnHide(self, disconnect):
        """Sets/unsets disconnection on hide event
        
        :param disconnect: (bool) whether or not to disconnect on hide event
        """
        if not self.visible() and disconnect == False:
            self.info("Ignoring setDisconnectOnHide to False because widget is not visible")
            return
        self._disconnect_on_hide = disconnect
    
    def hideEvent(self, event):
        """Override of the QWidget.hideEvent()
        """
        if self._disconnect_on_hide:
            try:
                if self.getModelName():
                    self._detach()
                event.accept()
            except Exception,e:
                self.warning("Exception received while trying to hide")
                self.traceback()

    def showEvent(self, event):
        """Override of the QWidget.showEvent()"""
        if self._disconnect_on_hide:
            try:
                if self.getModelName():
                    self._attach()
                event.accept()
            except Exception,e:
                self.warning("Exception received while trying to show")
                self.traceback()

    def closeEvent(self, event):
        """Override of the QWidget.closeEvent()"""
        try:
            self._detach()
            event.accept()
        except Exception,e:
            self.warning("Exception received while trying to close")
            self.traceback()
            
    def handleException(self, e):
        for h in self._exception_listener:
            h.exceptionReceived(e)

    def _findSetTextMethod(self):
        """Determine if this widget is able to display the text value of the tau
        model. It searches through the possible Qt methods to display text.
           
        :return: (callable) a python method or None if no suitable method is found.
        """
        setMethod = None
        try:
            setMethod = getattr(self,'setText')
        except AttributeError:
            try:
                setMethod = getattr(self,'setTitle')
            except AttributeError:
                try:
                    setMethod = getattr(self,'display')
                except AttributeError:
                    # it seems the widget has no way to update a value
                    pass
                
        return setMethod
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods: default implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
         
    def updatePendingOpsStyle(self):
        """This method should be reimplemented by derived classes that want to
        change their appearance depending whether there are pending operations or not"""
        pass
        
    def emitValueChanged(self, *args):
        """Connect the specific XXXXChanged signals from derived classes to this
        method in order to have a unified signal which can be used by Tau Widgets"""
        self.emit(Qt.SIGNAL('valueChanged'))
        self.updatePendingOpsStyle() #by default, update its own style
    
    def safeApplyOperations(self, ops = None):
        """Applies the given operations (or the pending operations if None passed)
        
        :param ops: (sequence<tau.core.TauOperation> or None) list of operations to apply. 
                    If None is given (default) the component fetches the pending operations
        """
        
        if ops is None: ops = self.getPendingOperations()
        
        #Check if we need to take care of dangerous operations
        if self.getForceDangerousOperations(): dangerMsgs = []
        else: dangerMsgs = [op.getDangerMessage() for op in ops if op.getDangerMessage() is not None]
        #warn the user if need be
        if len(dangerMsgs)==1:
            result = Qt.QMessageBox.warning(self, "Potentially dangerous action",
                                                "%s\nProceed?"%dangerMsgs[0],
                                                Qt.QMessageBox.Ok|Qt.QMessageBox.Cancel,
                                                Qt.QMessageBox.Ok)
            if result != Qt.QMessageBox.Ok:
                return
                
        elif len(dangerMsgs)>1:
            warningDlg = Qt.QMessageBox(Qt.QMessageBox.Warning, " %d potentially dangerous actions"%len(dangerMsgs),
                                     "You are about to apply %d actions that may be potentially dangerous. Proceed?"%len(dangerMsgs),
                                     Qt.QMessageBox.Ok|Qt.QMessageBox.Cancel,
                                     self)
            details = "\n".join(dangerMsgs)
            warningDlg.setDetailedText(details)
            result = warningDlg.exec_()
            if result != Qt.QMessageBox.Ok:
                return
        self.applyPendingOperations(ops)
        
