
#!/usr/bin/env python

#############################################################################
##
## This file is part of Tau, a Tango User Interface Library
##
## http://www.tango-controls.org/static/tau/latest/doc/html/index.html
##
## (copyleft) CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## This is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This software is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
###########################################################################

"""
tauvalue.py: 
"""


from PyQt4 import Qt
import PyTango
import tau.core

from tau.widget import TauWidgetFactory, TauBaseWidget
from tau.widget import TauValueLabel, TauConfigLabel, TauValueSpinBox, TauValueLineEdit, TauBoolLed
from tau.widget import TauValueCheckBox, TauStateLed, TauLauncherButton
from tau.widget import TauFrame, TauWheelEdit
from valuestable import TauValuesTable
from tau.widget.qwt import TauPlot, TauArrayEditor
from tau.widget.resources import qrc_extra_icons_designer

class defaultTauValueCheckBox(TauValueCheckBox):
    def __init__(self,*args):
        TauValueCheckBox.__init__(self,*args)
        self.setShowText(False)
        
class defaultLabelWidget(TauConfigLabel):
    def __init__(self,*args):
        TauConfigLabel.__init__(self,*args)
        self.setAlignment(Qt.Qt.AlignRight)
        self.setSizePolicy(Qt.QSizePolicy.Preferred,Qt.QSizePolicy.Maximum)
        self.setStyleSheet('border-style: solid; border-width: 1px; border-color: transparent; border-radius: 4px;')
    def setModel(self, model):
        try: config = self.tauValueBuddy.getLabelConfig()
        except Exception: config = 'label'
        config = self.tauValueBuddy.getLabelConfig()
        TauConfigLabel.setModel(self, model + "?configuration=%s"%config)
    def sizeHint(self):
        return Qt.QSize(Qt.QLabel.sizeHint(self).width(), 18)
    def contextMenuEvent(self,event):   
        """ The label widget will be used for handling the actions of the whole TauValue
        
        see :meth:`QWidget.contextMenuEvent`"""
        menu = Qt.QMenu(self)  
        menu.addMenu(tau.widget.actions.ConfigurationMenu(self.tauValueBuddy)) #@todo: This should be done more Tau-ish 
        if self.tauValueBuddy.isModifiableByUser():
            cr_action = menu.addAction("Change Read Widget",self.tauValueBuddy.onChangeReadWidget)
            cw_action = menu.addAction("Change Write Widget",self.tauValueBuddy.onChangeWriteWidget)
            cw_action.setEnabled(not self.tauValueBuddy.isReadOnly()) #disable the action if the tauValue is readonly
        menu.exec_(event.globalPos())
        event.accept()
    def mousePressEvent(self, event):
        '''reimplemented to provide drag events'''
        Qt.QLabel.mousePressEvent(self, event)
        if event.button() == Qt.Qt.LeftButton:
            self.dragStartPosition = event.pos()
    def mouseMoveEvent(self, event):
        '''reimplemented to provide drag events'''
        if not event.buttons() & Qt.Qt.LeftButton:
            return
        mimeData = Qt.QMimeData()
        mimeData.setText(self.tauValueBuddy.getModelName())
        drag = Qt.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        dropAction = drag.start(Qt.Qt.CopyAction)
    

class defaultUnitsWidget(TauConfigLabel):
    def __init__(self,*args):
        TauConfigLabel.__init__(self,*args)
        self.setNoneValue('')
        self.setSizePolicy(Qt.QSizePolicy.Preferred,Qt.QSizePolicy.Maximum)
    def setModel(self, model):
        TauConfigLabel.setModel(self, model + "?configuration=unit")
    def sizeHint(self):
        #print "UNITSSIZEHINT:",Qt.QLabel.sizeHint(self).width(), self.minimumSizeHint().width(), Qt.QLabel.minimumSizeHint(self).width()
        return Qt.QSize(Qt.QLabel.sizeHint(self).width(), 24)
 
class TauPlotButton(TauLauncherButton):
    '''A button that launches a TauPlot'''
    def __init__(self, parent = None, designMode = False):
        TauLauncherButton.__init__(self, parent = parent, designMode = designMode, widget = TauPlot(), icon=':/designer/qwtplot.png', text = 'Show')
        
class TauArrayEditorButton(TauLauncherButton):
    '''A button that launches a TauArrayEditor'''
    def __init__(self, parent = None, designMode = False):
        TauLauncherButton.__init__(self, parent = parent, designMode = designMode, widget = TauArrayEditor(), icon=':/designer/arrayedit.png', text = 'Edit')
        
class TauValuesTableButton(TauLauncherButton):
    '''A button that launches a TauValuesTable'''
    def __init__(self, parent = None, designMode = False):
        TauLauncherButton.__init__(self, parent = parent, designMode = designMode, widget = TauValuesTable(), icon=':/designer/table.png', text = 'Show')

class TauStatusLabel(TauValueLabel):
    '''just a tauVaueLabel but showing the state as its background by default'''
    def __init__(self, parent = None, designMode = False, background = 'state'):
        TauValueLabel.__init__(self, parent = parent, designMode = designMode, background = background)
        

class TauValue(Qt.QWidget, TauBaseWidget):
    
    __pyqtSignals__ = ("modelChanged(const QString &)",)
    
    def __init__(self, parent = None, designMode = False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TauBaseWidget, name, designMode=designMode)
        
        self._designMode=designMode
        
        #This is a hack to show something usable when in designMode
        if self._designMode:
            layout = Qt.QHBoxLayout(self)
            dummy = TauConfigLabel()
            layout.addWidget(dummy)
            dummy.setUseParentModel(True)
            dummy.setModel("?configuration=attr_fullname") 
            dummy.setPrefixText("< TauValue: ")
            dummy.setSuffixText(" >")
        else:    
            self.setVisible(False)
            self.setFixedSize(1,1)

        self._parentLayout=None
        
        self._labelWidget = None
        self._readWidget = None
        self._writeWidget = None
        self._unitsWidget = None
        
        self.labelWidgetClassID = 'Auto'
        self.readWidgetClassID = 'Auto'
        self.writeWidgetClassID = 'Auto'
        self.unitsWidgetClassID = 'Auto'
        self.setPreferredRow(-1)
        self._row = None
        
        self._allowWrite = True
        self._minimumHeight = None
        self._labelConfig = 'label'
        self.setModifiableByUser(False)
        
        if parent is not None:
            self.setParent(parent)
    
    def labelWidget(self):
        return self._labelWidget
    
    def readWidget(self):
        return self._readWidget
    
    def writeWidget(self):
        return self._writeWidget
    
    def unitsWidget(self):
        return self._unitsWidget
    
    def setParent(self, parent):
   
        #make sure that the parent has a QGriLayout
        pl=parent.layout()
        if pl is None:
            pl = Qt.QGridLayout(parent) #creates AND sets the parent layout
        if not isinstance(pl, Qt.QGridLayout):
            raise ValueError('layout must be a QGridLayout (got %s)'%type(pl))
        self._parentLayout = pl
        
        if self._row is None:
            self._row = self.getPreferredRow()  #@TODO we should check that the Preferred row is empty in self._parentLayout
            if self._row < 0:
                self._row = self._parentLayout.rowCount()
        #print 'ROW:',self, self.getRow()
        
        #insert self into the 0-column
        self._parentLayout.addWidget(self, self._row, 0) #this widget is invisible (except in design mode)
        
        #Create/update the subwidgets (this also inserts them in the layout)
        if not self._designMode:  #in design mode, no subwidgets are created
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
        
        #do the base class stuff too    
        Qt.QWidget.setParent(self,parent)
        
    def getAllowWrite(self):
        return self._allowWrite
    
    @Qt.pyqtSignature("setAllowWrite(bool)")
    def setAllowWrite(self, mode):
        self._allowWrite = mode
    
    def resetAllowWrite(self):
        self._allowWrite = True
    
    def getPreferredRow(self):
        return self._preferredRow
    
    @Qt.pyqtSignature("setPreferredRow(int)")
    def setPreferredRow(self,row):
        self._preferredRow=row
        
    def resetPreferredRow(self):
        self.setPreferredRow(-1)
        
    def getRow(self):
        return self._row
    
    def setMinimumHeight(self, minimumHeight):
        self._minimumHeight = minimumHeight
        
    def minimumHeight(self):
        return self._minimumHeight
    
    def getDefaultLabelWidgetClass(self):
        return defaultLabelWidget
     
    def getDefaultReadWidgetClass(self, returnAll=False):
        '''
        Returns the default class (or classes) to use as read widget for the
        current model.
        
        :param returnAll: (bool) if True, the return value is a list of valid
                          classes instead of just one class
        
        :return: (class or list<class>) the default class  to use for the read
                 widget (or, if returnAll==True, a list of classes that can show
                 the attribute ). If a list is returned, it will be loosely
                 ordered by preference, being the first element always the
                 default one.
        '''
        modelobj = self.getModelObj()
        if modelobj is None: 
            if returnAll: return [TauValueLabel]
            else: return TauValueLabel

        config = modelobj.getConfig()
        #print "---------ATTRIBUTE OBJECT:----------\n",modelobj.read()
        if config.isScalar():
            configType = config.getType() 
            if  configType == PyTango.ArgType.DevBoolean:
                result = [TauBoolLed, TauValueLabel]
            elif configType == PyTango.ArgType.DevState:
                result = [TauStateLed, TauValueLabel]
            elif str(self.getModel()).lower().endswith('/status'):
                result = [TauStatusLabel, TauValueLabel]
            else:
                result = [TauValueLabel]
        elif config.isSpectrum():
            result = [TauPlotButton, TauValueLabel]
        elif config.isImage():
            result = [TauValuesTableButton, TauValueLabel]
        else:
            self.warning('Unsupported attribute type')
            result = None
        
        if returnAll: return result
        else: return result[0]
        
    def getDefaultWriteWidgetClass(self, returnAll=False):
        '''
        Returns the default class (or classes) to use as write widget for the
        current model.
        
        :param returnAll: (bool) if True, the return value is a list of valid
                          classes instead of just one class
        
        :return: (class or list<class>) the default class  to use for the write
                 widget (or, if returnAll==True, a list of classes that can show
                 the attribute ). If a list is returned, it will be loosely
                 ordered by preference, being the first element always the
                 default one.
        '''
        if self.isReadOnly(): 
            if returnAll: return []
            else: return None
        modelobj = self.getModelObj()
        if modelobj is None:
            if returnAll: return [TauValueLineEdit]
            else: return TauValueLineEdit
        config = modelobj.getConfig()
        if config.isScalar():
            configType = config.getType() 
            if configType == PyTango.ArgType.DevBoolean:
                result = [defaultTauValueCheckBox, TauValueLineEdit]
            #elif PyTango.is_numerical_type(configType ):
            #    result = TauWheelEdit
            else:
                result = [TauValueLineEdit, TauValueSpinBox, TauWheelEdit]
        elif config.isSpectrum():
            result = [TauArrayEditorButton, TauValueLineEdit]
        else:
            self.warning('Unsupported attribute type')
            result = [None]
            
        if returnAll: return result
        else: return result[0]
    
    def getDefaultUnitsWidgetClass(self):
        return defaultUnitsWidget
    
    def onChangeReadWidget(self):
        classnames = ['None', 'Auto']+[c.__name__ for c in self.getDefaultReadWidgetClass(returnAll=True)]
        cname, ok = Qt.QInputDialog.getItem(self, 'Change Read Widget', 'Choose a new read widget class', classnames, 1, True)
        if ok:
            self.setReadWidgetClass(str(cname))
            
    def onChangeWriteWidget(self):
        classnames = ['None', 'Auto']+[c.__name__ for c in self.getDefaultWriteWidgetClass(returnAll=True)]
        cname, ok = Qt.QInputDialog.getItem(self, 'Change Write Widget', 'Choose a new write widget class', classnames, 1, True)
        if ok:
            self.setWriteWidgetClass(str(cname))
        
    def _newSubwidget(self, oldWidget, newClass):
        '''eliminates oldWidget and returns a new one.
        If newClass is None, None is returned
        If newClass is the same as the olWidget class, nothing happens'''
        if oldWidget.__class__ == newClass: return oldWidget
        if oldWidget is not None:
            oldWidget.hide()
            oldWidget.setParent(None)
            oldWidget.destroy()
        if newClass is None: result = None
        else: result = newClass()
        return result

    def labelWidgetClassFactory(self, classID):
        if classID is None or classID is 'None': return None
        if isinstance(classID, type): return classID
        elif str(classID) == 'Auto': return self.getDefaultLabelWidgetClass()
        else: return TauWidgetFactory().getTauWidgetClass(classID)

    def readWidgetClassFactory(self, classID):
        if classID is None or classID is 'None': return None
        if isinstance(classID, type): return classID
        elif str(classID) == 'Auto': return self.getDefaultReadWidgetClass()
        else: return TauWidgetFactory().getTauWidgetClass(classID)
    
    def writeWidgetClassFactory(self, classID):
        if classID is None or classID is 'None': return None
        if isinstance(classID, type): return classID
        elif str(classID) == 'Auto': return self.getDefaultWriteWidgetClass()
        else: return TauWidgetFactory().getTauWidgetClass(classID)
        
    def unitsWidgetClassFactory(self, classID):
        if classID is None or classID is 'None': return None
        if isinstance(classID, type): return classID
        elif str(classID) == 'Auto': return self.getDefaultUnitsWidgetClass()
        else: return TauWidgetFactory().getTauWidgetClass(classID)
        
    def updateLabelWidget(self):
        #get the class for the widget and replace it if necessary
        klass = self.labelWidgetClassFactory(self.labelWidgetClassID)
        self._labelWidget = self._newSubwidget(self._labelWidget, klass)
        
        #take care of the layout
        self.addLabelWidgetToLayout() 
        
        if self._labelWidget is not None:
            #give the new widget a reference to its buddy TauValue object
            self._labelWidget.tauValueBuddy = self
            
            #tweak the new widget
            if self.minimumHeight() is not None:
                self._labelWidget.setMinimumHeight(self.minimumHeight())
            
            #set the model for the subwidget
            if hasattr(self._labelWidget,'setModel'):
                self._labelWidget.setModel(self.getModelName())
            
    def updateReadWidget(self):
        #get the class for the widget and replace it if necessary
        klass = self.readWidgetClassFactory(self.readWidgetClassID)
        self._readWidget = self._newSubwidget(self._readWidget, klass)
        
        #take care of the layout
        self.addReadWidgetToLayout() 
        
        if self._readWidget is not None:
            #give the new widget a reference to its buddy TauValue object
            self._readWidget.tauValueBuddy = self
            
            #tweak the new widget
            if self.minimumHeight() is not None:
                self._readWidget.setMinimumHeight(self.minimumHeight())
            
            #set the model for the subwidget
            if hasattr(self._readWidget,'setModel'):
                self._readWidget.setModel(self.getModelName())

    def updateWriteWidget(self):
        #get the class for the widget and replace it if necessary
        klass = self.writeWidgetClassFactory(self.writeWidgetClassID)
        self._writeWidget = self._newSubwidget(self._writeWidget, klass)
        
        #take care of the layout
        self.addReadWidgetToLayout() #this is needed because the writeWidget affects to the readWritget layout
        self.addWriteWidgetToLayout()
        
        if self._writeWidget is not None:
            #give the new widget a reference to its buddy TauValue object
            self._writeWidget.tauValueBuddy = self
            
            #tweak the new widget
            ##hide getPendingOperations of the writeWidget so that containers don't get duplicate lists
            #self._writeWidget._getPendingOperations = self._writeWidget.getPendingOperations 
            #self._writeWidget.getPendingOperations = lambda : [] 
            self.connect(self._writeWidget, Qt.SIGNAL('valueChanged'),self.updatePendingOpsStyle)
            self._writeWidget.setDangerMessage(self.getDangerMessage())
            self._writeWidget.setForceDangerousOperations(self.getForceDangerousOperations())
            if self.minimumHeight() is not None:
                self._writeWidget.setMinimumHeight(self.minimumHeight())
            
            #set the model for the subwidget
            if hasattr(self._writeWidget,'setModel'):
                self._writeWidget.setModel(self.getModelName())
        
    def updateUnitsWidget(self):
        #get the class for the widget and replace it if necessary
        klass = self.unitsWidgetClassFactory(self.unitsWidgetClassID)
        self._unitsWidget = self._newSubwidget(self._unitsWidget, klass)
        
        #take care of the layout
        self.addUnitsWidgetToLayout() 
        
        if self._unitsWidget is not None:
            #give the new widget a reference to its buddy TauValue object
            self._unitsWidget.tauValueBuddy = self
            #tweak the new widget
            if self.minimumHeight() is not None:
                self._unitsWidget.setMinimumHeight(self.minimumHeight())
            
            #set the model for the subwidget
            if hasattr(self._unitsWidget,'setModel'):
                self._unitsWidget.setModel(self.getModelName())

    def addLabelWidgetToLayout(self):
        if self._labelWidget is not None:
            self._parentLayout.addWidget(self._labelWidget, self._row, 1)
    
    def addReadWidgetToLayout(self):
        if self._readWidget is not None: 
            if self._writeWidget is None:
                self._parentLayout.addWidget(self._readWidget, self._row, 2,1,2)
            else:
                self._parentLayout.addWidget(self._readWidget, self._row, 2)
    
    def addWriteWidgetToLayout(self):
        if self._writeWidget is not None:
            self._parentLayout.addWidget(self._writeWidget, self._row, 3)
    
    def addUnitsWidgetToLayout(self):
        if self._unitsWidget is not None:
            self._parentLayout.addWidget(self._unitsWidget, self._row, 4)

    @Qt.pyqtSignature("parentModelChanged(const QString &)")
    def parentModelChanged(self, parentmodel_name):
        """Invoked when the parent model changes
        
        :param parentmodel_name: (str) the new name of the parent model
        """
        TauBaseWidget.parentModelChanged(self, parentmodel_name)
        if not self._designMode:     #in design mode, no subwidgets are created
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()

    @Qt.pyqtSignature("setLabelWidget(QString)")
    def setLabelWidgetClass(self,classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TauWidget class name, or any class'''
        self.labelWidgetClassID = classID
        self.updateLabelWidget()
            
    def getLabelWidgetClass(self):
        return self.labelWidgetClassID
    
    def resetLabelWidgetClass(self):
        self.labelWidgetClassID = 'Auto'
    
    @Qt.pyqtSignature("setReadWidget(QString)")
    def setReadWidgetClass(self,classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TauWidget class name, or any class'''
        self.readWidgetClassID = classID
        self.updateReadWidget()
            
    def getReadWidgetClass(self):
        return self.readWidgetClassID
    
    def resetReadWidgetClass(self):
        self.readWidgetClassID = 'Auto'
        
    @Qt.pyqtSignature("setWriteWidget(QString)")
    def setWriteWidgetClass(self,classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TauWidget class name, or any class'''
        self.writeWidgetClassID = classID
        self.updateWriteWidget()
    
    def getWriteWidgetClass(self):
        return self.writeWidgetClassID
    
    def resetWriteWidgetClass(self):
        self.writeWidgetClassID = 'Auto'
        
    @Qt.pyqtSignature("setUnitsWidget(QString)")
    def setUnitsWidgetClass(self,classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TauWidget class name, or any class'''
        self.unitsWidgetClassID = classID
        self.updateUnitsWidget()
    
    def getUnitsWidgetClass(self):
        return self.unitsWidgetClassID
    
    def resetUnitsWidgetClass(self):
        self.unitsWidgetClassID = 'Auto'
    
        
    def isReadOnly(self):
        if not self.getAllowWrite(): return True 
        modelObj = self.getModelObj()
        if modelObj is None: return False 
        return not modelObj.isWritable()

    def getModelClass(self):
        import tau.core
        return tau.core.TauAttribute
    
    def destroy(self):
        if not self._designMode:
            for w in [self._labelWidget, self._readWidget, self._writeWidget, self._unitsWidget]:
                if isinstance(w,Qt.QWidget):
                    w.setParent(self)   #reclaim the parental rights over subwidgets before destruction
        Qt.QWidget.setParent(self,None)
        Qt.QWidget.destroy(self)

    def createConfig(self, allowUnpickable=False):
        '''
        extending  :meth:`TauBaseWidget.createConfig` to store also the class names for subwidgets
               
        :param alllowUnpickable:  (bool) 
        
        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).
        
        .. seealso: :meth:`TauBaseWidget.createConfig`, :meth:`applyConfig` 
        '''
        configdict = TauBaseWidget.createConfig(self, allowUnpickable=allowUnpickable)
        #store the subwidgets classIDs
        for key in ('LabelWidgetClass', 'ReadWidgetClass', 'WriteWidgetClass', 'UnitsWidgetClass'):
            classID = getattr(self, 'get%s'%key)() # calls self.getLabelWidgetClass, self.getReadWidgetClass,...
            if isinstance(classID, (str, Qt.QString)) or allowUnpickable:
                configdict[key] = classID
            else:
                self.info('createConfig: %s not saved because it is not Pickable (%s)'%(key, str(classID)))
        
        return configdict
    
    def applyConfig(self, configdict, **kwargs):
        """extending :meth:`TauBaseWidget.applyConfig` to restore the subwidget's classes
        
        :param configdict: (dict)
        
        .. seealso:: :meth:`TauBaseWidget.applyConfig`, :meth:`createConfig`
        """
        #first do the basic stuff...
        TauBaseWidget.applyConfig(self, configdict, **kwargs)
        #restore the subwidgets classIDs
        for key in ('LabelWidgetClass', 'ReadWidgetClass', 'WriteWidgetClass', 'UnitsWidgetClass'):
            if key in configdict:
                getattr(self, 'set%s'%key)(configdict[key])
        
    @Qt.pyqtSignature("setModel(QString)")
    def setModel(self, model):
        model = str(model)
        TauBaseWidget.setModel(self,model)
        if not self._designMode:     #in design mode, no subwidgets are created
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
        
    def isValueChangedByUser(self):
        try:
            return self._writeWidget.isValueChangedByUser()
        except AttributeError:
            return False
    
    def setDangerMessage(self, dangerMessage=None):
        TauBaseWidget.setDangerMessage(self, dangerMessage)
        try:
            return self._writeWidget.setDangerMessage(dangerMessage)
        except AttributeError:
            pass
    
    def setForceDangerousOperations(self, yesno):
        TauBaseWidget.setForceDangerousOperations(self, yesno)
        try:
            return self._writeWidget.setForceDangerousOperations(yesno)
        except AttributeError:
            pass
        
    def hasPendingOperations(self):
        '''self.getPendingOperations will alwaysd return an empty list, but still
        self.hasPendingOperations will look at the writeWidget's operations.
        If you want to ask the TauValue for its pending operations, call
        self.writeWidget().getPendingOperations()'''
        w = self.writeWidget()
        if w is None: return []
        return w.hasPendingOperations()
        
    
    def updatePendingOpsStyle(self):
        if self._labelWidget is None: return
        if self.hasPendingOperations():
            self._labelWidget.setStyleSheet(
                'border-style: solid ; border-width: 1px; border-color: blue; color: blue; border-radius:4px;')
        else:
            self._labelWidget.setStyleSheet(
                'border-style: solid; border-width: 1px; border-color: transparent; color: black;  border-radius:4px;')
            
    def getLabelConfig(self):
        return self._labelConfig
    
    @Qt.pyqtSignature("setLabelConfig(QString)")
    def setLabelConfig(self, config):
        self._labelConfig = config
        self.updateLabelWidget()
        
    def resetLabelConfig(self):
        self._labelConfig = 'label'
        self.updateLabelWidget()
    
    
    ########################################################
    ## Qt properties (for designer)
    model = Qt.pyqtProperty("QString", TauBaseWidget.getModel,  setModel, TauBaseWidget.resetModel)
    preferredRow = Qt.pyqtProperty("int", getPreferredRow, setPreferredRow, resetPreferredRow)
    labelWidgetClass = Qt.pyqtProperty("QString", getLabelWidgetClass, setLabelWidgetClass, resetLabelWidgetClass)
    readWidgetClass = Qt.pyqtProperty("QString", getReadWidgetClass, setReadWidgetClass, resetReadWidgetClass)
    writeWidgetClass = Qt.pyqtProperty("QString", getWriteWidgetClass, setWriteWidgetClass, resetWriteWidgetClass)
    unitsWidgetClass = Qt.pyqtProperty("QString", getUnitsWidgetClass, setUnitsWidgetClass, resetUnitsWidgetClass)
    labelConfig = Qt.pyqtProperty("QString", getLabelConfig, setLabelConfig, resetLabelConfig)
    allowWrite = Qt.pyqtProperty("bool", getAllowWrite, setAllowWrite, resetAllowWrite)
    modifiableByUser = Qt.pyqtProperty("bool", TauBaseWidget.isModifiableByUser, TauBaseWidget.setModifiableByUser, TauBaseWidget.resetModifiableByUser)

class TauValuesFrame(TauFrame):
    '''This is a container specialiced into containing TauValue widgets.
    It should be used Only for TauValues'''
    
    _model = Qt.QStringList()
    @Qt.pyqtSignature("setModel(QStringList)")    
    def setModel(self, model):
        self._model = model
        for tv in self.getTauValues():
            tv.destroy()
        for m in self._model:
            tauvalue = TauValue(self, self.designMode)
            tauvalue.setMinimumHeight(20)
            tauvalue.setModel(str(m))
            tauvalue.setModifiableByUser(self.isModifiableByUser())
    def getModel(self):
        return self._model
                
    def resetModel(self):
        self.setModel(Qt.QStringList())
    
    def getTauValueByIndex(self, index):
        '''returns the TauValue item at the given index position'''
        return self.getTauValues()[index]
    
    def getTauValues(self):
        '''returns the list of TauValue Objects contained by this frame'''
        return [obj for obj in self.children() if isinstance(obj,TauValue)]
    

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties 
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
   
    model = Qt.pyqtProperty("QStringList", getModel, setModel, resetModel)


#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
if __name__ == "__main__":
    
    
    import sys

    app = Qt.QApplication(sys.argv)
    form = Qt.QMainWindow()
    #ly=Qt.QVBoxLayout(form)
    #container=Qt.QWidget()
    container = TauValuesFrame()
    container.setModifiableByUser(True)
    #from tau.widget import TauForm
    #container = TauForm()
    #ly.addWidget(container)
    form.setCentralWidget(container)
    
    ##set a model list  
    if len(sys.argv)>1:
        models=sys.argv[1:]
        container.setModel(models)
    else:
        models = []
        
    #models assigned for debugging.... comment out when releasing
    #models=['bl97/pc/dummy-01/Current','bl97/pysignalsimulator/1/value1','bl97/pc/dummy-02/RemoteMode','bl97/pc/dummy-02/CurrentSetpoint','bl97/pc/dummy-02/Current']
    #models=['bl97/pc/dummy-01/Current']
    #models=['bl97/pc/dummy-01/CurrentSetpoint','bl97/pc/dummy-02/Current','bl97/pc/dummy-02/RemoteMode','bl97/pysignalsimulator/1/value1']
    #models=['bl97/pc/dummy-01/CurrentSetpoint','bl97/pc/dummy-02/RemoteMode']
    #container.setModel(models)
    
    #container.getTauValueByIndex(0).writeWidget().setDangerMessage('BOOO') #uncomment to test the dangerous operation support
    #container.getTauValueByIndex(0).readWidget().setShowState(True)
    #container.getTauValueByIndex(0).setWriteWidgetClass(TauValueLineEdit)
    
    form.show()
    
    #show an Attributechooser dialog if no model was given
    if models == []:
        from tau.widget.dialog import AttrChooser
        attrChooser = AttrChooser()
        form.connect(attrChooser, Qt.SIGNAL("UpdateAttrs"), container.setModel)
        attrChooser.show()
    
    sys.exit(app.exec_())